function evalinfo = evalbatch(cfg, algcfg)
% Evaluate a batch of DP/RL experiments, plotting results

% TODO implement: console output
% Also check for convergence/divergence etc?

CFG.alg = [];                       % currently supported: llrfittedqi
CFG.startat = 1;                    % offset for experiment index
CFG.repeat = 1;                     % #experiment repetitions
CFG.out = [];                       % plot this quantity
CFG.in = [];                        % as a function of this quantity
% in and out may be given as structures with specific configurations
% the priority is: default values from below, overwritten by parameter defaults, overwritten by
% these explicit settings
CFG.distribute = {};                % instead of being treated as expanded, these params should be
    % distributed one-by-one to the experiments; the same number of values as the number of
    % experiments must be provided
CFG.type = 'exp';                   
    % 'exp': plot across parameter 'in'; it should be the only one expanded
    % 'inner': plot along inner index, such as time for online algo
    % 'opt': plot across parameter 'in', and for each of its values, optimize across the other parameters
% experiment names and data saving configuration
CFG.idfields = [];                  % include ONLY these fields in the experiment id 
    % (by default, alg name, all showable&set parameters, and the repetition index are shown)
CFG.idprefix = '';
CFG.idinfix = '';
CFG.idpostfix = '';
CFG.verb = 3;
    % >= 0: display console output
    % 1, 2, 3: display details about the evaluation, with increasing verbosity
% cache
CFG.cachefile = 'd:\work\m\approxrl\batch\cache.mat';
CFG.skipcache = 0;                  % completely ignore cache: don't load, don't update
CFG.overridecache = 0;              % override cache and recompute anyway (saving to cache afterwards)
% output config
CFG = setfigprop(CFG, 'addfields');
CFG.legendprefix = [];              % add this prefix to legend entries
CFG.plotindex = 0;                 % plot style index
CFG.plotconf = 2;                   % confidence interval: 0 = nothing, 1 = edges, 2 = edges + shading
% support for global transforms was removed October 2012
% CFG.intransform = [];               % if set, this transform will be applied to the input
%     % note that if the variable has a transform of its own, /that/ transform will be applied first
% CFG.outtransform = [];              % same, for the output

cfg = parseconfig(cfg, CFG);
dispx('EVALBATCH will run with the following config:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1, 'cfg', 'config');

% #### 1. Initialize ####

algfun = eval(['@' cfg.alg '_exp']);

% get algorithm params and process algorithm configs
PARAMS = algfun('params');
[params, expconfigs, used, expanded] = processparams(algcfg, PARAMS, cfg.distribute);
nexp = length(expconfigs);
% get possible outputs for this algorithm
OUTPUTS = algfun('outputs');

% get input and output ID allowing for structure configs
if isstruct(cfg.in), inid = cfg.in.id; else inid = cfg.in; end;
if isstruct(cfg.out), outid = cfg.out.id; else outid = cfg.out; end;

switch cfg.type,
    case 'exp',
        if length(expanded) > 1,
            error('EVALBATCH: [exp] mode can only plot across one expanded parameter.');
        elseif ~strcmp(inid, expanded),
            error('EVALBATCH: [exp] mode: input is not among expanded parameters.');
        end;
    case 'inner',
        INPUTS = algfun('inputs');      % note algos that don't use inner do not need to implement this
        if length(expanded) > 0,
            error('EVALBATCH: [inner] mode can only work for a single experiment');
        elseif ~isfield(INPUTS, inid),
            error('EVALBATCH: [inner] mode: input variable unknown');
        end;
    case 'opt',
        if length(expanded) < 2 || ~any(strcmp(inid, expanded)),
            error('EVALBATCH: [opt] must plot along one expanded param, with at least another param expanded.');
        end;
    otherwise,
        error('EVALBATCH: Evaluation type %s unknown.', cfg.type);
end;
if isfield(cfg, 'intransform') || isfield(cfg, 'outtransform'), 
    error('Global transforms not supported, implement transform for that parameter'); 
end;

% default config for input (horizontal axis) (in addition to standard parameter config) 
IN.id = [];          % set according to param id
IN.label = [];       % set by default to id
IN.units = [];       % no units
IN.transform = [];   % no transformation applied
IN.scale = 'lin';    % linear scale (options: lin, log)

% default config for output (vertical axes and plot styles)
OUT = IN;
OUT.conflabel = '95% confidence bound';
OUT.params = {};     % params required on the config
% mean style (or complete curve for deterministic experiments)
OUT.meansty = {{'-', 'Color', [.4 .4 .4], 'LineWidth', 2, 'Marker', 'o'}, ...
    {'k-', 'Marker', 's'}, ....
    {'k-', 'Marker', 'x'}, ...
    {'k--', 'LineWidth', 2, 'Marker', '^'}, ... % starting here, no markers
    {'-', 'Color', [.4 .4 .4], 'LineWidth', 2}, ...
    {'k-'}, ....
    {'k--'}, ...
    {'k:'}};
% lower (confidence or minimum) bound 
OUT.lowersty = {{'k--'}, {'k--'}, {'k--'}, {'k--'}, {'k:'}, {'k:'}, {'k:'}, {'k-.'}};
% upper (confidence or maximum) bound 
OUT.uppersty = [];      % use lower by default
% shading of min-max or confidence interval
OUT.intervalsty = {[.7 .7 .7], 'EdgeColor' , 'none'};

% prepare in and out variables
switch cfg.type,
    % process input based on whether it's a structure or just a parameter id
    case {'exp', 'opt'},    in = processvar(params.(inid), IN, 'in');
    case 'inner',           in = processvar(INPUTS.(inid), IN, 'in');
end;
if isstruct(cfg.in), in = checkparams(cfg.in, in); end;
out = processvar(OUTPUTS.(outid), OUT, 'out');
if isstruct(cfg.out), out = checkparams(out, cfg.out); end;

% prepare cache
% silence var not found warning - we rely on that to determine if data is cached
warning('off', 'MATLAB:load:variableNotFound');
cache_newdata = [];

% #### 2. Compute output ####

% process experiments, evaluating the output
if any(strcmp(cfg.type, {'exp', 'opt'})), 
    out.val = nan(cfg.repeat, nexp);     % in 'inner' mode it will be dynamically initialized
end;    
for rep = cfg.startat:(cfg.startat+cfg.repeat-1),
    dispx(sprintf('Evaluating set%d, type [%s]', rep, cfg.type), cfg.verb, 2);
    for iexp = 1:nexp,        
        % pick up experiment config
        expcfg = expconfigs{iexp};
        % generate ID; this is also the datafile name
        expid = generateexpid(cfg, params, expcfg, rep);
        expcfg.datafile = expid;
        % check if datafile exists
        if ~(exist([expid '.mat'], 'file')), 
            error('Set%d/exp%d, id=%s: datafile not found!', rep, iexp, expid);
        end;
        % FIXME: note these fields do not get propagate to cache content if it was saved on an
        % earlier run, for a different output variable
        % add necessary configuration fields on the experiment config
        if ~isempty(out.params), expcfg = copyfields(algcfg, expcfg, out.params); end;
        
        % check if data is available on the cache
        if ~cfg.skipcache,
            cache_id = expidtocacheid(expid);
            if exist(cfg.cachefile, 'file'),
                lastwarn(''); load(cfg.cachefile, cache_id);
                if isempty(lastwarn), 
                    cache_content = eval(cache_id);
                    if isfield(cache_content, out.id), 
                        cache_status = 1;   % cache field found
                    else cache_status = 0;  % cache variable found, but field not found
                    end;
                else cache_status = -1;     % cache variable not found
                end;
            else cache_status = -2;         % cache file not found
            end;
            % cache status message
            if cache_status == 1 && ~cfg.overridecache,
                dispx(sprintf('Set%d/exp%d, id=%s: using cache.', rep, iexp, expid), cfg.verb, 4);
            elseif cache_status == 1 && cfg.overridecache,
                dispx(sprintf('Set%d/exp%d, id=%s: overriding cache.', rep, iexp, expid), cfg.verb, 4);
            else
                dispx(sprintf('Set%d/exp%d, id=%s: not cached.', rep, iexp, expid), cfg.verb, 4);
            end;
            % create cache variable if needed
            if cache_status <= -1, cache_content = expcfg; cache_content.rep = 1; end;
        end;
        
        % retrieve or compute data
        if ~cfg.skipcache && cache_status == 1 && ~cfg.overridecache,
            % retrieve cached data (also the input is relevant if type 'inner')
            outval = cache_content.(out.id);
            if cfg.type(1) == 'i', inval = cache_content.(in.id); end;
        else
            % call experiment handling function to evaluate
            % note behavior is different for exp and inner
            dispx(sprintf('Evaluating set%d/exp%d, id=%s', rep, iexp, expid), cfg.verb, 3);
            if cfg.type(1) == 'e' || cfg.type(1) == 'o',      % exp type
                outval = algfun(out.id, expcfg);
                % flag for saving to cache
                if ~cfg.skipcache,
                    cache_content.(out.id) = outval;
                    cache_newdata.(cache_id) = cache_content;
                end;
            elseif cfg.type(1) == 'i',  % inner type
                [inval outval] = algfun(out.id, expcfg);
                % flag for saving to cache
                if ~cfg.skipcache,
                    cache_content.(in.id) = inval;
                    cache_content.(out.id) = outval;
                    cache_newdata.(cache_id) = cache_content;
                end;
            end;
        end;
        % save data on input/output structures
        if cfg.type(1) == 'e' || cfg.type(1) == 'o',      % exp type
            out.val(rep, iexp) = outval;
        elseif cfg.type(1) == 'i',  % inner type
            if isempty(in.val), 
                % first encounter
                disp('Init values');
                in.val = inval;
                out.val = nan(cfg.repeat, nexp, length(in.val));
            else
                if ~isequal(in.val, inval), error('EVALBATCH: Mismatch in input length for type [inner]'); end;
            end;
            out.val(rep, iexp, :) = outval;
        end;        
    end;    % FOR experiments
end;        % FOR repetition

% update cache if modified
if ~cfg.skipcache && ~isempty(cache_newdata),
    % save non-compressed (v6) for fast loading
    dispx('Saving new data to cache...', cfg.verb, 2);
    cache_ids = fieldnames(cache_newdata); structtovars(cache_newdata);
    if exist(cfg.cachefile, 'file'),    save(cfg.cachefile, cache_ids{:}, '-append', '-v6'); 
    else                                save(cfg.cachefile, cache_ids{:}, '-v6'); 
    end;
	clear(cache_ids{:});
    dispx([8 ' saved.'], cfg.verb, 2);
    % restore var not found warning
    warning('on', 'MATLAB:load:variableNotFound');
end;

% finalize the output data dimensions
switch cfg.type,
    case {'exp', 'opt'}
    	if size(out.val, 3) > 1, error('EVALBATCH: Cannot have vector values per experiment in type [exp/opt]'); end;
        nval = nexp;
        out.val = reshape(out.val, cfg.repeat, nval);
    case 'inner',
    	if size(out.val, 2) > 1, error('EVALBATCH: Cannot have several experiments in type [inner]'); end;
        if any(isnan(out.val(:))), error('EVALBATCH: NaN values on out.val -- size mismatches?'); end;
        nval = size(out.val, 3);
        out.val = reshape(out.val, cfg.repeat, nval);
end;

% 1 oct 2014 -- original location of in transform was here, but when parameters are optimized,
% the data cannot be found anymore -- so transform was moved after processing
% transform variables if needed - note transform functions must be vectorized
if ~isempty(out.transform), out.val = feval(out.transform, out.val); end;
% if ~isempty(cfg.intransform), in.val = feval(cfg.intransform, in.val); disp('check code'); end;
% if ~isempty(cfg.outtransform), out.val = feval(cfg.outtransform, out.val); error('check code'); end;
% compute mean and confidence interval if more than one repetition
if cfg.repeat > 1,
    out.mean = mean(out.val);
    out.conf = zeros(2, nval);
    % find confidence interval, taking into account that all values may be identical
    for ival = 1:nval,
        if all(out.mean(ival) == out.val(:, ival)), 
            out.conf(:, ival) = [out.mean(ival); out.mean(ival)];
        else
            [h, p, out.conf(:, ival)] = ttest(out.val(:, ival));
        end;
    end;
    % below assumes normal distribution (so is not very appropriate for small sample sizes)
    % confwidth = 1.96*std(out.val) ./ sqrt(cfg.repeat);
    % out.conf = [out.mean - confwidth; out.mean + confwidth];
end;

% optimize if mode 'opt'
if strcmp(cfg.type, 'opt'),
    optimizedids = setdiff(expanded, in.id);    % expanded but not plotting across them
    expindices = cell(1, length(in.val));   % where exp with each value of 'in' are found
    
    % pick up original values for all experiments, prepare the optimized values
    ninval = length(in.val);
    allvals = out.val;
    out.val = nan(cfg.repeat, ninval);
    if cfg.repeat > 1,
        allmean = out.mean; allconf = out.conf;
        out.mean = nan(1, ninval); out.conf = nan(2, ninval);
    end;
    optindices = nan(1, ninval);
    optparams = cell(1, ninval);
    
    % prepare console out
    cout = cell(length(optimizedids), ninval+1);
    for i = 1:length(optimizedids), cout{i, 1} = params.(optimizedids{i}).shortid; end;
    
    % run through the input values
    for ival = 1:ninval,
        inputvalue = in.val(ival);       % note this test is limited to vector values!!!
        % find the experiments matching this input value
        for iexp = 1:nexp,  
            if isequal(eval(sprintf('expconfigs{iexp}.%s', in.path)), inputvalue), 
                expindices{ival}(end+1) = iexp; 
            end;
        end;
        % find the best value across the other parameters,
        % updating the data along the way
        if cfg.repeat == 1,
            % find the largest output value for this input value
            [out.val(ival), i2max] = max(allvals(expindices{ival})); 
            optindices(ival) = expindices{ival}(i2max);
        else
            % find the largest output UCB for this input value
            [dummy, i2max] = max(allconf(2, expindices{ival}));
            optindices(ival) = expindices{ival}(i2max);
            out.val(:, ival) = allvals(:, optindices(ival));
            out.mean(ival) = allmean(optindices(ival));
            out.conf(:, ival) = allconf(:, (optindices(ival)));
        end;
        % save param values for the best experiment
        for i = 1:length(optimizedids),
            optparams{ival} = struct;
            paramvalue = eval(sprintf('expconfigs{optindices(ival)}.%s', params.(optimizedids{i}).path));
            optparams{ival}.(optimizedids{i}) = paramvalue;
            % add to console out
            cout{i, ival+1} = paramvalue;
        end;
    end;
end;

% console output 
if cfg.verb >= 0,
    disp(['Experiment ' cfg.idinfix]);
    switch cfg.type,
        case 'exp',
            disp('For inputs:')
            invalues = in.val
            if cfg.repeat == 1, 
                disp('The following values are obtained:')
                outvalues = out.val
            else
                disp('The following outputs are optimal [mean, +-conf]:')
                outvalues = [out.mean; diff(out.conf)./2]
            end;
        case 'opt',
            disp('For inputs:')
            invalues = in.val
            if cfg.repeat == 1, 
                disp('The following outputs are optimal:')
                outvalues = out.val
            else
                disp('The following outputs are optimal [mean, +-conf]:')
                outvalues = [out.mean; diff(out.conf)./2]
            end;
            disp('and are obtained for the parameter values:');
            cout
    end;
end;

% #### 3. Plot output ####

% format inputs
% transform if needed
if ~isempty(in.transform), in.val = feval(in.transform, in.val);  end;

% if input is non-numeric, or a format function is provided, process using it
if any(strcmp(in.type, {'char', 'vector', 'cell'})) || isfield(in, 'formatfun'),
    % make numeric vectors into cells for uniform treatment
    if isnumeric(in.val), in.val = num2cell(in.val); end;
    xval = 1:length(in.val);
    xticklabels = cell(length(in.val), 1);
    for i = 1:length(in.val),
        [fid, xticklabels{i}] = formatparam(in, in.val{i}); 
        if isempty(fid) && isempty(xticklabels{i}), error('EVALBATCH: Input value cannot be formatted!'); end;
    end;
    clear fid;
elseif any(strcmp(in.type, {'generic', 'scalar'})),
    % assume it's a numeric vector and hope for the best...
    xticklabels = [];
    xval = in.val;      
end;

dispx('Plotting data...', cfg.verb, 1);

% choose plot function
switch [in.scale out.scale],
    case 'linlin', plotfun = @plot;
    case 'linlog', plotfun = @semilogy;
    case 'loglin', plotfun = @semilogx;
    case 'loglog', plotfun = @loglog;
end;

% create or get figure
if cfg.addtocrt, figh = gcf; else figh = figure; end;

% choose style index
nstyles = length(out.meansty);
if cfg.plotindex == 0,             % a default style if value not explicitly set
    if cfg.addtocrt 
        if (nstyles > 1),   cfg.plotindex = 1 + unidrnd(nstyles - 1); % choose at random
        else                cfg.plotindex = 1;  % just one style available
        end;
    else cfg.plotindex = 1; % not adding to current figure -- use first style
    end;
end;
cfg.plotindex = mod(cfg.plotindex - 1, nstyles) + 1; % make sure within allowable range

lh = []; leg = {}; 
if ~isempty(cfg.legendprefix), legpref = [cfg.legendprefix ', ']; else legpref = ''; end;

% plot data remembering legend entries
if cfg.repeat == 1,
    % plot values
    lh(end+1) = plotfun(xval, out.val, out.meansty{cfg.plotindex}{:});
    leg{end+1} = [legpref out.label]; 
    hold on; 
else    % cfg.repeat > 1
    % plot mean values
    lh(end+1) = plotfun(xval, out.mean, out.meansty{cfg.plotindex}{:});
    leg{end+1} = [legpref 'mean ' out.label]; 
    hold on; 
    % plot confidence level, if several repetitions and configured to do so
    if cfg.plotconf >= 1,
        plotfun(xval, out.conf(1, :), out.lowersty{cfg.plotindex}{:}); 
        plotfun(xval, out.conf(2, :), out.uppersty{cfg.plotindex}{:});
        % replace list line by the two lines below to get a legend label for the confidence bound
        % note only one boundary gets a legend (since both have same style & meaning)
        % lh(end+1) = plotfun(xval, out.conf(2, :), out.uppersty{cfg.plotindex}{:});
        % leg{end+1} = [legpref out.conflabel]; 
        % add shading if configured to do so -- note negative z coordinate to prevent overlaps with lines
        if cfg.plotconf >= 2,
            patch([xval fliplr(xval)], [out.conf(1, :) fliplr(out.conf(2, :))], ...
                -.1 * [xval xval], out.intervalsty{:});
        end;
    end;
end;

% set various labels
if isempty(cfg.addtocrt) || ~cfg.addtocrt,   % create legend and set axis labels
    lhh = legend(lh, leg{:});
    set(lhh, 'Location', 'Best'); 
    if ~isempty(in.units),  xlabel(sprintf('%s [%s]', in.label, in.units));
    else                    xlabel(in.label);
    end;
    if ~isempty(out.units), ylabel(sprintf('%s [%s]', out.label, out.units));
    else                    ylabel(out.label);
    end;
    if ~isempty(xticklabels), set(gca, 'XTick', xval, 'XTickLabel', xticklabels); end;
else                % add to already existing legend (do not overwrite axis labels)
    [dummy1 dummy2 oldobjecth oldleg] = legend; clear dummy1 dummy2;
    if ~isempty(oldleg), lhh = legend([oldobjecth(:); lh(:)], oldleg{:}, leg{:}); 
    else lhh = legend(lh, leg{:}); end;
    set(lhh, 'Location', 'Best');
end;
% set some informative figure title
cfg.figname = sprintf('%s - %s %s', cfg.alg, algcfg.problem, cfg.legendprefix);

% finalize and possibly save figure 
setfigprop(cfg);
saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);

% final message
dispx([8 ' done.'], cfg.verb, 1);

% output input and output data
evalinfo.figh = figh;
evalinfo.in = in;
evalinfo.out = out;

end
% end evalbatch function


% #########################
% #### Local functions ####


function v = processvar(v, V, inouttype)
% process variable style (add defaults etc.)
fields = fieldnames(V);
for i = 1:length(fields), 
    if ~isfield(v, fields{i}), v.(fields{i}) = V.(fields{i}); end;
end;
if isempty(v.label), v.label = v.id; end;
switch inouttype,
    case 'out', 
        if isempty(v.uppersty), v.uppersty = v.lowersty; end;
end;
end

function v = expidtocacheid(v)
% transform experiment ID to cache variable ID (name) -- remove illegal characters
invalidchars = '.# ,/?^';    % the rest of the characters are not currently used
for ic = invalidchars, v = v(v ~= ic); end;
end