function bruteforceopt(cfg)
% Brute-force optimal control computation, for a given set of initial states
% Requires episodic task, discrete set of actions

if nargin < 1, cfg = struct; end;

% Default config
CFG.problem = '';
CFG.gamma = [];
CFG.datafile = 'bruteforceopt';
CFG.steps = 10;
CFG.progdisp = .05;
CFG.X0 = [];
CFG.model_params = {};
CFG.verb = 5;
CFG.silent = 0;
CFG.usesubseq = 1;
CFG.improvedatafile = [];

if ischar(cfg), cfg = str2cfg(cfg, fieldnames(CFG)); end;
cfg = checkparams(cfg, CFG);
if cfg.silent, cfg.verb = -Inf; end;
cfg.improve = ~isempty(cfg.improvedatafile);
cfg

% create the model
model = feval(cfg.problem, 'model', cfg.model_params{:});
% get the set of initial states
if ischar(cfg.X0),
    cecfg = feval(cfg.problem, 'ce', cfg.X0);
    X0 = cecfg.X0;
else  % don't care about X0, just get U
    cecfg = feval(cfg.problem, 'ce');
    X0 = cfg.X0;
end;
% shorthand vars
U = cecfg.U; gamma = cecfg.gamma; Kmax = cfg.steps;
M = size(U, 2);     % number of actions
% precompute discounting vector
disc = cumprod([1 gamma+zeros(1, Kmax)]);

% init tables for all x0
N0 = size(X0, 2);
JSTAR = zeros(N0, 1); KSTAR = zeros(N0, 1); PSTAR = nan(N0, Kmax);
USTAR = nan(N0, model.q, Kmax);
XSTAR = nan(N0, model.p, Kmax+1); RSTAR = nan(N0, Kmax);
TERM = false(N0, 1);
found = false(N0, 1);     % logical vector storing "optimal found" flags

if cfg.improve,     % improve on existing results by lengthening sequence
    dispx('Loading earlier data', cfg.verb, 1);
    vars = {'JSTAR', 'KSTAR', 'PSTAR', 'USTAR', 'XSTAR', 'TERM', 'RSTAR', 'found', 'cfg'};
    % this will give a warning if TERM was not set in the old datafile
    loadwithprefix(cfg.improvedatafile, 'old', vars);
    
    % these will give errors unless dimensions are correct; that is intended
    % to be so
    oldsteps = oldcfg.steps;
    JSTAR(:) = oldJSTAR;
    KSTAR(:) = oldKSTAR;
    PSTAR(:, 1:oldsteps) = oldPSTAR;        % the rest remain NaN
    USTAR(:, :, 1:oldsteps) = oldUSTAR;     % the rest remain NaN
    XSTAR(:, :, 1:oldsteps+1) = oldXSTAR;   % the rest remain NaN
    RSTAR(:, 1:oldsteps) = oldRSTAR;     % the rest remain NaN
    found = oldfound;                       % this should be all true if run was completed
    if exist('oldTERM', 'var'),
        TERM(:) = oldTERM;
    else        % support for datafiles without TERM flags
        % Safe choice: trajectories for which the length was equal to the max
        % # of steps are assumed not terminated
        TERM(KSTAR == oldsteps) = false;
    end;
    % this will force recomputation for un-terminated states
    found(~TERM) = false;
    % clear loaded variables
    for i = 1:length(vars), clear(['old' vars{i}]); end;
    dispx(['Improving for ' num2str(sum(~found)) ' states'], cfg.verb, 1);
end;

dispx('Starting optimization', cfg.verb, 1);
prog = cfg.progdisp;
hits = 0;                       % subsequence hits -- subsequences not currently supported
reliable = 1;                   % results reliability flag
unrX0 = zeros(model.p, 0);      % memory for unreliable states
for i0 = 1:N0,
    if found(i0), continue; end;
    % initialize best optimal score
    Jstar = -Inf; clear Xstar Rstar Kstar termstar Pstar Ustar
    % initialize history for current traj, the initial state is always the same
    X = zeros(model.p, Kmax+1); X(:, 1) = X0(:, i0); R = zeros(Kmax, 1);
    % loop over all possible action sequences of length Kmax
    for ip = 0:(M^Kmax-1),
        P = dec2base(ip, M, Kmax) - '0' + 1;   % indices into the discrete actions for each step
        for k = 1:Kmax,
            [X(:, k+1) R(k) terminal] = feval( model.fun, model, X(:, k), U(:, P(k)) );
            if terminal, break; end;    % no other rewards contribute (equivalently, they are all 0)
        end;        % steps loop
        Jx = disc(1:k) * R(1:k);
        if Jx > Jstar,     % replace old optimum with current
            Jstar = Jx;
            Xstar = X(:, 1:k+1); Rstar = R(1:k);
            Kstar = k; termstar = terminal;
            Pstar = P(1:k); Ustar = U(:, Pstar);
        end;
    end;            % policies loop
    found(i0) = true;
    JSTAR(i0) = Jstar;                        % optimal return
    KSTAR(i0) = Kstar;                        % # of steps to termination
    XSTAR(i0, :, 1:Kstar+1) = Xstar;          % state traj
    USTAR(i0, :, 1:Kstar) = Ustar;            % policy / actual actions
    PSTAR(i0, 1:Kstar) = Pstar;               % policy / action indices
    RSTAR(i0, 1:Kstar) = Rstar;               % rewards
    TERM(i0) = termstar;                      % whether best trajectory terminated
    % check if optimal episode terminated
    if ~termstar,
        warning(['Best trajectory did not terminate for x0=[' num2str(X0(:, i0)') ']. Results are unreliable.']); 
        % flag unreliable and memorize unreliable initial state
        reliable = 0;
        unrX0(:, end+1) = X0(:, i0);
    end;
% 	% use Bellman's optimality principle to find optimal subsequences (if
% 	% configured to do so)
%     if cfg.usesubseq,
%         Xstar = X(:, 1:kstar+1, ipstar);
%         Rstar = R(1:kstar, ipstar);
%         remaining = find(~found);
%         for k = 2:kstar,
%             eq = repmat(Xstar(:, k), 1, length(remaining)) == X0(:, remaining);
%             ix = find(all(eq), 1);  % X0 contains unique elements, search for 1 only
%             if ~isempty(ix),
%                 % found a match; determine its original index in X0
%                 f0 = remaining(ix);
%                 % mark as found
%                 found(f0) = true;
%                 % length of optimal sequence
%                 fkstar = kstar - k + 1;
%                 KSTAR(f0) = fkstar;
%                 % the rhs-s below is displaced k steps to the right
%                 JSTAR(f0) = disc(1:fkstar) * Rstar(k:kstar);
%                 PSTAR(f0, 1:fkstar) = P(ipstar, k:kstar);            % policy / action indices
%                 XSTAR(f0, :, 1:fkstar+1) = Xstar(:, k:kstar+1);      % policy / actual actions
%                 % use the just-set PSTAR to retrieve the open-loop actions
%                 USTAR(f0, :, 1:fkstar) = U(:, PSTAR(f0, 1:fkstar));   % policy / actual actions
%                 % all sub-trajectories terminated IFF main trajectory terminated
%                 TERM(f0) = TERM(i0);        
%                 % update the remaining indices
%                 remaining = find(~found);
%                 hits = hits+1;
%             end;
%         end;
%     end;
    if i0 / N0 > prog,
        dispx([num2str(i0/N0 * 100) '% done.'], cfg.verb, 2);
%         dispx([num2str(i0/N0 * 100) '% done, ' num2str(hits) ' optimal subsequences found.'], cfg.verb, 2);
        prog = prog + cfg.progdisp;
        dispx(['Saving intermediate data to ' cfg.datafile], cfg.verb, 1);
        save(cfg.datafile);
    end;
end;                % states loop

dispx(['Optimization complete. Saving to ' cfg.datafile], cfg.verb, 1);
save(cfg.datafile);
    
