function [Q, info] = simpleqlearn(cfg, model)
% Q-learning
%   Q = SIMPLEQLEARN(CFG, MODEL)
%   [Q, INFO] = SIMPLEQLEARN(CFG, MODEL)
% Inputs:
%   CFG             - structure with the fields as commented in the code below
% Outputs:
%   Q               - computed Q-function
%   INFO            - (optional) other information, on a structure

% -----------------------------------------------
% Process configuration structure

% default config
% script config
CFG.datafile = 'qdata';             % save data this file
CFG.gamma = 0.98;                   % discount factor
CFG.alpha = .1;                     % learning rate
CFG.explor = .9;                    % exploration rate
CFG.explordecay = .9;               % exploration decay (per trial)
CFG.lambda = 0;                     % eligibility trace parameter (0 to disable)
CFG.eps = -1;                       % threshold for Q-table convergence (negative to run for maxtrials)
CFG.maxtrials = 100;                % max # trials
CFG.maxsteps = 1000;                % max # steps in a trial
CFG.reset = 'rand';                 % trial reset
CFG.randseed = [];                  % set the random seed (for repeatable results)
% display config
CFG.verb = 2;                       % verbosity: the higher, the more detailed the messages displayed

% TODO somehow, implement the possibility of providing an initial Q-table (in case this is the
% original task)

% install default values for fields not specified on the input config
fn = fieldnames(CFG); 
if isempty(cfg), cfg = struct; end;
for i = 1:length(fn), if ~isfield(cfg, fn{i}), cfg.(fn{i}) = CFG.(fn{i}); end; end;

% Echo config
if cfg.verb >= 1, 
    disp('Q-learning called with the following configuration:');
    cfg
end;

% -----------------------------------------------
% Initialize

% Compute dimensionality data
% dimx = dimensions of each state variable, N total #states; similarly, dimu and M for actions
DIMS.dimx = zeros(model.p, 1); for i = 1:model.p, DIMS.dimx(i) = length(model.X{i}); end;
DIMS.N = prod(DIMS.dimx);
DIMS.dimu = zeros(model.q, 1); for i = 1:model.q, DIMS.dimu(i) = length(model.U{i}); end;
DIMS.M = prod(DIMS.dimu);
Xflat = flat(model.X); 
Uflat = flat(model.U);
p = model.p; q = model.q;

% init Q-function, elig trace, histories etc.
Q = zeros(DIMS.N, DIMS.M);
E = zeros(DIMS.N, DIMS.M);
X = NaN(p, cfg.maxsteps+1, cfg.maxtrials);
U = NaN(q, cfg.maxsteps+1, cfg.maxtrials);
R = NaN(1, cfg.maxsteps+1, cfg.maxtrials);
% Q-function history
Qh = cell(cfg.maxtrials+1, 1);  
Qh{1} = Q;                      % save Q_0 on the stats
deltah = NaN(cfg.maxtrials+1, 1);

trial = 1;
explor = cfg.explor;
alpha = cfg.alpha;

% controlled random seed initialization, if configured
if ~isempty(cfg.randseed), rand('twister', cfg.randseed); end;

% -----------------------------------------------
% Perform Q-learning
if cfg.verb >= 1,
    if cfg.lambda > 0, disp('Performing Q(lambda)-learning...');
    else disp('Performing Q-learning...');
    end;
end;

conv = 0;   % convergence flag
while trial <= cfg.maxtrials && ~conv,       % main loop

    % initialize trial
    k = 1; terminal = 0;
    % reset state
    X(:, 1, trial) = feval(model.fun, 'reset', model);
    % find index of first state
    xind = findflat(X(:, 1, trial), Xflat, 1, 'first');
    % reset elig trace
    E = 0 .* E;
    
    while k <= cfg.maxsteps && ~terminal,        

        % choose action according to current state
        exploring = rand < explor;
        if exploring,   % exploratory action
            uind = unidrnd(DIMS.M);
            % reset elig tr if using it
            if cfg.lambda > 0, E(:) = 0; end;
        else            % greedy action
            [Qmax uind] = find(Q(xind, :) == max(Q(xind, :)));
            if length(uind) > 1, uind = uind(unidrnd(length(uind))); end;  % break any ties
            % decay elig trace if using it
            if cfg.lambda > 0, E = (cfg.gamma * cfg.lambda) .* E; end;
        end;
        % transform action index into into the actual action
        U(:, k, trial) = Uflat(:, uind);

        % apply the action
        [X(:, k+1, trial), R(1, k+1, trial), terminal] = ...
            feval(model.fun, 'step', model, X(:, k, trial), U(:, k, trial), k);
        xindplus = findflat(X(:, k+1, trial), Xflat, 1, 'first');

        % compute temporal difference and update Q-function
        tempdiff = R(1, k+1, trial) + cfg.gamma * max(Q(xindplus, :)) - Q(xind, uind);
        if cfg.lambda > 0,  % eligibility trace based update
            E(xind, uind) = 1;
            Q = Q + (alpha * tempdiff) .* E;
        else                % no elig tr, implement a fast single-element update
            Q(xind, uind) = Q(xind, uind) + alpha * tempdiff;        
        end;

        % visualization
        if cfg.verb >= 3, 
            feval(model.fun, 'visualizestep', model, X(:, k, trial), U(:, k, trial), ...
                X(:, k+1, trial), R(1, k+1, trial), k); 
        end;

        % increment time step
        k = k + 1; 
        % next state index becomes current
        xind = xindplus;
    end;

    % store Q-function at the end of the trial on the history
    Qh{trial+1} = Q;
    % compute max-norm difference
    deltah(trial+1) = max(max(abs(Q - Qh{trial})));
    % convergence if difference smaller than eps
    conv = deltah(trial+1) < cfg.eps;

    % console feedback of algorithm progress
    if cfg.verb >= 2, disp(sprintf('Trial #%d completed, delta=%.3f', trial, deltah(trial+1))); end;

    % increment trial counter
    trial = trial + 1;
    % anneal exploration
    explor = explor * cfg.explordecay;
end;        % while not converged and allowed more iterations

if cfg.verb >= 1,
    if conv,	disp('Convergence detected. Algorithm stopped.');
    else        disp(sprintf('maxtrials=%d exhausted. Algorithm stopped.', cfg.maxtrials));
    end;
end;

% -----------------------------------------------
% Compose "info" variable
info.cfg = cfg;
info.conv = conv;
info.trials = trial - conv;    % subtract 1 if we converged, since we incremented once extra
info.Qh = Qh;
info.deltah = deltah;
info.X = X;
info.U = U;
info.R = R;

save(cfg.datafile);
if cfg.verb >= 1, disp(sprintf('Data was saved to file [%s].', cfg.datafile)); end;

end         % simpleqlearn
