function [eJ, J] = mc_rbfq(c, rad, theta, model, X0, cfg)
% Monte-Carlo evaluation of RBF parameters and corresponding linear weights resulting from
% Q-iteration
%   [EJ, J] = MC_RBFQ(C, RAD, THETA, MODEL, X0, CFG)
% Monte-Carlo evaluation of RBF parameters and corresponding linear weights resulting from
% Q-iteration.
% This is used in two ways:
%   - as a part of cross-entropy optimization of Q-iteration.
%   - called by mc_rbfqi
% In its current form evaluates only deterministic systems.
%
% Parameters:
%   C       - RBF centers, p x N where p state variable size, N number of basis functions
%   RAD     - RBF radii, p x N where p state variable size, N number of basis functions
%   THETA   - optimal parameter vector computed by Q-iteration with the RBF parameters C, RAD.
%       NxM matrix of Q-values.
%   MODEL   - problem model
%   X0      - distribution of initial states. For now, assumed to be uniform over a 
%       set of discrete values, each given in a column of X0
%   CFG     - configuration parameters. Requires fields (subject to change):
%       U, mc_maxsteps, - see cerbfqi for description
%
% Returns:
%   EJ      - expected cost
%   E       - cost of every initial state sample

n0 = size(X0, 2); N = size(c, 2);
J = zeros(n0, 1);
% precompute discounting vector, pre-initialize reward vector
disc = cumprod([1 cfg.gamma+zeros(1, cfg.mc_maxsteps)]);
r = zeros(cfg.mc_maxsteps, 1);

for i = 1:n0,
    x = X0(:, i);
    for k = 1:cfg.mc_maxsteps,
        % compute crisp (approximate) optimal action and apply to system
        Qa = nrbf(x, N, c, rad) * theta;
        [x r(k) terminal] = feval(model.fun, model, x, cfg.U(:, find(Qa == max(Qa), 1)));
        if terminal, break; end;    % no other rewards contribute (equivalently, they are all 0)
    end;
    
    % return of x0 is inner product of discounting and rewards
    J(i) = disc(1:k) * r(1:k);
end;

% expected value is the mean
eJ = mean(J);

% END rbfmceval() RETURNING expected cost eJ, cost vector J ===========================