function  varargout = doosoo(cfg)
% DOOSOO Optimistic optimization with DOO or SOO
%   [USTAR, [INFO, [DATA]]] = DOOSOO(CFG)
% Returns solution, and optionally info about the optimization process and the
% optimization data itself.

% !!!! CHECK IF depth(T) limit is handled correctly

% Author: Lucian Busoniu
% Versions:

CFG.method = '';                % which method to use: 'doo' or 'soo' / SOO not yet supported
% common options
CFG.fun = [];                   % function to optimize
CFG.funargs = {};               % any additional arguments to call the function with (besides the current point)
CFG.n = [];                     % expansion budget
CFG.maxu = [];                  % variable bounds -- will be used to infer number of variables
% DOO options
CFG.diam = [];                  % function to compute diameter of a node's set
% SOO options
CFG.dmax = @(tau) sqrt(tau);    % function dmax(iter), by default sqrt(iter)
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);

% parameter and error checks
if ~any(strcmp(cfg.method, {'doo', 'soo'})), error('Unsupported method [%s]', cfg.method); end;
if isempty(cfg.fun) || isempty(cfg.n) || isempty(cfg.maxu), error('fun, n, and maxu are required'); end;
if cfg.method(1) == 'd' && isempty(cfg.diam), error('diam function required for DOO'); end;

% quick-access to config parameters, and other helper vars
usingsoo = strcmp(cfg.method, 'soo');
q = length(cfg.maxu); 
fun = cfg.fun; funargs = cfg.funargs; diam = cfg.diam;

% generate 2^q (all) combinations of flags 0, 1 up to dimension q 
% jth column contains flags: whether the pth dimension set is the right one or not (i.e. or left)
rightset = logical(flat(num2cell(repmat([0 1], q, 1), 2)));

% start counting time
t = cputime;

% init data structures
maxtreesize = 1 + cfg.n * 2^q;                % number of nodes eventually on the tree
% initlen = 5; initsize = 10;                 % smaller values for debugging
% U intervals: [start; end] X #vars X #nodes
TINT = NaN(2, q, maxtreesize);                    
% center points
TC = NaN(q, maxtreesize);
% center point values
TR = zeros(1, maxtreesize);
% depths
TDEPTH = zeros(1, maxtreesize);
% for DOO, diameters (stored to avoid recomputing which may be somewhat costly)
if ~usingsoo, TDIAM = zeros(1, maxtreesize); end;
% leaf flags
TLEAF = false(1, maxtreesize);     

% initialize with a big set covering the whole space
TINT(:, :, 1) = [-cfg.maxu'; cfg.maxu'];
TC(:, 1) = (TINT(1, :, 1) + TINT(2, :, 1)) ./ 2; 
TR(1) = fun(TC(:, 1), funargs{:});
TDEPTH(1) = 0;
if ~usingsoo, TDIAM(1) = diam(TINT(:, :, 1)); end;
TLEAF(1) = true;
tsize = 1;          % current tree size
tmaxdepth = 0;      % current max depth
n = 0;              % budget of expansions consumed so far
   
    % Nested function to expand a node
    function expand(node)
        for j = 1:2^q,
            % starting points: = to parent if left set, center if right set
            TINT(1, ~rightset(:, j), tsize+j) = TINT(1, ~rightset(:, j), node);
            TINT(1, rightset(:, j), tsize+j) = TC(rightset(:, j), node);
            % end points: center if left set, = to parent if right set
            TINT(2, ~rightset(:, j), tsize+j) = TC(~rightset(:, j), node);
            TINT(2, rightset(:, j), tsize+j) = TINT(2, rightset(:, j), node);
            % finalize node: compute center, evaluate, diam, set leaf flag
            TC(:, tsize+j) = (TINT(1, :, tsize+j) + TINT(2, :, tsize+j)) ./ 2; 
            TR(tsize+j) = fun(TC(:, tsize+j), funargs{:});
            TDEPTH(tsize+j) = TDEPTH(node) + 1;
            if ~usingsoo, TDIAM(tsize+j) = diam(TINT(:, :, tsize+j)); end;
            TLEAF(tsize+j) = true;
        end;
        TLEAF(node) = false;     % parent is no longer leaf
        tsize = tsize + 2^q;
        tmaxdepth = max(TDEPTH(node)+1, tmaxdepth);
    end

%#### Main loop
stopflag = false; 
if usingsoo, tau = 1; d = 0; rmax = -Inf; nodefoundattau = false; end;
while ~stopflag,
    
    %## Select set/node to expand
    if usingsoo, % SOO algorithm
        node = 0;
        % continue searching across depths until a node is found
        while ~node,
            % IF DEBUGGING: declare variables INT = []; C = []; DEPTH = []; LEAF = []; R = [];
            leavesd = find(TLEAF(1:tsize) & (TDEPTH(1:tsize) == d));
            if ~isempty(leavesd),           % it might be that all nodes at the depth were expanded
                [rmaxd, irmaxd] = max(TR(leavesd)); % best leaf at depth d (breaking ties for the first found)
                if rmaxd >= rmax,                   % if it's not dominated by those above it
                    node = leavesd(irmaxd);             % we'll expand it
                    rmax = rmaxd;                       % its value becomes the new dominating value
                    nodefoundattau = true;
                end;
            end;
            if ~node,                       % node not found, continue with next depth
                d = d + 1;
                if (d > tmaxdepth) || (d > cfg.dmax(tau)), % unless we have exceeded the allowed range
                    % move to the next iteration, and restart at the root
                    tau = tau + 1; d = 0; rmax = -Inf; nodefoundattau = false;
                    if cfg.verb >= 5, fprintf('SOO old: moving to next iter, depth(T)=%d, dmax(tau)=%.1f\n', tmaxdepth, cfg.dmax(tau)); end;
                end;
            end;
        end;
        if cfg.verb >= 5, fprintf('SOO old: tau=%d, expanding node %d at d=%d, r=%.2f\n', tau, node, d, rmax); end;
        % note the children of the expanded node now become candidate leaves 
    else    % DOO algorithm
        % all leaves are candidates
        leaves = find(TLEAF(1:tsize));
        % compute b-values, choose a set with the maximal b-value, breaking ties in favor of first
        [bmax, ibmax] = max(TR(leaves) + TDIAM(leaves)); node = leaves(ibmax);
    end;
    
    %## Expand the chosen node
    expand(node);
    n = n + 1;
    
    % view current status if configured
    if cfg.verb >= 4, 
        viewsets(TINT, TC, TR, TLEAF, node, tsize-2^q+1:tsize); 
        pause(0.1); 
    end;
    
    % stop when the budget has been exhausted: 
    stopflag = n >= cfg.n;    
    
end;    % end main tree construction loop

%#### Find best action sequence on the tree, breaking ties in favor of the first found
[rstar, istar] = max(TR(1:tsize));
ustar = TC(:, istar);

% create stats, data if requested on output
if nargout >= 2,
    optinfo.n = n;
    optinfo.treesize = tsize;
    optinfo.trun = cputime - t;
    optinfo.message = sprintf('Optimization done, n=%d expansions, treesize (# fun calls)=%d, trun=%.2f', ...
        optinfo.n, optinfo.treesize, optinfo.trun); 
end;
if nargout >= 3,
    optdata.TINT = TINT(:, :, 1:tsize);
    optdata.TC = TC(:, 1:tsize);
    optdata.TR = TR(:, 1:tsize);
    optdata.TDEPTH = TDEPTH(1:tsize);
    if ~usingsoo, optdata.TDIAM = TDIAM(1:tsize); end;
    optdata.TLEAF = TLEAF(1:tsize);
end;
% construct the output based on calling mode
switch nargout,
    case 1, varargout = {ustar};
    case 2, varargout = {ustar, optinfo};
    case 3, varargout = {ustar, optinfo, optdata};
end;

 
end % main function

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Visualization
% sequences in the U space
function viewsets(TINT, TC, TR, TLEAF, expanded, inew)
    if size(TINT, 2) > 2, return; end;  % only plots <= 2D
    figure(10); clf;

    if size(TINT, 2) == 1,
        for i = [expanded find(TLEAF)],
            if i == expanded, c = 'r'; lw = 2;
            elseif any(i == inew), c = 'b'; lw = 1.5;
            else c = 'k'; lw = 1;
            end;
            plot(TINT(:, 1, i), [0 0], c, 'LineWidth', lw, 'Marker', '+', 'MarkerFaceColor', c);
            if i == expanded, hold on; end;
            plot([TC(1, i) TC(1, i)], [0 TR(i)], c);
        end;
        set(gca, 'ylim', [-1 max(TR(TLEAF))+1]);
    else        
        for i = [expanded find(TLEAF)],
            if i == expanded, c = 'r'; lw = 2;
            elseif any(i == inew), c = 'b'; lw = 1.5;
            else c = 'k'; lw = 1;
            end;
            patch(TINT([1 2 2 1], 1, i), TINT([1 1 2 2], 2, i), [0 0 0 0], 'EdgeColor', c, 'FaceColor', c, 'FaceAlpha', 0.5, 'LineWidth', lw);
            plot3([TC(1, i) TC(1, i)], [TC(2, i) TC(2, i)], [0 TR(i)], c);
            if i == expanded, hold on; end;
        end;
        set(gca, 'zlim', [0 max(TR(TLEAF))*1.1]);
    end;
end

%             T_DIMSEL = T_DIMSEL + [1 cputime-t2]; % 1st elem count, 2nd elem total