function magenmpc(cfg)
% Generic planning/predictive control for multiagent systems
%   MAGENMPC(CFG)
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see he str2cfg
% 
% Can use a leader, which should always be the first agent.
% A different problem and planning algorithm are used for the leader. 
% The requirement is that the dynamics of this leader are the same as those of the first
% agent in the multiagent problem.

% Note the leader 'seq' mode is currently untested

if nargin < 1, cfg = struct(); end;

% -----------------------------------------------
% Process configuration structure

% default config
% algo config
CFG.run = 0;                        % run control
CFG.plottrial = 0;                  % show some trajectories
CFG.problem = '';                   % what problem to solve
CFG.model_params = {};              % extra parameters for problem calling in 'model' mode
CFG.online = 1;                     % 1: control online; currently only online control is meaningful
CFG.X = [];                         % may contain set of states for reset (or its ID)
CFG.gamma = 0.98;                   % discount factor
CFG.planalg = 'maopd';              % planning algorithm
CFG.planparams = struct;            % other planning algorithm parameters
CFG.comm = 'prev';                  % communication config:
    % 'none'; 'prev'ious plans; 'seq'uential comm of plans from (1..i-1) to i (i+1..n previous preds)
CFG.randseed = [];                  % set the random seed (for repeatable results)
CFG.trials = 1;                     % # trials (control episodes)
CFG.maxsteps = 100;                 % max # steps in a trial
CFG.reset = 'rand';                 % trial reset
% leader config
CFG.leadproblem = [];               % if using a leader, set its problem type 
CFG.leadmodel_params = {};          % leader model params
CFG.leadplanalg = [];               % leader planning algorithm
CFG.leadplanparams = struct;        % leader planning params
CFG.leadreset = 'rand';
% save, stats config
CFG.datadir = '';                   % data dir
CFG.datafile = '';                  % save data to file
CFG.saveplandata = 0;               % remember the data created by planning algorithm -- memory-intensive!
% display config
CFG.verb = 0;                       % verbosity: the higher, the more detailed the messages displayed
CFG.visualize = 0;                  % for online, visualization level (0 = none, 1 = trial-level, 2 = step-level)
CFG.viscfg = struct;                % visualization config options
CFG.trialdisp = 10;
CFG.sampledisp = .05;
% figure config
CFG = setfigprop(CFG, 'addfields'); % figure config fields

% List of fields that define the problem and therefore may NOT be overwritten on load
KEEPFIELDS = {'problem', 'gamma', 'online', 'planalg', 'planparams', 'X'};

% Install function defaults for everything else
cfg = parseconfig(cfg, CFG);

if ~cfg.run,        % load data file
    loadvars = {'cfg', 'model', 'planinfoh'};
    if cfg.plottrial,
        loadvars = {loadvars{:}, 'X', 'U', 'R', 'GRAPHS', 'p', 'q'};
    end;
    % make sure that cfg is not overwritten
    cfg1 = cfg; kf = KEEPFIELDS;
    load(cfg.datafile, loadvars{:});
    % Overwrite problem-defining fields from loaded config, keep the rest as in the (current) cfg1;
    % the result becomes the new config
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['Data loaded from [' cfg.datafile '].'], cfg.verb, 1);
end;

if cfg.run && ~cfg.online, error('MAGENMPC does not implement offline mode'); end;

% Echo config
dispx('MAGENMPC called with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1, 'cfg', 'config');
planparams = cfg.planparams; dispx(planparams, cfg.verb, 1, 'planparams', 'config'); clear planparams;

% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

% -----------------------------------------------
% Main algorithm
if cfg.run,
    
    % create model, get its dimensionality data, other shorthand variables
    model = feval(cfg.problem, 'model', cfg.model_params{:});
    nagents = model.nagents;
    pranges = model.pranges; qranges = model.qranges;
    p = sum(model.p); q = sum(model.q);
    
    % init constant fields planning config, copying also the settings from config.planparams
    plancfg = struct;
    plancfg.gamma = cfg.gamma;
    plancfg = copyfields(cfg.planparams, plancfg);
    
    % init random generator if configured
    if ~isempty(cfg.randseed), rand('twister', cfg.randseed); end;
    
    % if X is string identifier, get actual states from the problem
    if ischar(cfg.X), cfg.X = feval(cfg.problem, 'X0', cfg.X); end;
    % process state set
    X = cfg.X; if iscell(X), X = flat(X); end;
    % if "smart" reset is used, copy states to reset structure 
    % -- unless overridden by existing setting
    if isstruct(cfg.reset) && ~isfield(cfg.reset, 'x0'), cfg.reset.x0 = X; end;
    
    % only ONLINE control, along trajectories of the system, is supported

    % init histories etc.
    X = NaN(p, cfg.maxsteps+1, cfg.trials);
    U = NaN(q, cfg.maxsteps+1, cfg.trials);
    R = NaN(nagents, cfg.maxsteps+1, cfg.trials);
    GRAPHS = NaN(nagents, nagents, cfg.maxsteps+1, cfg.trials); % for uniformity, keep this even for var-conn problems
    % planinfoh = struct; -- cannot do this, we don't know the fields in advance
    if cfg.saveplandata, 
        plandatah = cell(nagents, cfg.maxsteps, cfg.trials); 
        % Upredictionsh = NaN(q, 0, cfg.maxsteps+1, cfg.trials);     % can remain empty
    end;

    % leader initialization
    uselead = ~isempty(cfg.leadproblem);
    if uselead,
        leadmodel = feval(cfg.leadproblem, 'model', cfg.leadmodel_params{:});
        leadplancfg = struct;
        leadplancfg.gamma = cfg.gamma;
        leadplancfg.actiontype = 'seq';
        leadplancfg = copyfields(cfg.leadplanparams, leadplancfg);
        if isstruct(cfg.leadreset) && ~isfield(cfg.leadreset, 'x0'), 
            cfg.leadreset.x0 = cfg.reset.x0(pranges(1)+1:pranges(2), :); 
        end;
        % note there are no separate state and action trajectories for the leader: they are stored
        % along with the other agents'
        % rewards in the single-agent leader problem may nevertheless differ from multiagent ones
        LEADR = NaN(1, cfg.maxsteps+1, cfg.trials);
        % allow use of reference model for leader
        if leadmodel.userefmodel, LEADREFX = NaN(leadmodel.refp, cfg.maxsteps+1, cfg.trials); end;
    end;
    
    trial = 1;
    timestat.run = 0;
    timestat.planning = nan(cfg.maxsteps, cfg.trials);

    % init visualization config if needed
    if cfg.visualize,
        vcfg = cfg.viscfg;
        vcfg.gview = [];
    end;

    % -----------------------------------------------
    % Run control
    dispx('Performing MAGENMPC online control...', cfg.verb, 0);

    t = cputime;
    while trial <= cfg.trials,       % main loop

        % initialize trial
        k = 1; terminal = 0;
        % initial state
        % if "smart" reset is used, copy trial index as it may be needed
        if isstruct(cfg.reset), cfg.reset.trial = trial; end;
        [X(:, 1, trial), GRAPHS(:, :, 1, trial)] = feval(cfg.problem, 'reset', model, cfg.reset);
        
        % for variable connectivity: compute minimal robust graph in case method uses it
        %   - but only for linear agents, else it is meaningless
        if strcmp(model.matype, 'varconn') && strcmp(model.id, 'malinear'),
            plancfg.Grob = feval(cfg.problem, 'minrobgraph', model, X(:, 1, trial));
        end;
        
        % separately reset leader, if using one, checking that its state matches
        if uselead,
            if isstruct(cfg.leadreset), cfg.leadreset.trial = trial; end;
            if leadmodel.userefmodel, 
                [xlead, LEADREFX(:, 1, trial)] = feval(cfg.leadproblem, 'reset', leadmodel, cfg.leadreset);
            else
                xlead = feval(cfg.leadproblem, 'reset', leadmodel, cfg.leadreset);
            end;
            if any(xlead ~= X(pranges(1)+1:pranges(2), 1, trial)), 
                error('Reset state mismatch between leader model and multiagent model');
            end;
        end;

        timestat.run = timestat.run + (cputime - t);
        % visualize initial state
        if cfg.visualize >= 2,
            vcfg.mampcstep = 1;
            vcfg.mampctrial = 0;
            vcfg.k = k - 1;
            vcfg.trial = trial;
            vcfg.ktotal = k - 1 + cfg.maxsteps * (trial - 1); % this only works in continuing tasks
            [figh vcfg.gview] = feval(model.visualizefun, vcfg);
        end;
        t = cputime;

        while k <= cfg.maxsteps && ~terminal,        

            % perform planning for every agent
            plancfg.k = k;
            plancfg.G = GRAPHS(:, :, k, trial);   
            % send agent predictions if available; step 1 has already been applied so it's removed
            % note state predictions start from the NEXT state for step 2
            if k > 1 && size(Xagents, 2) >= 3,  Xpredictions = Xagents(:, 3:end);
            else                                Xpredictions = zeros(p, 0);
            end;
            if k > 1 && size(Uagents, 2) >= 2,  Upredictions = Uagents(:, 2:end); 
            else                                Upredictions = zeros(q, 0); 
            end;
            Xagents = zeros(p, 0); Uagents = zeros(q, 0); 
            
            for i = 1:nagents,
                
                % if the first agent is the leader, use single-agent algorithm to control it
                if i == 1 && uselead,
                    leadplancfg.k = k;
                    if leadmodel.userefmodel, leadplancfg.refx = LEADREFX(:, k, trial); end;
                    t2 = cputime;
                    % note we retrieve plandata anyway, as the 4-outputs calling mode is how 
                    % we signal the single-agent planning algorithm that it should return state and
                    % action sequences
                    [ui, xi, planinfo, plandata] = feval(cfg.leadplanalg, ...
                        X(pranges(1)+1:pranges(2), k, trial), leadmodel, leadplancfg);
                else
                    % configuration
                    plancfg.iagent = i;
                    % handle communication
                    if cfg.comm(1) ~= 'n', 
                        plancfg.Xpredictions = Xpredictions; plancfg.Upredictions = Upredictions; 
                    end;
                    % choose actions using the configured algorithm
                    t2 = cputime;
                    if k == 2 && i == 3,
                        plancfg.brstop = 1;
                    else
                        plancfg.brstop = 0;
                    end;
                    if cfg.saveplandata,
                        [ui, xi, planinfo, plandata] = feval(cfg.planalg, X(:, k, trial), model, plancfg);
                    else
                        [ui, xi, planinfo] = feval(cfg.planalg, X(:, k, trial), model, plancfg);
                    end;
                end;
                % already stop counting planning time
                timestat.planning(k, trial) = timestat.planning(k, trial) + cputime - t2;                
                % save sequences, making sure every uninitialized element is NaN
                Xagents(:, size(Xagents, 2)+1:size(xi, 2)) = NaN; % no effect if... 
                Uagents(:, size(Uagents, 2)+1:size(ui, 2)) = NaN; % ...X/Uagents longer than xi/ui
                Xagents(pranges(i)+1:pranges(i+1), 1:size(xi, 2)) = xi;
                Uagents(qranges(i)+1:qranges(i+1), 1:size(ui, 2)) = ui;
                % if communication is sequential, agent i+1 receives the plan of agents 1, ..., i
                if cfg.comm(1) == 's',
                    % pad with NaNs
                    Xpredictions(:, size(Xpredictions, 2)+1:size(xi, 2)-1) = NaN;
                    Upredictions(:, size(Upredictions, 2)+1:size(ui, 2)) = NaN;
                    % note again we just care about the NEXT states of the transitions, so we ignore
                    % the first element of Xpredictions
                    Xpredictions(pranges(i)+1:pranges(i+1), 1:size(xi, 2)-1) = xi(:, 2:end);
                    Upredictions(qranges(i)+1:qranges(i+1), 1:size(ui, 2)) = ui;
                    % overwrite sequences' tails with null values 
                    % to avoid keeping any values previously there (from previous-step predictions)
                    Xpredictions(pranges(i)+1:pranges(i+1), size(xi, 2):end) = NaN;
                    Upredictions(qranges(i)+1:qranges(i+1), size(ui, 2)+1:end) = NaN;
                end;
                planinfoh(i, k, trial) = planinfo;
                if cfg.saveplandata, plandatah{i, k, trial} = plandata; end;
            end;    % agents loop
            
            % save predictions if configured
            % -- requires some work to properly handle sequential communication
            % if cfg.saveplandata && size(Uagents, 2) > 1, 
            %    Upredictionsh(:, 1:size(Uagents, 2), k, trial) = Uagents; 
            % end;
                        
            if any(isnan(Uagents(:, 1))),   % handle action error
                dispx(sprintf('Agent control failed, stopping. k=%d: x=%s u=%s: ', k, ...
                    num2strx(X(:, k, trial), 'vector'), num2strx(Uagents(:, 1), 'vector')), cfg.verb, 3);
                break;
            end;

            % apply the action
            U(:, k, trial) = Uagents(:, 1);
            [X(:, k+1, trial), R(:, k+1, trial), terminal, GRAPHS(:, :, k+1, trial)] = ...
                feval(model.fun, model, X(:, k, trial), U(:, k, trial), 'real');

            % if using leader, check consistency with its model
            if uselead,
                % apply the action
                if leadmodel.userefmodel,
                    [xlead, LEADR(1, k+1, trial), leadterminal, dummypplus, LEADREFX(:, k+1, trial)] = ...
                        feval(leadmodel.fun, leadmodel, X(pranges(1)+1:pranges(2), k, trial), ...
                        U(qranges(1)+1:qranges(2), k, trial), [], LEADREFX(:, k, trial));
                else
                    [xlead, LEADR(1, k+1, trial), leadterminal] = feval(leadmodel.fun, leadmodel, ...
                        X(pranges(1)+1:pranges(2), k, trial), U(qranges(1)+1:qranges(2), k, trial));
                end;
                % consistency checks
                if any(xlead ~= X(pranges(1)+1:pranges(2), k+1, trial)),
                    error('State transition mismatch between leader model and multiagent model');
                elseif leadterminal && ~terminal, 
                    error('Leader terminated before multiagent system');
                end;
            end;
            
            % for var-conn problems: check for and signal loss of connectivity
            if model.matype(1) == 'v',
                if k == 1, lastconn = 1; end;
                conn = sum(abs(eig(diag(sum(plancfg.G)) - plancfg.G)) <= 1e-10) <= 1;
                if ~conn && lastconn,       dispx(sprintf('Lost connectivity at k=%d!', k), cfg.verb, 3); 
                elseif conn && ~lastconn,   dispx(sprintf('Regained connectivity at k=%d.', k), cfg.verb, 3); 
                end;
                lastconn = conn;
            end;

%             if isfield(plancfg, 'Upredictions'), 
%                 disp(['prediction match: ' num2str(squeeze(U(:, k, trial))' == squeeze(plancfg.Upredictions(:, 1))')]);
%             end;
            
            % update time stats
            timestat.run = timestat.run + (cputime - t);

            % log
            if cfg.verb >= 3,
                if ~any(diff(model.p)) && ~any(diff(model.q)), 
                    % display per-agent x and u
                    disp(sprintf('k=%d:', k));
                    xs = num2str(reshape(X(:, k, trial), model.p(1), nagents)'); 
                    us = num2str(reshape(U(:, k, trial), model.q(1), nagents)'); 
                    inds = '0'+(1:nagents)';
                    s = [repmat('  x', nagents, 1) inds repmat(' = [', nagents, 1) xs ...
                        repmat('],     u', nagents, 1) inds repmat(' = [', nagents, 1) us repmat(']', nagents, 1)];
                    disp(s);
                else
                    % display compact, one-line state and action
                    disp(sprintf('k=%d: x=%s u=%s', k, ...
                        num2strx(X(:, k, trial), 'vector'), num2strx(U(:, k, trial), 'vector')));
                end;
                drawnow;
            end;

            % visualization
            if cfg.visualize >= 2,
                vcfg.mampcstep = 1;
                vcfg.mampctrial = 0;
                vcfg.k = k;
                vcfg.trial = trial;
                vcfg.ktotal = k - 1 + cfg.maxsteps * (trial - 1); % this only works in continuing tasks
                [figh vcfg.gview] = feval(model.visualizefun, vcfg);
            end;

            % finalize loop
            t = cputime; 
            k = k + 1;
        end;

        % update stats
        timestat.run = timestat.run + (cputime - t);

        % visualization
        if cfg.visualize >= 1,
            vcfg.mampctrial = 1;
            vcfg.mampcstep = 0;
            vcfg.trial = trial;
            vcfg.ktotal = k - 1 + cfg.maxsteps * (trial - 1); % this only works in continuing tasks
            [figh vcfg.gview] = feval(model.visualizefun, vcfg);
        end;

        % console feedback of algorithm progress
        if ~mod(trial, cfg.trialdisp), 
            dispx(sprintf('Trial #%d completed.', trial), cfg.verb, 2);
        end;

        % start counting time again, increment iteration counter
        t = cputime;
        trial = trial + 1;
    end;        % while not converged and allowed more iterations

    dispx(sprintf('trials=%d exhausted. Algorithm stopped.', cfg.trials), cfg.verb, 0);

    % finalize visualizer
    if cfg.visualize,
        vcfg.mampctrial = 0;
        vcfg.mampcstep = 0;
        vcfg.finalize = 1;
        [figh vcfg.gview] = feval(model.visualizefun, vcfg);
    end;
    
end;

% Plot a trial (only makes sense if algorithm ran online)
% Note the default plot mode will likely be a jungle, it's best to implement a plot function for the
% multiagent problem
if cfg.plottrial && cfg.online,
    Ns = find(~isnan(U(1, :, cfg.plottrial)), 1, 'last');
    hist.Ns = Ns;
    hist.t = (0:Ns) * model.Ts; 
    hist.x = X(:, 1:Ns+1, cfg.plottrial); 
    hist.u = U(:, 1:Ns+1, cfg.plottrial); 
    hist.r = R(:, 1:Ns+1, cfg.plottrial);
    hist.G = GRAPHS(:, :, 1:Ns+1, cfg.plottrial);
    % might want to add the leader rewards and perhaps reference state here
    % hist.R = discreturn(cfg, hist.r, Ns, terminal);
    mpfigh = [];
    % retroactively add plot function
    if exist([model.id '_plot'], 'file'), model.plotfun = [model.id '_plot']; end;
    % if the model provides a plot function, use it
    if isfield(model, 'plotfun'), 
        mpcfg = struct;
        mpcfg.traj = hist;
        mpcfg.datasource = 'caller';
        mpcfg = copyfields(cfg, mpcfg, fieldnames(setfigprop(struct, 'addfields')));
        mpfigh = feval(model.plotfun, mpcfg);
    end;
    % otherwise, or if the plot function produced nothing, make a generic plot
    if ~isempty(mpfigh), figh = mpfigh;
    else figh = plothistory(hist); 
    end;
    setfigprop(cfg);
    % save if requested
    saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);
end;

% -----------------------------------------------
% Backup data

if cfg.run && ~isempty(cfg.datafile),
    % save into the same directory as the problem
    if isempty(cfg.datadir),
        datadir = fileparts(which(cfg.problem));
    else
        datadir = cfg.datadir;
        if (datadir(end) == '\'), datadir = datadir(1:end-1); end;
    end;
    cfg.datafile = [datadir '\' cfg.datafile];
    save(cfg.datafile);
    dispx(['Data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;


% END qi() RETURNING varargout =================================================================
