function  [useq, xseq, planinfo, plandata] = maopd_predictneighbors(xroot, model, cfg)
% OPD Optimistic planning with deterministic policies for multiagent systems
% Uses predictions of neighbor trajectories (either communicated directly or via action plans)

% Uses arrays; does not support tree or memory reuse
% Author: Lucian Busoniu
% Versions:

CFG.gamma = [];                 % discount factor
CFG.n = [];                     % expansion budget
CFG.rob = 0;                    % 1: consider only neighbors on robust graph; 0: all neighbors
CFG.criterion = 'optim';        % expansion criterion: optim(istic) or unif(orm)
CFG.iagent = [];                % index of this agent
CFG.maxpredlength = Inf;        % if the agent predictions should be cut at this length 
    % (to simulate communication limitations)
CFG.G = [];                     % graph of interagent connections: adjacency matrix
CFG.Grob = [];                  % minimal robust graph (required if rob=1)
CFG.U = [];                     % discrete, flat action spaces of ALL agents, in a cell array
CFG.Upredictions = [];          % agent action predictions (vector of flat actions)
CFG.Xpredictions = [];          % state predictions, similar structure
    % any unavailable predictions should be set to NaN; U will be assumed zero in these cases
CFG.xpredtype = 'sim';          % neighbors' state predictions:
    % sim: always simulate using the action predictions
    % comm: use communicated (in Xpredictions) as long as available, simulate afterwards
    % TODO IMPLEMENT flatcomm: use communicated as long as available, constant afterwards
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.gamma; n = cfg.n; optimistic = strcmp(cfg.criterion, 'optim');
iagent = cfg.iagent;

% retrieve and check graph connections of current agent, stopping if main connections interrupted
% NOTE these may not be entirely suitable for fixed-conn problems
errormessage = '';
if cfg.rob,
    g = cfg.Grob(iagent, :);
    if ~all(cfg.G(iagent, g)),
        errormessage = sprintf('WARNING! Robust graph connection of [%d] with [%d] broken, cannot run planning', ...
            iagent, find(g & ~cfg.G(iagent, :)));
    end;
else
    g = cfg.G(iagent, :);
    if ~any(g),  
        errormessage = sprintf('WARNING! All connections of [%d] broken, cannot run planning', iagent); 
    end;
end;
if ~isempty(errormessage)
    dispx(errormessage, cfg.verb, 0);
    useq = NaN(model.q(iagent), 1);
    planinfo.n = []; planinfo.depth = []; planinfo.trun = []; planinfo.treesize = [];
    planinfo.message = errormessage; 
    plandata = [];
    return;
end;

ineighbors = find(g);
% form state value from this agents and neighbors, and neighbor action and state indices
x = xroot(model.pranges(iagent)+1:model.pranges(iagent+1)); uindices = []; xindices = [];
for ineighbor = ineighbors, 
    x = [x; xroot(model.pranges(ineighbor)+1:model.pranges(ineighbor+1))]; 
    uindices = [uindices model.qranges(ineighbor)+1:model.qranges(ineighbor+1)];
    xindices = [xindices model.pranges(ineighbor)+1:model.pranges(ineighbor+1)];
end;
uindices = uindices(:); xindices = xindices(:);
x0neighb = xroot(xindices);     % save this in case it's needed later
xroot = x;                      % agent + neighbors
dimx = sum(model.p([iagent ineighbors])); % number of state vars (computed differently as verification)
% form action spaces
if iscell(cfg.U{iagent}), U = flat(cfg.U{iagent}); else U = cfg.U{iagent}; end;
M = size(U, 2);           % number of discrete actions for agent

if size(cfg.Xpredictions, 2) ~= size(cfg.Upredictions, 2), error('Different X and U pred lengths'); end;     % shouldn't happen
predlen = size(cfg.Xpredictions, 2);
if cfg.maxpredlength <= predlen, predlen = cfg.maxpredlength; end; % explicit limitation on prediction length
if cfg.xpredtype(1) == 's', xpred = []; % should resimulate everything (to emulate lack of state comm)
elseif cfg.xpredtype(1) == 'f', 
    % up to predlen, we take the comunicated predictions
    xpred = cfg.Xpredictions(xindices, 1:predlen);
    xpred(:, predlen+1:cfg.n) = nan(size(xpred, 1), cfg.n-predlen);
    % the rest take the last known values, 
    % padding with enough data that even if depth=budget, we have a prediction
    % note that if some xpred were nan BEFORE predlen, this code still takes care of filling them in
    for ix = 1:size(xpred, 1),
        ilastknown = find(~isnan(xpred(ix, :)), 1, 'last');
        if isempty(ilastknown), xlastknown = x0neighb(ix); ilastknown = 0;
        else                    xlastknown = xpred(ix, ilastknown); 
        end;
        xpred(ix, ilastknown+1:cfg.n) = xlastknown;
    end;
else
    % up to predlen, we take the comunicated predictions
    % the rest will be automatically simulated within the dynamics function ***_mg
    xpred = cfg.Xpredictions(xindices, 1:predlen);
end;
upred = cfg.Upredictions(uindices, 1:predlen);

% below here, predlen is not used anymore so we are sure to adapt to the case where we initialize
% more data than predlen
clear predlen

% start counting time
t = cputime;

% prepare tree data 
maxtreesize = n*M + 1;
Tparent = zeros(2, maxtreesize, 'uint32');        % parent and parent action
Tchildren = zeros(M, maxtreesize, 'uint32');      % child index for each action
Tdepth = zeros(1, maxtreesize, 'uint16');         % depth (uint16 as depth > 65536 considered unfeasible)
Tleaf = false(1, maxtreesize);                    % leaf flag
Tx = NaN(dimx, maxtreesize, 'double');            % state
Tr = NaN(1, maxtreesize, 'double');               % reward
TR = NaN(1, maxtreesize, 'double');               % partial return up to node (cumulative reward)
Tterm = false(1, maxtreesize);                    % terminal flag
Tb = NaN(1, maxtreesize, 'double');               % b-values

% initialize the root node (note 0 = null pointer)
root = 1;
Tparent(:, root) = 0;
Tchildren(:, root) = 0;
Tdepth(root) = 0;
Tleaf(root) = true;
Tx(:, root) = xroot;
Tr(root) = 0;
TR(root) = 0;
Tterm(root) = false;    % presumably, we are not asked to make a decision in a terminal state
Tb(root) = 1/(1-discount);

treesize = 1;

% 2) Main tree construction loop
for tau = 1:n,
    
    % 2a) select candidate leaves for expansion
    leaves = find(Tleaf);
    if optimistic, 
        % optimistic: max b-value RELATIVE TO THE ROOT
        B = TR(leaves) + discount.^double(Tdepth(leaves)) .* Tb(leaves);
        leaves = leaves(B == max(B));
        % Notes: 
        % - due to OPD structure we are able to directly go for the max-b
        % leaf, equivalent to iteratively developing all optimistic paths
        % - without memory use, Tb(leaves) will always be 1/1-gamma
    else
        % uniform: smallest depth
        leaves = leaves(Tdepth(leaves) == min(Tdepth(leaves)));
    end;
    % select only nonterminal leaves; if none, nothing left to do
    leaves = leaves(~Tterm(leaves)); if isempty(leaves); break; end;
    % break any ties between remaining candidates deterministically
    nodetoexpand = leaves(1);
        
    % 2b) expand the chosen node
    % retrieve predicted states based on config
    if Tdepth(nodetoexpand)+1 > size(xpred, 2),     % prediction unavailable
        xplusneighbors = [];
        xnavail = false; xnset = false;
    else
        % use available states, resimulate unavailable ones
        % (NaNs from Xpredictions are taken over for unavailable states)
        % depth 0-based, prediction indexing 1-based
        xplusneighbors = xpred(:, Tdepth(nodetoexpand)+1);
        xnavail = ~any(isnan(xplusneighbors)); xnset = false;
        if all(isnan(xplusneighbors)), xplusneighbors = []; end; % may help the simulation
    end;
    % neighb simulation may require the actions (0s assumed beyond set pred length)
    if Tdepth(nodetoexpand)+1 > size(upred, 2),
        uneighbors = zeros(size(uindices));
    else
        uneighbors = upred(:, Tdepth(nodetoexpand)+1);
        uneighbors(isnan(uneighbors)) = 0;
    end;
    
    % node is no longer a leaf
    Tleaf(nodetoexpand) = false;
    % run through the discrete actions 
    for j = 1:M,
%         if iagent == 3 && tau == 1 && cfg.k == 7  && any(j == [1 2]), end;
    
        % find index where new node will be, save it on the children table of current node
        child = treesize + 1;
        Tchildren(j, nodetoexpand) = child;

        if cfg.brstop && tau == 1,
            model.brstop = 1;
        else
            model.brstop = 0;
        end;
        
        % simulate transition and add data to tree
        [Tx(:, child), Tr(child), Tterm(child)] = feval(model.fun, model, Tx(:, nodetoexpand), ...
            [U(:, j); uneighbors], 'pred', iagent, g, xplusneighbors);
        % if Tr(child) < 0 || Tr(child) > 1, error('Rewards must be in [0, 1]'); end;
        % if any neighbor states were unavailable, set them and flag the change
        if ~xnavail && ~xnset, xplusneighbors = Tx(model.p(1)+1:end, child); xnset = true; end;

        % create new node
        Tparent(:, child) = [nodetoexpand; j];      % parent is expanded node, along jth action
        Tchildren(:, child) = 0;                    % init children pointers to "null"
        Tdepth(child) = Tdepth(nodetoexpand) + 1;   % one deeper than the parent
        Tleaf(child) = true;                        % new child is a leaf
        TR(child) = TR(nodetoexpand) + discount^double(Tdepth(nodetoexpand)) * Tr(child);
        if Tterm(child),    Tb(child) = 0;          % values of terminal nodes exactly 0
        else                Tb(child) = 1/(1-discount);
        end;
            
        % we just added a new node, increase tree size
        treesize = treesize + 1;
    end;
    % if any newly simulated neighbors states, save them on the predictions table
    if xnset, xpred(:, Tdepth(nodetoexpand)+1) = xplusneighbors; end;

    % 2c) update b-values affected by expanded node (path to root)
    node = nodetoexpand;
    while node ~= 0,       % when node is 0, it means the root has already been processed
        children = Tchildren(:, node);
        Tb(node) = max(Tr(children) + discount .* Tb(children));
        node = Tparent(1, node);
    end;

    if cfg.verb >= 4,
        visualizetree(Tparent, Tchildren, Tdepth, Tleaf, Tx, Tr, Tterm, Tb, TR, U, nodetoexpand); 
        set(gcf, 'Name', sprintf('k=%d: Tree visualization', cfg.k));
        pause;
    end;

end;    % end main tree construction loop


% 4) Find near-optimal action sequence and corresponding state sequence
% choose a leaf with maximal u-value (i.e., truncated return)
leaves = find(Tleaf); [Rmax, jmax] = max(TR(leaves)); node = leaves(jmax);
% navigate greedy path upwards in the tree until level 1, just after root, saving the actions
% this procedure avoids the need to store and backup util-values while building the tree
useq = NaN(model.q(iagent), Tdepth(node));
xseq = NaN(model.p(iagent), Tdepth(node)+1);
maxpath = NaN(1, Tdepth(node)+1); maxpath(end) = node;
while node ~= root,
    % save the action used to arrive at this node by its parent
    useq(:, Tdepth(node)) = U(:, Tparent(2, node));
    % save the state
    xseq(:, Tdepth(node)+1) = Tx(1:model.p(iagent), node);
    maxpath(Tdepth(node)+1) = node;
    node = Tparent(1, node); 
end;
% first (root) state not yet processed, copy it to the sequence
xseq(:, 1) = Tx(1:model.p(iagent), node); % node = root
maxpath(1) = node;

% output stats if requested
if nargout >= 2,
    planinfo.n = tau;
    planinfo.depth = max(Tdepth);
    planinfo.trun = cputime - t;
    planinfo.treesize = treesize;
    planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d, treesize=%d', ...
        tau, planinfo.depth, planinfo.treesize);
end;
% output only tree data (no memory)
if nargout >= 3,
    plandata.Tparent = Tparent(:, 1:treesize);
    plandata.Tchildren = Tchildren(:, 1:treesize);
    plandata.Tdepth = Tdepth(1:treesize);
    plandata.Tleaf = Tleaf(1:treesize);
    plandata.Tx = Tx(:, 1:treesize);
    plandata.Tr = Tr(1:treesize);
    plandata.TR = TR(1:treesize);
    plandata.Tb = Tb(1:treesize);
    plandata.Tterm = Tterm(1:treesize);
end;

end % main function


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local functions

% Plot a graphical representation of the current tree
% (argument list is a mess but whatever)
function visualizetree(Tparent, Tchildren, Tdepth, Tleaf, Tx, Tr, Tterm, Tb, TR, ...
    U, expandednode)
persistent figh;
figsize = [800, 600];
if isempty(figh), figh = figurex(struct('size', figsize)); else figure(figh); end;
clf;

% initialize helper variables for nested functions
M = size(U, 2);
nodestoexpand = {};
% build and customize tree
if verLessThan('Matlab', '7.10'),   guitree = uitree('Root', rendernode(1));
else                                guitree = uitree('v0', 'Root', rendernode(1));
end;
guitree.Position(3) = figsize(1);   % expand to figure width
% expand nodes
for i = 1:length(nodestoexpand), guitree.expand(nodestoexpand{i}); end;
% done

    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Nested function to render a node recursively
    function guinode = rendernode(i, jp)
    % create label
    if i == 1,  label = 'root: ';
    else        label = ['u=' num2strx(U(:, jp)', 'v') ' --> '];
    end;
    label = [label sprintf('x=%s, r=%.4f, R=%.4f, b=%.4f', num2strx(Tx(:, i)', 'v'), Tr(i), TR(i), Tb(i))];
    if Tterm(i), label = [label ' T']; end;
    if i == expandednode,label = ['<html><font color="red">' label '</font></html>'];    % expanded
    end;
    % add tree(i) to tree
    if verLessThan('Matlab', '7.10'),   guinode = uitreenode(label, label, [], Tleaf(i));
    else                                guinode = uitreenode('v0', label, label, [], Tleaf(i));
    end;
    if ~Tleaf(i),
        % remember we need to expand this node
        nodestoexpand{end+1} = guinode;
        for jj = 1:M, guinode.add(rendernode(Tchildren(jj, i), jj)); end;
    end;
    end     % function rendernode
end     % function visualizetree


% Action prediction figure (goes at end of main function body): This code may be out of date
% persistent predfigh; 
% if cfg.verb >= 3; % && any(cfg.k == 3:6), 
%     if isempty(predfigh), predfigh = figurex(struct('size', [900 750])); else figure(predfigh); end;
%     maxpath = [1 maxpath];  % add the root
%     agentcolors = {'b', 'r', 'g', 'k', 'm', 'y', 'c'};
%     iall = [iagent ineighbors];
%     subplot(2, 2, 1); cla; subplot(2, 2, 2); cla; subplot(2, 2, 3); cla; subplot(2, 2, 4); cla; 
%     ulen = size(useq, 2);
%     % pad prediction with zeros
%     if size(cfg.Upredictions, 2) < ulen, cfg.Upredictions(1:sum(model.p), ulen) = 0; end;
%     for i = 1:length(iall),
%         % double integrator assumed here!
%         subplot(2, 2, 1); plot(Tx(2*(i-1)+1, maxpath), agentcolors{iall(i)}); 
%         hold on;
%         subplot(2, 2, 2); plot(Tx(2*i, maxpath), agentcolors{iall(i)});
%         hold on;
%         subplot(2, 2, 3); 
%         if iall(i) == iagent,
%             stairs(useq, agentcolors{iall(i)});
%         else
%             % displace other agent's graphs so they can be seen better
%             stairs(cfg.Upredictions(model.qranges(iall(i))+1, 1:ulen)+rand*0.005, agentcolors{iall(i)});
%         end;
%         hold on;
%         leg{i} = sprintf('agent %d', iall(i));  
%     end;
%     subplot(2, 2, 4); plot(2:length(maxpath), Tr(maxpath(2:end)), agentcolors{1});
%     % labels etc.
%     subplot(2, 2, 1); ylabel('pos'); 
%     title(sprintf('k = %d', cfg.k));
%     legend(leg{:});
%     subplot(2, 2, 2); ylabel('vel'); 
%     subplot(2, 2, 3); ylabel('u'); 
%     subplot(2, 2, 4); ylabel('r'); 
%     pause;
% end;


