function ooconsensus(cfg)
% Consensus with optimistic optimization

if nargin < 1, cfg = struct(); end;

% -----------------------------------------------
% Process configuration structure

% default config
% algo config
CFG.run = 0;                        % run control
CFG.plottrial = 0;                  % show some trajectories
CFG.problem = '';                   % what problem to solve
CFG.model_params = {};              % extra parameters for problem calling in 'model' mode
CFG.alpha = [];                     % alpha for building P
CFG.P = [];                         % matrix P which generates reference behavior
CFG.K = [];                         % #steps over which agents are controllable
CFG.norm = 2;                       % norm for optimization (default 2)
CFG.optparams = struct;             % other optim algorithm parameters
    % must specify at least: 'maxu', and if DOO: 'diam' the diameter fcn
CFG.X = [];                         % may contain set of states for reset (or its ID)
CFG.trials = 1;                     % # trials (control episodes)
CFG.maxsteps = 100;                 % max # steps in a trial
CFG.reset = [];                     % trial reset
% save, stats config
CFG.datadir = '';                   % data dir
CFG.datafile = '';                  % save data to file
CFG.saveoptdata = 0;                % remember the data created by optimization algorithm -- memory-intensive!
CFG.plotparams = struct;            % custom plot params
% display config
CFG.verb = 0;                       % verbosity: the higher, the more detailed the messages displayed
CFG.visualize = 0;                  % for online, visualization level (0 = none, 1 = trial-level, 2 = step-level)
CFG.viscfg = struct;                % visualization config options -- note we "piggyback" on magenmpc visualization support
CFG.trialdisp = 10;
% figure config
CFG = setfigprop(CFG, 'addfields'); % figure config fields

% List of fields that define the problem and therefore may NOT be overwritten on load
KEEPFIELDS = {'problem', 'optparams', 'X'};

% Install function defaults for everything else
cfg = parseconfig(cfg, CFG);

if ~cfg.run,        % load data file
    loadvars = {'cfg', 'model', 'optinfoh'};
    if cfg.plottrial,
        loadvars = {loadvars{:}, 'X', 'U', 'GRAPHS', 'pall', 'qall'};
    end;
    cfg1 = cfg; kf = KEEPFIELDS;
    load(cfg.datafile, loadvars{:});
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['Data loaded from [' cfg.datafile '].'], cfg.verb, 1);
end;

% Echo config
dispx('OOCONSENSUS called with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1, 'cfg', 'config');
dispx(cfg.optparams, cfg.verb, 1, 'optparams', 'config');


% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

% -----------------------------------------------
% Main algorithm
if cfg.run,
    
    % create model, get its dimensionality data, other shorthand variables
    model = feval(cfg.problem, 'model', cfg.model_params{:});
    if model.matype(1) == 'v', error('OOCONSENSUS does not yet support variable connectivity'); end;
    nagents = model.nagents;
    % pranges = model.pranges; qranges = model.qranges;
    pall = sum(model.p); qall = sum(model.q); pagent = model.p(1); qagent = model.q(1); K = cfg.K;
    
    % init P if given as a constant alpha
    if isempty(cfg.P) && ~isempty(cfg.alpha),
        cfg.P = model.G .* cfg.alpha; for i = 1:model.nagents, cfg.P(i, i) = 1-sum(cfg.P(i, :)); end;
    end;
    
    % P checks
    if any(cfg.P(:) < 0 | cfg.P(:) > 1), error('P values not in range'); end;
    if any(sum(cfg.P, 2) ~= 1), error('P not a stochastic matrix'); end;
    % note agents are allowed to use their own state even though there
    % is no explicit loop arc in the graph
    [i1, i2] = find(cfg.P & ~model.G);
    if any(i1 ~= i2), error('P has positive weights for non-neighboring agents'); end;
    
    % init constant fields planning config, copying also the settings from config.optparams
    optcfg = struct;
    optcfg = copyfields(cfg.optparams, optcfg);
    optcfg.fun = @ooconsensus_fun;
    
    % if X is string identifier, get actual states from the problem
    if ischar(cfg.X), cfg.X = feval(cfg.problem, 'X0', cfg.X); end;
    % process state set
    X = cfg.X; if iscell(X), X = flat(X); end;
    % if "smart" reset is used, copy states to reset structure 
    % -- unless overridden by existing setting
    if isstruct(cfg.reset) && ~isfield(cfg.reset, 'x0'), cfg.reset.x0 = X; end;
    
    % init histories etc.
    X = NaN(pall, cfg.maxsteps+1, cfg.trials);
    U = NaN(qall, cfg.maxsteps+1, cfg.trials);
    % R = NaN(nagents, cfg.maxsteps+1, cfg.trials);
    GRAPHS = NaN(nagents, nagents, cfg.maxsteps+1, cfg.trials); % for uniformity, keep this even for fixed-conn problems
    % optinfoh = struct; -- cannot do this, we don't know the fields in advance
    if cfg.saveoptdata, optdatah = cell(nagents, cfg.maxsteps, cfg.trials); end;

    trial = 1;
    timestat.run = 0;
    timestat.optimization = nan(cfg.maxsteps, cfg.trials);

    % init visualization config if needed
    if cfg.visualize,
        vcfg = cfg.viscfg;
        vcfg.gview = [];
    end;

    % -----------------------------------------------
    % Run control
    1; % overall initialization
    dispx('Performing OOCONSENSUS online control...', cfg.verb, 0);

    t = cputime;
    while trial <= cfg.trials,       % main loop

        % initialize trial
        k = 1; terminal = 0;
        % initial state
        % if "smart" reset is used, copy trial index as it may be needed
        if isstruct(cfg.reset), cfg.reset.trial = trial; end;
        [X(:, 1, trial), GRAPHS(:, :, 1, trial)] = feval(cfg.problem, 'reset', model, cfg.reset);
        
        timestat.run = timestat.run + (cputime - t);
        % visualize initial state -- note we use the MAPMC options to do
        % this as our algorithm is the same from the visualization point of view
        if cfg.visualize >= 2,
            vcfg.mampcstep = 1;
            vcfg.mampctrial = 0;
            vcfg.k = k - 1;
            vcfg.trial = trial;
            vcfg.ktotal = k - 1 + cfg.maxsteps * (trial - 1); % for multiple trials, this only works in continuing tasks
            [figh vcfg.gview] = feval(model.visualizefun, vcfg);
        end;
        t = cputime;

        2; % trial initialization
        while k <= cfg.maxsteps && ~terminal,        
            
            % compute ref state
            xkagents = reshape(X(:, k, trial), pagent, nagents);
            xstaragents = (cfg.P * xkagents')';
            
            3; % ref state computation

            % perform optimization (planning/mpc) for every agent
            useqagents = NaN(qagent, nagents, K);
            for i = 1:nagents,
                % configuration updates
                optcfg.iagent = i;
                optcfg.funargs = {cfg, model, xkagents(:, i), xstaragents(:, i), K};
                % choose actions using the OO algorithm
                t2 = cputime;
                if cfg.saveoptdata,     [ui, optinfo, optdata] = doosoo(optcfg);
                else                    [ui, optinfo] = doosoo(optcfg);
                end;
                useqagents(:, i, :) = reshape(ui, qagent, K);
                 4; % ith agent optimization completed

                % already stop counting planning time
                timestat.optimization(k, trial) = timestat.optimization(k, trial) + cputime - t2;                
                optinfoh(i, k, trial) = optinfo;
                if cfg.saveoptdata, optdatah{i, k, trial} = optdata; end;
            end;    % agents loop
            
            if any(isnan(useqagents(:))),   % handle action error
                dispx(sprintf('Agent control failed, stopping. k=%d: x=%s u=%s: ', k, ...
                    num2strx(X(:, k, trial), 'vector'), num2strx(useqagents(:, 1), 'vector')), cfg.verb, 3);
                break;
            end;

            % apply the action sequence
            U(:, k:k+K-1, trial) = reshape(useqagents, qall, K);
            for j = 0:K-1,
                [X(:, k+j+1, trial), dummyreward, terminal, GRAPHS(:, :, k+j+1, trial)] = ...
                    feval(model.fun, model, X(:, k+j, trial), U(:, k+j, trial), 'real');
                
                % note we need to log transitions inside this loop
                % update time stats
                timestat.run = timestat.run + (cputime - t);
                % log
                if cfg.verb >= 3,
                    if ~any(diff(model.p)) && ~any(diff(model.q)), 
                        % display per-agent x and u
                        disp(sprintf('k=%d:', k+j));
                        xs = num2str(reshape(X(:, k+j, trial), pagent, nagents)'); 
                        us = num2str(reshape(U(:, k+j, trial), qagent, nagents)'); 
                        inds = '0'+(1:nagents)';
                        s = [repmat('  x', nagents, 1) inds repmat(' = [', nagents, 1) xs ...
                            repmat('],     u', nagents, 1) inds repmat(' = [', nagents, 1) us repmat(']', nagents, 1)];
                        disp(s);
                    else
                        % display compact, one-line state and action
                        disp(sprintf('k=%d: x=%s u=%s', k+j, ...
                            num2strx(X(:, k+j, trial), 'vector'), num2strx(U(:, k+j, trial), 'vector')));
                    end;
                    drawnow;
                end;
                % visualization
                if cfg.visualize >= 2,
                    vcfg.mampcstep = 1;
                    vcfg.mampctrial = 0;
                    vcfg.k = k;
                    vcfg.trial = trial;
                    vcfg.ktotal = k - 1 + cfg.maxsteps * (trial - 1); % this only works in continuing tasks
                    [figh vcfg.gview] = feval(model.visualizefun, vcfg);
                end;
                5; % K+kth step simulated
                t = cputime;    % reset time counter for the next step of simulation
            end;

            % finalize loop
            k = k + K;
        end;

        % update stats
        timestat.run = timestat.run + (cputime - t);

        % visualization
        if cfg.visualize >= 1,
            vcfg.mampctrial = 1;
            vcfg.mampcstep = 0;
            vcfg.trial = trial;
            vcfg.ktotal = k - 1 + cfg.maxsteps * (trial - 1); % this only works in continuing tasks
            [figh vcfg.gview] = feval(model.visualizefun, vcfg);
        end;

        % console feedback of algorithm progress
        if ~mod(trial, cfg.trialdisp), 
            dispx(sprintf('Trial #%d completed.', trial), cfg.verb, 2);
        end;

        % start counting time again, increment iteration counter
        t = cputime;
        trial = trial + 1;
    end;        % while not converged and allowed more iterations

    dispx(sprintf('trials=%d exhausted. Algorithm stopped.', cfg.trials), cfg.verb, 0);

    % finalize visualizer
    if cfg.visualize,
        vcfg.mampctrial = 0;
        vcfg.mampcstep = 0;
        vcfg.finalize = 1;
        [figh vcfg.gview] = feval(model.visualizefun, vcfg);
    end;
    
end;

% Plot a trial
% Note the default plot mode will likely be a jungle, it's best to implement a plot function for the
% multiagent problem
if cfg.plottrial,
    Ns = find(~isnan(U(1, :, cfg.plottrial)), 1, 'last');
    hist.Ns = Ns;
    hist.t = (0:Ns) * model.Ts; 
    hist.x = X(:, 1:Ns+1, cfg.plottrial); 
    hist.u = U(:, 1:Ns+1, cfg.plottrial); 
    hist.r = NaN(model.nagents, Ns+1, 1);     % dummy rewards in order to obey the standard
    hist.G = GRAPHS(:, :, 1:Ns+1, cfg.plottrial);
    % might want to add the leader rewards and perhaps reference state here
    % hist.R = discreturn(cfg, hist.r, Ns, terminal);
    mpfigh = [];
    % retroactively add plot function
    if exist([model.id '_plot'], 'file'), model.plotfun = [model.id '_plot']; end;
    % if the model provides a plot function, use it
    if isfield(model, 'plotfun'), 
        mpcfg = struct;
        mpcfg.traj = hist;
        mpcfg.datasource = 'caller';
        mpcfg = copyfields(cfg, mpcfg, fieldnames(setfigprop(struct, 'addfields')));
        mpcfg = copyfields(cfg.plotparams, mpcfg);
        mpfigh = feval(model.plotfun, mpcfg);
    end;
    % otherwise, or if the plot function produced nothing, make a generic plot
    if ~isempty(mpfigh), figh = mpfigh;
    else figh = plothistory(hist); 
    end;
    setfigprop(cfg);
    % save if requested
    saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);
end;

% -----------------------------------------------
% Backup data

if cfg.run && ~isempty(cfg.datafile),
    % save into the same directory as the problem
    if isempty(cfg.datadir),
        datadir = fileparts(which(cfg.problem));
    else
        datadir = cfg.datadir;
        if (datadir(end) == '\'), datadir = datadir(1:end-1); end;
    end;
    cfg.datafile = [datadir '\' cfg.datafile];
    save(cfg.datafile);
    dispx(['Data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;

end
% END RETURNING varargout =================================================================


% Optimization function for OO consensus. Distance between Kth state and
% the desired state xstar
function r = ooconsensus_fun(useq, cfg, model, xk, xstar, K)
% useq is provided as a ``raw'' vector by the optimization algorithm
useq = reshape(useq, model.q(1), K);
xseq = feval(model.simagent, model, xk, useq, K);
% cfg.norm interpreted as p in Lp norm
r = -norm(xseq(:, K+1) - xstar, cfg.norm);     
end