function varargout = mainvpend_problem(what, varargin)
% Integrator problem setup.
%   OUT = IPSETUP_PROBLEM(WHAT)
% This function conforms to the specifications established by SAMPLE_PROBLEM.
%
% Signature for 'model' mode:
%   MODEL = DC_PROBLEM('model', CFG)

switch what
    case 'shortid',
        varargout = {'maip'};
    case 'info',    % basic info about the model (without creating the actual model)
        info.id = 'maip'; info.problem = @mainvpend_problem; info.det = 1; 
        varargout = {info};
                
    case 'model'
        % default params
        % 1) multirobot system parameters
        CFG.nagents = [];                       % number of agents (robots)
        CFG.G = [];                             % communication graph
        
        % config params
        % bounds
        CFG.maxomega = 15*pi;
        CFG.maxu = 3;         
        % electrical parameters
        CFG.K = 53.6e-3;            % torque constant [Nm/A]
        CFG.R = 9.50;               % rotor resistance [Ohm]
        CFG.Km = CFG.K / CFG.R;
        % mechanical
        CFG.b = 3.0e-6;             % viscous damping [Nms/rad]
        CFG.J = 9.3960e-005;        % (rotor and disk) for metal (guesswork...)
        CFG.l = 0.042;              % radius where weight is placed
        CFG.m = 0.055;
        % moment of inertia
        CFG.Jm = CFG.m * CFG.l^2;   % asymmetrical weight
        CFG.J = CFG.J + CFG.Jm;     % total
        % numerical integration
        CFG.Ts = .05; CFG.odemethod = 'fixed-ode'; CFG.odesolver = 'ode4'; CFG.odesteps = 1;
        
        % process any user-configured fields
        model = parseconfig(varargin{1}, CFG);
        
        % create standard model variables (these are "set in stone") 
        % suffix "1" means the variables are for a single agent
        model.id = 'mainvpend';                 % an identifier
        model.matype = 'fixedconn';             % connection graph does not vary
        model.p1 = 2; model.q1 = 1;             % always 2 states, 1 action per robot
        model.p = model.p1 + zeros(1, model.nagents); model.q = model.q1 + zeros(1, model.nagents); 
        model.pranges = [0 cumsum(model.p)];  model.qranges = [0 cumsum(model.q)];
        % set bounds in standard format
        model.maxx1 = [pi; model.maxomega]; 
        model.maxu1 = model.maxu; 
        model.maxx = repmat(model.maxx1, model.nagents, 1);
        model.maxu = repmat(model.maxu1, model.nagents, 1);  
        model.simagent = @mainvpend_simagent;
        model.fun = @mainvpend_mg;
        model.mgl = model.m * 9.81 * model.l;    % helper var
        
        % create discretization helper vars
        switch model.odemethod,
            case 'ode',         model.odet = [0 model.Ts/2 model.Ts]; model.odeopt = odeset;
            case 'fixed-ode',   model.odet = 0 : model.Ts / model.odesteps : model.Ts;
            case 'euler',       % do nothing
        end;
        
        varargout = {model};
         
    case 'reset',   % handle state reset
        RES.type = [];
        RES.x0 = [];
        m = varargin{1};
        res = varargin{2};
        if ischar(res) && any(strcmp(res, {'rand', 'randangles', 'values'})),   % handle simple method identifier
            res = struct('type', res);
        elseif isnumeric(res),
            res = struct('type', 'values', 'x0', res);
        end;
        
        res = parseconfig(res, RES);
        switch res.type,
            case 'rand',    x0 = (2*rand(sum(m.p), 1) - 1) .* m.maxx;
            case 'randangles', 
                x0 = [(2*rand(1, m.nagents) - 1) .* pi;
                    zeros(1, m.nagents)];
            case 'values',  x0 = res.x0;
        end;          
        % just return the fixed connection graph
        G = m.G;
    
        % flatten and output x, output G
        varargout = {x0(:), G};
end;

end % END doubleint_problem(), RETURNING out ====================================

