function figh = malinear_plot(cfg)
% Plot multiagent linear system evolution, with a focus on flocking

% default arguments
if nargin < 1, cfg = ''; end;
% support for default calling mode -- i.e. only history
if isstruct(cfg) && isfield(cfg, 't') && isfield(cfg, 'x'),
    hist = cfg; cfg = struct; cfg.traj = hist;
    cfg.datasource = 'caller';
end;

CFG.problem = 'malinear_problem';
CFG.datasource = '';
CFG.datafile   = '';    % data source can be given as a data file here as well
CFG.traj = [];          % plot history (given as standard history structure)
CFG.state = [];         % show a schematic of a given state
CFG.showranges = 0;     % in 'state' mode, whether ranges should be shown
CFG.enhancedposplot = 1;% in 'traj' mode, whether pos plot should include graph, final vels etc.
% plot configuration
CFG = setfigprop(CFG, 'addfields');  % add figure settings fields to CFG

% process config
cfg = parseconfig(cfg, CFG);
if ~isempty(cfg.datafile), cfg.datasource = cfg.datafile; end;

cfg

% needed vars
vars = {'model', 'cfg'};
if cfg.state, vars = {vars{:}, 'X', 'GRAPHS'}; end;
% load model, grids, etc. from the data source
cfg1 = cfg;
switch cfg.datasource,
    case 'caller',              % load needed vars from caller space
        for i=1:length(vars), cv.(vars{i}) = evalin('caller', vars{i}); end;
        structtovars(cv);
    otherwise                   % load from file
        load(cfg.datasource, vars{:});
end;
expcfg = cfg; cfg = cfg1; clear cfg1; % swap config

% color style
cs = struct;
cs.agentcolors = {'b', 'r', 'g', 'k', 'm', 'y', 'c', [.5 .5 .5], [1 .5 .5]};
cs.agenttraces = repmat({{'LineStyle', ':'}}, 1, length(cs.agentcolors));
cs.agentlines = repmat({{}}, 1, length(cs.agentcolors));
cs.agentsize = .3;
cs.connections = {'Color', [.5 .5 .5], 'LineStyle', '-', 'LineWidth', 0.5};
cs.velocity = {'LineStyle', '-', 'LineWidth', 2};
cs.range = {'LineStyle', '--', 'LineWidth', 1};
cs.return = {'Color', 'g', 'LineWidth', 1.5};
% grayscale style starting from color style
gs = cs;
% choose style
if cfg.grayscale, sty = gs; else sty = cs; end;
% labels
labels = struct;
labels.pos = {'p_1', 'p_2'}; labels.vel = {'v_1', 'v_2'}; labels.u = {'u'}; labels.time = 'k*T_s [s]';


% %%%%%%%%%%%%%%%%%%%%%%
% PLOT DATA

figh = [];

if ~isempty(cfg.traj),
    figh(end+1) = figure;
    traj = cfg.traj;
    % pick up positions, velocities
    x = reshape(traj.x, model.p(1), model.nagents, traj.Ns+1);
    pos = x(1:model.dim, :, :); vel = x(model.dim+1:end, :, :);
    switch model.dim,
        case 1,
            % expand into 2D showing agent evolution along horiz axis
%             offs = symequidgrid(model.nagents, 1);
%             offs = repmat(offs', 1, traj.Ns+1) + repmat(0:traj.Ns, model.nagents, 1);
            if cfg.enhancedposplot,
                offs = repmat(0:traj.Ns, model.nagents, 1);
                pos2 = [reshape(offs, [1, size(offs)]); pos]; 
                vel2 = [0*vel; vel];
                tracesubplot = {1, 2, 1};
            else
                possubplots = {{1, 2, 1}};
            end;
            velsubplots = {{1, 2, 2}};
        case 2,
            if ~cfg.enhancedposplot, error('regular position plot for 2D agents not implemented'); end;
            pos2 = pos;
            vel2 = vel;
            tracesubplot = {2, 2, [1 2]};
            velsubplots = {{2, 2, 3}, {2, 2, 4}};
        otherwise,
            error('malinear_plot: up to 2D systems supported');
    end;
    
    % positions plot
    if cfg.enhancedposplot, % enhanced variant
        subplot(tracesubplot{:});
        for i = 1:model.nagents,
            % traces; connections; final positions; final velocities
            plot(squeeze(pos2(1, i, :)), squeeze(pos2(2, i, :)), 'Color', sty.agentcolors{i}, sty.agenttraces{i}{:}); hold on; 
            lastpos = pos2(:, :, end); lastvel = vel2(:, :, end); lastG = traj.G(:, :, end);
            connx = []; conny = [];
            for j = i+1:model.nagents,
                if lastG(i, j), 
                    connx = [connx [lastpos(1, i); lastpos(1, j)]];
                    conny = [conny [lastpos(2, i); lastpos(2, j)]];
                end;
            end;
            line(connx, conny, sty.connections{:});
            hh = rectangle('Position', [lastpos(:, i)' 0 0] + [-.5 -.5 1 1] * sty.agentsize, 'Curvature', [1,1]);
            set(hh, 'EdgeColor', 'none', 'FaceColor', sty.agentcolors{i});
            hh = line([lastpos(1, i) lastpos(1, i) + lastvel(1, i)], [lastpos(2, i) lastpos(2, i) + lastvel(2, i)]);
            set(hh, 'Color', sty.agentcolors{i}, sty.velocity{:});
        end;
        if model.dim == 1,  xlabel(labels.time); ylabel(labels.pos{1});
        else
            xlabel(labels.pos{1}); ylabel(labels.pos{2}); 
            % both positions have same units in 2D
            axis equal;
        end;
    else % regular position plot
        for d = 1:model.dim,
            subplot(possubplots{d}{:}); 
            for i = 1:model.nagents,
                plot(traj.t, squeeze(pos(d, i, :)), 'Color', sty.agentcolors{i}, sty.agentlines{i}{:}); 
                hold on;
            end;
            xlabel(labels.time); ylabel(labels.pos{d});
        end;
    end;
    
    % velocities plot (always regular)
    for d = 1:model.dim,
        subplot(velsubplots{d}{:}); 
        for i = 1:model.nagents,
            plot(traj.t, squeeze(vel(d, i, :)), 'Color', sty.agentcolors{i}, sty.agentlines{i}{:}); 
            hold on;
        end;
        xlabel(labels.time); ylabel(labels.vel{d});
    end;
end;

% state
if ~isempty(cfg.state),
    % display initial state
    % this code is very similar to that in malinear_visualize
    figh(end+1) = figure;
    x = reshape(X(:, 1, 1), model.p(1), model.nagents);
    G = GRAPHS(:, :, 1, 1);
    pos = x(1:model.dim, :);
    vel = x(model.dim+1:end, :);
    % expand into 2D if 1D agents
    if model.dim == 1, pos = [pos; 0*pos]; vel = [vel; 0*vel]; end;
    % show agents at updated (next) positions, update velocities
    % at the same time make the connection lines
    connx = []; conny = [];
    for i = 1:model.nagents,
        % draw bullet representing current state, and velocity line
        gview.agents(i) = rectangle('Position', [pos(:, i)'-sty.agentsize/2 sty.agentsize sty.agentsize], ...
            'Curvature', [1,1]);
        set(gview.agents(i), 'EdgeColor', 'none', 'FaceColor', sty.agentcolors{i});
        gview.velocities(i) = line([pos(1, i) pos(1, i) + vel(1, i)], [pos(2, i) pos(2, i) + vel(2, i)]);
        set(gview.velocities(i), 'Color', sty.agentcolors{i}, sty.velocity{:});
        if cfg.showranges,
            gview.ranges(i) = rectangle('Position', [-1 -1 2 2] * model.range + [pos(:, i)' 0 0], ...
                'Curvature', [1,1]);
            set(gview.ranges(i), 'EdgeColor', sty.agentcolors{i}, 'FaceColor', 'none', sty.range{:});
        end;
        % note we only consider superdiagonal adjacencies to avoid repeated lines
        for j = i+1:model.nagents,
            if G(i, j),
                connx = [connx [pos(1, i); pos(1, j)]];
                conny = [conny [pos(2, i); pos(2, j)]];
            end;
        end;
    end;
    % note connx conny might also be empty!
    if ~isempty(connx), gview.connections = line(connx, conny, sty.connections{:}); end;
    axis equal off;
end;

% save last figure if indicated
if ~isempty(figh), saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget); end;

end
% END RETURNING array of figure handles =================================================

