function  varargout = manonhol_problem(mode, varargin)
% Markov game with decoupled nonholonomic agents.
%   OUT = MANONHOL_PROBLEM(MODE, CONFIG)
%   [X0, G] = MANONHOL_PROBLEM('reset', MODEL, RESETCONFIG);
%   [GSTAR, RHOSTAR, LAMBDA2STAR] = MANONHOL_PROBLEM('minrobgraph', MODEL, X);


switch mode,

    % Offer basic info about the model (without creating the actual model)
    case 'info',    
        info.id = 'manonhol';               % an identifier
        info.problem = @manonhol_problem;   % handle to model function
        info.det = 1;                       % whether problem is deterministic or stochastic
        varargout = {info};

    case 'shortid',    
        varargout = {'manonhol'};               % an identifier
    
    % The 'model' mode is required
    case 'model',
        % these should stay fixed
        m.id = 'manonhol';                  % an identifier
        m.problem = @manonhol_problem;      % handle to m function
        m.det = 1;                          % whether problem is deterministic or stochastic        
        m.fun = 'manonhol_mg';              % dynamics and rewards function
        m.plotfun = 'manonhol_plot';       
        m.visualizefun = 'manonhol_visualize';
%         m.simagent = 'manonhol_simagent';
        
        % configurable
        m.nagents = [];                     % number of agents
        m.matype = 'varconn';               % 'varconn' or 'fixedconn' problem
        m.G = [];                           % connectivity graph, for fixed connectivity
        m.range = [];                       % communication range, for variable connectivity
        m.Ts = 1;                           % sample time
        m.type = '';                        % agent type: unused so far
        m.conn = 1:2;                       % connectivity variables
        m.agree = 3:4;                      % agreement variables
        m.rewtype = 'flock';                % always requiring flocking reward
        % reward parameters:
        m.Q = [];                           % agreement norm weighting
        m.beta = 0.5;                       % weighting of connectivity cost term
        
        % process any custom config
        if nargin >= 2, m = parseconfig(varargin{1}, m); end;
        
        switch m.matype,
            case 'varconn',
            case 'fixedconn',
            otherwise,
                error('Unknown multiagent type [%s]', m.matype);
        end;
        
        m.p = 4 + zeros(1, m.nagents);
        m.q = 2 + zeros(1, m.nagents);
        m.pranges = [0 cumsum(m.p)]; m.qranges = [0 cumsum(m.q)];   % variable ranges

        switch m.rewtype,
            case 'flock',
                % error checks and defaults
                if isempty(m.range), error('Missing range for [flock] rewards');  end;
            otherwise,
                error('Unknown reward type [%s]', m.rewtype);
        end;
        varargout = {m};

    case 'reset',   % handle state reset -- only specific value supported
        m = varargin{1};
        x0 = varargin{2};
        % compute connections graph
        if m.matype(1) == 'v',  G = manonhol_buildgraph(m, x0);
        else                    G = m.G;
        end;
    
        % flatten and output x, output G
        varargout = {x0(:), G};
        
end;        % mode SWITCH

end % =================== MAIN FUNCTION

% END sample_problem() RETURNING out
% ===================================================