function optimexp(cfg)
% Run and evaluate experiments of policy optimization; or BF optimization for QI 
% Limitations: 
%   - only determ systems supported for Q-iteration w/ BF optimization
%   - does not allow score recomputation (as ceexp_eval did) for a different X0
%   - does not support best/worst policy/statistics, as ceexp_eval did
%   - not yet supported: plotting quantities for each experiment (such as stats)

% TODO cerbfqi, cefzqi "run" functionality not tested yet (cefzqi "plot"
% functions tested)

if nargin < 1, cfg = struct; end;

% common params
CFG.problem = '';
CFG.model_params = {};
CFG.alg = 'cerbfps';                % currently supported: cerbfps, cerbfqi, cefuzzqyqi, optrbfps
CFG.run = 0;                        % run algorithm
CFG.stats = 0;                      % display stats (=1 for all stats; or ='perf','runtime','iter' to select)
CFG.perf = 0;                       % plot performance
CFG.iter = 0;                       % plot number of iterations
CFG.runtime = 0;                    % plot runtime
CFG.cleanup = 0;                    % cleanup datafiles
CFG.cleanuptype = 'exper';          % 'exper' for experiment files; 'dir' for all files in cfg.datadir or crtdir
CFG.repeat = 1;                     % repeat each experiment this many times
CFG.startat = 1;                    % offset for experiment index
CFG.gamma = [];                     % discount factor
CFG.X0 = [];
CFG.mc_maxsteps = .005;             % when < 1, admissible error; when >= 1, trajectory length
CFG.mc_nsim = 10;                   % how many simulations for stochastic systems
CFG.model_params = {};      
CFG.Jstar = [];                     % optimal cost to compare to
%
% policy search options
CFG.N = [];
CFG.actsel = 'max';
PSFIELDS = {'X0', 'gamma', 'mc_maxsteps', 'mc_nsim', 'actsel', 'problem', 'model_params', 'datadir', 'verb'};
% ABF for Qiteration options
CFG.Np = [];
CFG.cost = 'mc';                    % cost type (mc Monte Carlo, br Bellman residual)
CFG.type = 'crad';                  % type of RBF radii for rbfqi
CFG.qi_eps = .001;                  % threshold for QI convergence
CFG.qi_maxiter = 500;               % max number of iterations for Q-iteration
CFG.qi_term = 'ignore';             % handling terminal states: zero or ignore
QIFIELDS = {'X0', 'gamma', 'mc_maxsteps', 'mc_nsim', 'cost', 'type', 'problem', 'model_params', 'datadir', 'verb', ...
    'qi_eps', 'qi_maxiter', 'qi_term'};
% CE optimization options
CFG.ce_c = 5;                       % Nsamples = c * #metaparams
CFG.ce_rho = 0.05;                  % percentile/number of elite samples
CFG.ce_alpha = 1;                   % smoothed update coefficient
CFG.ce_d = 5;                       % # of iterations score should be roughly constant
CFG.ce_eps = .005;                  % threshold for score increase
CFG.ce_maxiter = 50;                % max # iter
CEFIELDS = {'ce_c', 'ce_rho', 'ce_alpha', 'ce_d', 'ce_eps', 'ce_maxiter'};
% Generic optimization options (optrbfps)
CFG.solver = 'glcFast';
CFG.maxstarts = 1;
CFG.eps = .01;
CFG.solveropt = struct;
OPTFIELDS = {'solver', 'maxstarts', 'eps', 'solveropt'};
%
% datafile config
CFG.datadir = [];
CFG.fixedprefix = '';               % used this fixed prefix + param values (for compatibility reasons)
CFG.infix = '';
CFG.postfix = '';
CFG.hide = {};                      % prevent adding N, Np, rep to the file name
CFG.show = {};                      % force adding these params to the file name (besides N, Np)
CFG.miscinfo = '';                  % misc useful info about the experiment
% figure & output config
CFG.int = 'minmax';                 % 'minmax'; or 'conf' for 95% confidence intervals
CFG.conftype = 'gauss';             % 'gauss' or 'ttest'
CFG.statspec = '%g';                % specifier for printing stats values
CFG.legendprefix = '';              % if plotting multiple experiments
CFG = setfigprop(CFG, 'addfields');
CFG.verb = 5;

cfg = parseconfig(cfg, CFG);
% config dependencies
if isempty(cfg.addtocrt), cfg.addtocrt = 0; end;

% main varying parameter: number of basis functions has different names
switch cfg.alg,
    case {'cerbfps', 'optrbfps', 'cerbfqi'},    Nvar = 'N';
    case 'cefuzzyqi',                           Nvar = 'Np';
end;
nexp = length(cfg.(Nvar));  % # of experiments 

% === RUN EXPERIMENTS ===
if cfg.run,     % one repetition at a time, containing all different experiments
    for rep = cfg.startat:(cfg.startat+cfg.repeat-1),
        for iexp = 1:nexp,
            % form CE config
            algcfg = struct;
            algcfg.run = 1;
            switch cfg.alg,
                case 'cerbfps',
                    algcfg.N = cfg.N(iexp);
                    algcfg.datafile = gen_dfname(cfg, cfg.N(iexp), rep);
                    algcfg = copyfields(cfg, algcfg, CEFIELDS);
                    algcfg = copyfields(cfg, algcfg, PSFIELDS);
                case 'optrbfps',
                    algcfg.N = cfg.N(iexp);
                    algcfg.datafile = gen_dfname(cfg, cfg.N(iexp), rep);
                    algcfg = copyfields(cfg, algcfg, OPTFIELDS);
                    algcfg = copyfields(cfg, algcfg, PSFIELDS);                    
                case 'cerbfqi', 
                    algcfg.N = cfg.N(iexp);
                    algcfg.datafile = gen_dfname(cfg, cfg.N(iexp), rep);
                    algcfg = copyfields(cfg, algcfg, CEFIELDS);
                    algcfg = copyfields(cfg, algcfg, QIFIELDS);
                    % ceXXXqi has not been updated to new style of handling X0
                    algcfg.ce_params = {cfg.X0}; 
                case 'cefuzzyqi',     
                    algcfg.Np = cfg.Np(iexp);
                    algcfg.datafile = gen_dfname(cfg, cfg.Np(iexp), rep);
                    algcfg = copyfields(cfg, algcfg, CEFIELDS);
                    algcfg = copyfields(cfg, algcfg, QIFIELDS);
                    % ceXXXqi has not been updated to new style of handling X0
                    algcfg.ce_params = {cfg.X0}; 
            end;
            if exist([algcfg.datafile '.mat'], 'file'), 
                dispx(['Skipping exp [' algcfg.datafile '] -- found data file.'], cfg.verb, 5);
                continue; 
            end;
            dispx(['Running exp in set #' num2str(rep) ', df=' algcfg.datafile], cfg.verb, 2);
            feval(cfg.alg, algcfg);
        end;    % FOR experiments
    end;        % FOR repetition
    
end;            % RUN branch


% ==== COMPUTE STATISTICS ====
if any(cfg.stats) || cfg.perf || cfg.iter || cfg.runtime,
    J = zeros(cfg.repeat, nexp);
    T = zeros(cfg.repeat, nexp);
    K = zeros(cfg.repeat, nexp);
    % loop through files collecting interesting data
    for iexp = 1:nexp,
        for rep = cfg.startat:(cfg.startat+cfg.repeat-1),        
            fname = gen_dfname(cfg, cfg.(Nvar)(iexp), rep);
            if ~exist([cfg.datadir fname  '.mat'], 'file'), 
                error(['Filename [' fname '] cannot be found.']);
            end;
            % different algorithms save different data, and require different processing
            switch cfg.alg,
                case {'cerbfps', 'cerbfqi', 'cefuzzyqi'},
                    loadwithprefix([cfg.datadir fname], 'exp_', {'Jstar', 'trun', 'Jhat', 'conv'});
                    if ~exp_conv, dispx(['Warning! Lack of convergence in [' fname ']'], cfg.verb, 0); end;
                    J(rep-cfg.startat+1, iexp) = exp_Jstar;
                    T(rep-cfg.startat+1, iexp) = exp_trun;
                    K(rep-cfg.startat+1, iexp) = find(~isnan(exp_Jhat), 1, 'last');
                case 'optrbfps',
                    loadwithprefix([cfg.datadir fname], 'exp_', {'Jstar', 'trun', 'optres'});
                    J(rep-cfg.startat+1, iexp) = exp_Jstar;
                    T(rep-cfg.startat+1, iexp) = exp_trun;
                    K(rep-cfg.startat+1, iexp) = exp_optres.Iter;
            end;
        end;
    end;
    if cfg.repeat > 1,
        switch cfg.conftype,
            case 'gauss',
                confJ = 1.96*std(J)/sqrt(cfg.repeat);
                confT = 1.96*std(T)/sqrt(cfg.repeat);
                confK = 1.96*std(K)/sqrt(cfg.repeat);
            case 'ttest',
                % test for constant valued samples to avoid division by zero in ttest
                confJ = nan(2, nexp);
                ident = all(diff(J) == 0);
                confJ(:, ident) = [mean(J(:, ident)); mean(J(:, ident))];
                [hdummy, pdummy, confJ(:, ~ident)] = ttest(J(:, ~ident));
                confJ = diff(confJ)/2;
                %
                confT = nan(2, nexp);
                ident = all(diff(T) == 0);
                confT(:, ident) = [mean(T(:, ident)); mean(T(:, ident))];
                [hdummy, pdummy, confT(:, ~ident)] = ttest(T(:, ~ident));
                confT = diff(confT)/2;
                %
                confK = nan(2, nexp);
                ident = all(diff(K) == 0);
                confK(:, ident) = [mean(K(:, ident)); mean(K(:, ident))];
                [hdummy, pdummy, confK(:, ~ident)] = ttest(K(:, ~ident));
                confK = diff(confK)/2;
            otherwise,
                error('Unknown confidence interval type [%s]', cfg.conftype);
        end;
        % compute statistics: [minimum, mean, maximum, and 95% conf bound on the mean]
        stats.J = [min(J); mean(J); max(J); confJ];
        stats.T = [min(T); mean(T); max(T); confT];
        stats.K = [min(K); mean(K); max(K); confK];
    else
        % simply pick up the values, no statistics to compute
        stats.J = J; stats.T = T; stats.K = K;
    end;
    
    % publish the statistics to the global workspace
    global OPTIMEXP_STATS
    OPTIMEXP_STATS = stats;
end;    % COMPUTE STATS branch


% ===== OUTPUT HUMAN-READABLE STATS AT THE CONSOLE ===
if cfg.stats,
    out.J = cell(nexp, 3);
    out.T = cell(nexp, 3);
    out.K = cell(nexp, 3);
    quantids = 'JTK';
    for iexp = 1:nexp,
        for i = 1:length(quantids),
            qid = quantids(i);
            qval = stats.(qid)(:, iexp);    % pick up values for easy access
            % an informative label about the experiment
            out.(qid){iexp, 1} = sprintf('%s=%d', Nvar, cfg.(Nvar)(iexp));
            % mean and confidence interval on the mean (with possibly custom format spec)
            out.(qid){iexp, 2} = sprintf(sprintf(...
                '%s [%s, %s]', cfg.statspec, cfg.statspec, cfg.statspec), ...
                qval(2), qval(2)-qval(4), qval(2)+qval(4));
            % min and max values (with possibly custom format spec)
            out.(qid){iexp, 3} = sprintf(...
                sprintf('%s<*<%s', cfg.statspec, cfg.statspec), ...
                qval(1), qval(3));
        end;
    end;
	% output using informative names
    format long
    if isscalar(cfg.stats) || strcmp(cfg.stats, 'perf'),
        score = out.J
    end;
    if isscalar(cfg.stats) || strcmp(cfg.stats, 'runtime'),
        runtime = out.T
    end;
    if isscalar(cfg.stats) || strcmp(cfg.stats, 'iter'),
        iter = out.K
    end;
    format short
    clear score runtime iter
end;
    


% ==== PLOTS ====
figh = [];

% plot styles
if cfg.addtocrt,        % some styles differ depending on whether adding to an existing figure
    gs.mean = {'-x', 'Color', 'k', 'LineWidth', 1};
    gs.min = {'--x', 'Color', [.3 .3 .3], 'LineWidth', 1};
    gs.max = {'-.x', 'Color', [.3 .3 .3], 'LineWidth', 1};
    gs.conf = {':', 'Color', 'k'}; 
else
    gs.mean = {'-o', 'Color', 'k', 'MarkerEdgeColor', 'k', 'LineWidth', 1.5};
    gs.min = {'--o', 'Color', [.5 .5 .5], 'MarkerSize', 5, 'MarkerEdgeColor', [.3 .3 .3], 'LineWidth', 1.5};
    gs.max = {'-.o', 'Color', [.5 .5 .5], 'MarkerSize', 5, 'MarkerEdgeColor', [.3 .3 .3], 'LineWidth', 1.5};
    gs.conf = {'--', 'Color', 'k'}; 
end;
gs.opt = {'-', 'Color', 'k', 'LineWidth', 2};
gs.confshade = {[.7 .7 .7], 'EdgeColor' , 'none'};
gs.cm = gray(96); gs.cm = gs.cm(24:end);        % black to dirty white
% copy grayscale styles to color styles; make necessary changes
cs = gs;                                        
if cfg.addtocrt,        % styles for add to current
else
    cs.mean = {'-o', 'Color', 'b', 'MarkerEdgeColor', 'b', 'LineWidth', 1.5};
    cs.conf = {'--', 'Color', 'b'}; 
end;
cs.confshade = {[.8 .8 .9], 'EdgeColor' , 'none'};
% set style to color or grayscale depending on context
if cfg.grayscale, sty = gs; else sty = cs; end; 

% labels
labels.J = 'score'; labels.T = 'execution time'; labels.K = 'number of iterations';
labels.Jopt = 'optimal score';
labels.mean = 'mean '; labels.max = 'max '; labels.min = 'min ';    % reserve spaces for concatenation
labels.conf = '95% confidence bounds';
switch cfg.alg,
    case {'cerbfps','cerbfqi','optrbfps'},  labels.x = 'Number of RBFs';
    case 'cefuzzyqi',                       labels.x = 'Number of fuzzy sets on each axis';
end;
labels.Jy = 'Score';
labels.Ty = 'Execution time [s]';
labels.Ky = 'Number of iterations';

% plot-able quantities
quantnames = {'perf', 'runtime', 'iter'};
% their IDs, pointing to statistics field names, labels etc.
quantids = 'JTK';

figh = [];
for i = 1:length(quantnames),           % if flag requesting quantity plot is set
    if cfg.(quantnames{i}),
        xaxis = cfg.(Nvar);             % values on the X axis for the plot
        quantid = quantids(i);
        quantdata = stats.(quantid);
        if cfg.addtocrt,    figh(end+1) = gcf; hold on;
        else                figh(end+1) = figure;
        end;
        % for every graph: store to legend handles; add legend text
        if cfg.repeat == 1,     % single plot
            lh = cfg.plotfun(xaxis, quantdata, sty.mean{:});
            leg = {[cfg.legendprefix labels.(quantid)]};
        else                    % multiple experiments, statistics plot
            lh = cfg.plotfun(xaxis, quantdata(2, :), sty.mean{:}); 
            hold on;
            leg = {[cfg.legendprefix labels.mean labels.(quantid)]}; 
            if cfg.int(1) == 'm',       % plot min-max interval
                lh(end+1) = cfg.plotfun(xaxis, quantdata(1, :), sty.min{:}); 
                lh(end+1) = cfg.plotfun(xaxis, quantdata(3, :), sty.max{:});
                leg{end+1} = [cfg.legendprefix labels.min labels.(quantid)];
                leg{end+1} = [cfg.legendprefix labels.max labels.(quantid)];
            elseif cfg.int(1) == 'c',   % plot 95% confidence interval
                % plot shade at negative z to put it "behind" everything
                patch([xaxis fliplr(xaxis)], ...
                    [quantdata(2,:)-quantdata(4,:) fliplr(quantdata(2,:)+quantdata(4,:))], ...
                    -.1 * [xaxis xaxis], sty.confshade{:});
                % only one boundary gets a legend (since both have same style & meaning)
                cfg.plotfun(xaxis, quantdata(2,:)-quantdata(4,:), sty.conf{:}); 
                lh(end+1) = cfg.plotfun(xaxis, quantdata(2,:)+quantdata(4,:), sty.conf{:});
                leg{end+1} = [cfg.legendprefix labels.conf];
            end;
        end;

        % if plotting score: plot optimal when given
        if quantid == 'J' && ~isempty(cfg.Jstar),
            lh(end+1) = cfg.plotfun([xaxis(1) xaxis(end)], [cfg.Jstar cfg.Jstar], sty.opt{:});
            leg{end+1} = labels.Jopt;
        end;

        if ~cfg.addtocrt, % place legend, but only if > 1 plot on figure
            if cfg.repeat > 1, 
                lhh = legend(lh, leg{:});
                set(lhh, 'Location', 'Best'); 
            end;
            % axis labels: y label depends on the current quantity being plotted
            xlabel(labels.x); ylabel(labels.([quantid 'y'])); 
            cfg.xtick = xaxis; 
        else            % add to already existing legend
            [dummy1 dummy2 oldobjecth legold] = legend; clear dummy1 dummy2;
            if ~isempty(legold), lhh = legend([oldobjecth(:); lh(:)], legold{:}, leg{:}); 
            else lhh = legend(lh(:), leg{:}); end;
            set(lhh, 'Location', 'Best');
        end;
        % give some informative figure title
        set(figh, 'Name', [cfg.problem '/' cfg.alg '/' cfg.X0], 'NumberTitle', 'off');
        
        % finalize and possibly save figure 
        setfigprop(cfg);
    	saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);          
    end;
end;        % PLOT for/branch


% ==== DATAFILE CLEANUP ====
if cfg.cleanup,
    switch cfg.cleanuptype,        
        case 'exper',       % cleanup experiment files
            for iexp = 1:nexp,
                for rep = cfg.startat:(cfg.startat + cfg.repeat - 1),
                    fname = gen_dfname(cfg, cfg.(Nvar)(iexp), rep);
                    dispx(['Level-' num2str(cfg.cleanup) ' cleanup of [' fname ']'], cfg.verb, 1);
                    optimexp_cleanup(cfg.alg, fname, cfg.cleanup);
                end;
            end;    % experiments loop
            
        case 'dir',         % cleanup datadir or if none specified, current dir
            crtdir = pwd; 
            if ~isempty(cfg.datadir), cd(cfg.datadir); end;     % else stay in current dir
            files = dir;
            for i = 1:length(files),
                extpos = strfind(files(i).name, '.mat');
                if extpos,  % else not a MAT file so we skip it
                    fname = files(i).name(1:extpos-1);
                    dispx(['Level-' num2str(cfg.cleanup) ' cleanup of [' fname ']'], cfg.verb, 1);
                    optimexp_cleanup(cfg.alg, fname, cfg.cleanup);
                end;
            end;
            cd(crtdir);     % make sure we are in the original directory upon returning
            
    end;    % cleanuptype switch
end;        % CLEANUP branch


% === END OPTIMEXP FUNCTION ===
end     




% ==== LOCAL FUNCTIONS ====
function df = gen_dfname(cfg, n, rep)
% Subfunction to generate datafile
if ~isempty(cfg.fixedprefix),
    df = cfg.fixedprefix;
else
    % alg prefix
    switch cfg.alg,
        case 'cerbfps',       df = 'cerbfps';
        case 'cerbfqi',       df = 'cerbfqi';
        case 'cefuzzyqi',     df = 'cefzqi';
        case 'optrbfps',      df = 'optrbfps';
        otherwise,            df = cfg.alg;
    end;
    % problem prefix
    switch cfg.problem
        case 'ddi_problem',      df = [df 'dd'];
        case 'hillcar_problem',  df = [df 'hc'];
        case 'bike_problem',     df = [df 'bike'];
        case 'hiv_problem',      df = [df 'hiv'];
    end;
    % infix
    if ~isempty(cfg.infix), df = [df '_' cfg.infix]; end;

    % add X0 to name (if it is a string)
    if ischar(cfg.X0) && ~any(strcmp(cfg.hide, 'X0')), df = [df '_X0' cfg.X0]; end;
end;

% add # of BFs to name
switch cfg.alg,
    case {'cerbfps', 'optrbfps', 'cerbfqi'}
        if ~any(strcmp(cfg.hide, 'N')), df = [df '_N' num2str(n)]; end;
    case 'cefuzzyqi'
        if ~any(strcmp(cfg.hide, 'Np')), df = [df '_Np' num2str(n)]; end;
end;
 
% add any parameters that are forced to show
for i = 1:length(cfg.show),
    % test if parameter is on the config field and a scalar
    if isfield(cfg, cfg.show{i}) && isscalar(cfg.(cfg.show{i}))
        df = [df '_' cfg.show{i} num2strx(cfg.(cfg.show{i}))]; 
    end;    % else simply don't add this parameter
end;

% postfix
if ~isempty(cfg.postfix), df = [df '_' cfg.postfix]; end;
% repetition index, only if there's more than 1 rep to run, or forced show
if ~any(strcmp(cfg.hide, 'rep')) && (cfg.repeat > 1 || any(strcmp(cfg.show, 'rep'))),
    df = [df '_#' num2str(rep)];
end;
end     % function GEN_DFNAME


function optimexp_cleanup(alg, cleanupdf, level)
% Clean up datafile to save diskspace and loading time
% Parameters:
%   ALG         - algorithm for which to perform cleanup
%   CLEANUPDF   - filename
%   LEVEL       - cleanup level. Possible levels (from light to severe):
%       1   - smart cleanup, saves the best sample and the first elite sample
%       2   - cleanup the entire sample history; evolution replay no longer possible

if ~exist([cleanupdf '.mat'], 'file'), return; end;
% load file name; make sure the same filename variable is kept even 
% if a variable exists with the same name in the file
cdf2 = cleanupdf; load(cleanupdf); cleanupdf = cdf2; clear cdf2;
% check if already cleaned up
if exist('cleanedup', 'var') && (cleanedup >= level), return; end;  % nothing to do 

% cleanup procedure depends on algorithm
switch alg,
    case 'cerbfps',
        if ~exist('phistar', 'var'),
%             disp(['Warning! Final "star" variables not computed in [' cleanupdf ']. Skipping file.']);
%             return;
            disp(['Warning! Final "star" variables not computed in [' cleanupdf ']. Computing them.']);
            Jstar = max(max(J)); [istar, kstar] = find(J == Jstar, 1); 
            phistar = squeeze(PHI{kstar}(istar, :, :));
        end;
        % regardless of the level, cleanup the large inner-iteration variables
        clear j g ind   
        % note the above should also clear phi; but I did not do it early on and phi is not so
        % large, so I keep it as is to avoid changes in behavior across data directories
        if level == 1,
            % for every iteration, save the best parameters;
            % and the parameters of the first elite sample (the quantile)
            cK = find(~isnan(Jhat), 1, 'last');
            PHIbest = zeros(cK, 2*model.p + cfg.nbits, cfg.N);
            PHIelite = zeros(cK, 2*model.p + cfg.nbits, cfg.N);
            for ck = 1:cK,
                cj = J(:, ck); 
                % determine the best sample
                [cjmax, cibest] = max(cj); cibest = cibest(1);  % keep one element
                % determine the first elite sample
                if cfg.ce_rho < 1,  cielite = find(cj >= Jhat(ck), 1, 'last');
                else                cielite = cfg.ce_rho;
                end;
                % make an empty tensor for parameters, save only the best & elite
                PHIbest(ck, :, :) = PHI{ck}(cibest, :, :);
                PHIelite(ck, :, :) = PHI{ck}(cielite, :, :);
            end;
            clear ck cK cj cjmax cibest cielite
            % it is now safe to cleanup the sample history
            clear PHI 
        elseif level == 2,
            % nevermind evolution and history, just cleanup the samples
            % also cleanup any previusly found best and elite samples
            clear PHI PHIbest PHIelite
        end;
        
    case 'cefuzzyqi',
        if ~exist('thetastar', 'var'),
            disp(['Warning! Final "star" variables not computed in [' cleanupdf ']. Skipping file.']);
            return;
        end;
        % THE LINE BELOW FIXES A BUG:
        % for many exps, cgridstar was computed from the last sample c, instead of cstar
        cgridstar = gridvector2cell(reshape(cstar, 1, []), model.p, cranges, model.maxx);
        % regardless of the level, cleanup the large inner-iteration variables
        clear j g theta phi
        if level == 1,
            % Note this was not truly tested since there is no evolution replay for cefuzzyqi
            % for every iteration, save the best & elite parameters
            cK = find(~isnan(Jhat), 1, 'last');
            PHIbest = zeros(cK, Nfree); PHIelite = PHIbest;
            THETAbest = zeros(cK, cfg.N, cfg.M); THETAelite = THETAbest;
            for ck = 1:cK,
                cj = J(:, ck); 
                [cjmax, cibest] = max(cj); cibest = cibest(1);
                if cfg.ce_rho < 1,  cielite = find(cj >= Jhat(ck), 1, 'last');
                else                cielite = cfg.ce_rho;
                end;
                % make an empty tensor for parameters, save only the best & elite
                PHIbest(ck, :) = PHI{ck}(cibest, :);
                PHIelite(ck, :) = PHI{ck}(cielite, :);
                THETAbest(ck, :, :) = THETA{ck}(cibest, :, :);
                THETAelite(ck, :, :) = THETA{ck}(cielite, :, :);
            end;
            clear ck cK cj cjmax cibest cielite
            clear PHI THETA
        elseif level == 2,      % nevermind sol evolution; cleanup all the stuff
            clear PHI PHIbest PHIelite
            clear THETA THETAbest THETAelite
        end;
    case 'cerbfqi',
        if ~exist('thetastar', 'var'),
            disp(['Warning! Final "star" variables not computed in [' cleanupdf ']. Skipping file.']);
            return;
        end;
        clear j g theta phi
        if level == 1, disp('Level 1 cleanup is not supported for rbfqi.'); return;
        elseif level == 2, clear PHI THETA;
        end;
    otherwise,
        error(['Cleanup not supported for algorithm type [' cfg.alg ']']);
end;

cleanedup = level;
clear level alg;
save(cleanupdf);

end






