function out = learn_llssvr(mode, cfg, approx, x, v)
% Implements the local LS SVR approximator for OP
% Currently can be used in combination with OPMDP
% Required parameters on the config:
%   gamma               - discount factor
%   dimx (set automatically in the planner)
%   Knn, Snn, svrsigma, svrgamma - NN and SVR algo params

if mode(1) == 'i',      % 'init' mode
    % initialization consists of creating an empty memory
    approx = struct;
    approx.x = NaN(cfg.dimx, 0, 'double');     % state
    approx.v = NaN(1, 0, 'double');            % values
    approx.memsize = 0;                        % memory initially empty
    
    out = approx;
    
elseif mode(1) == 'e',  % 'estimateb' mode
    % quick-return when the database is too small 
    if approx.memsize < cfg.Knn, out = 1/(1-cfg.gamma); return; end;
    
    % search neighbors
    idx = knnsearchx(x', approx.x', cfg.Knn, cfg.Snn);

    % computation omega and H
    omega = kernel_matrix(approx.x(:,idx)','RBF_kernel', cfg.svrsigma);

    H = omega;
    H = H+diag((cfg.svrgamma^-1)*ones(cfg.Knn,1));
    v = H\approx.v(1,idx)';
    nu = H\ones(cfg.Knn,1);
    % Preconditioned Conjugate Gradients Method
    %   v  = pcg(H,model.ytrain(model.selector,i), 100*eps,model.nb_data);
    %   nu = pcg(H,ones(model.nb_data,i), 100*eps,model.nb_data);
    s = ones(1,cfg.Knn)*nu(:,1);
    beta = (nu(:,1)'*approx.v(idx)')./s;
    alpha = v(:,1)-(nu(:,1)*beta);

    % compute and output value
    kx = kernel_matrix(approx.x(:,idx)', 'RBF_kernel', cfg.svrsigma, x');
    v  = kx'*alpha+ones(size(kx,2),1)*beta;
    v  = min(v,1/(1-cfg.gamma));

    out = v;

elseif mode(1) == 'u',  % 'update' mode
    % reject anything inside the smallest hyperbox containing the new samples
    LB = repmat(min(x, [], 2), 1, approx.memsize);
    UB = repmat(max(x, [], 2), 1, approx.memsize);
    keep = any(approx.x < LB | approx.x > UB, 1);

    % remove remaining duplicates (either due to duplicates on the tree, or because very close to
    % but outside from the edges of hyperbox)
    [approx.x, approx.v] = learn_removeduplicates([approx.x(:, keep), x], [approx.v(keep), v], cfg);
    approx.memsize = size(approx.x, 2);
    
    out = approx;
        
else
    error('Unknown mode [%s]', mode);
end;



% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local functions
function omega = kernel_matrix(Xtrain,kernel_type, kernel_pars,Xt)
% Construct the positive (semi-) definite and symmetric kernel matrix
%
% >> Omega = kernel_matrix(X, kernel_fct, sig2)
%
% This matrix should be positive definite if the kernel function
% satisfies the Mercer condition. Construct the kernel values for
% all test data points in the rows of Xt, relative to the points of X.
%
% >> Omega_Xt = kernel_matrix(X, kernel_fct, sig2, Xt)
%
%
% Full syntax
%
% >> Omega = kernel_matrix(X, kernel_fct, sig2)
% >> Omega = kernel_matrix(X, kernel_fct, sig2, Xt)
%
% Outputs
%   Omega  : N x N (N x Nt) kernel matrix
% Inputs
%   X      : N x d matrix with the inputs of the training data
%   kernel : Kernel type (by default 'RBF_kernel')
%   sig2   : Kernel parameter (bandwidth in the case of the 'RBF_kernel')
%   Xt(*)  : Nt x d matrix with the inputs of the test data
%
% See also:
%  RBF_kernel, lin_kernel, kpca, trainlssvm, kentropy


% Copyright (c) 2011,  KULeuven-ESAT-SCD, License & help @ http://www.esat.kuleuven.be/sista/lssvmlab

[nb_data,d] = size(Xtrain);


if strcmp(kernel_type,'RBF_kernel'),
    if nargin<4,
        XXh = sum(Xtrain.^2,2)*ones(1,nb_data);
        XXh=single(XXh);
        Xtrain=single(Xtrain);
        omega = XXh+XXh'-2*(Xtrain*Xtrain');
        omega = exp(-omega./(2*kernel_pars(1)));
        
        %memory method
%          XXh = sum(Xtrain.^2,2)*ones(1,nb_data);
%          XXhT=transpose(XXh);
%          XX=XXh+XXhT;
%          clear XXh;
%          DP=2*(Xtrain*Xtrain');
%          clear Xtrain;
%          omega1 = XX-DP;
%          clear DP;
%          omega = exp(-omega1./(2*kernel_pars(1)));
         %omega=full(omega);
    else
        XXh1 = sum(Xtrain.^2,2)*ones(1,size(Xt,1));
        XXh2 = sum(Xt.^2,2)*ones(1,nb_data);
        omega = XXh1+XXh2' - 2*Xtrain*Xt';
        omega = exp(-omega./(2*kernel_pars(1)));
    end
    
elseif strcmp(kernel_type,'RBF4_kernel'),
    if nargin<4,
        XXh = sum(Xtrain.^2,2)*ones(1,nb_data);
        omega = XXh+XXh'-2*(Xtrain*Xtrain');
        omega = 0.5*(3-omega./kernel_pars).*exp(-omega./(2*kernel_pars(1)));
    else
        XXh1 = sum(Xtrain.^2,2)*ones(1,size(Xt,1));
        XXh2 = sum(Xt.^2,2)*ones(1,nb_data);
        omega = XXh1+XXh2' - 2*Xtrain*Xt';
        omega = 0.5*(3-omega./kernel_pars).*exp(-omega./(2*kernel_pars(1)));
    end
    
% elseif strcmp(kernel_type,'sinc_kernel'),
%     if nargin<4,
%         omega = sum(Xtrain,2)*ones(1,size(Xtrain,1));
%         omega = omega - omega';
%         omega = sinc(omega./kernel_pars(1));
%     else
%         XXh1 = sum(Xtrain,2)*ones(1,size(Xt,1));
%         XXh2 = sum(Xt,2)*ones(1,nb_data);
%         omega = XXh1-XXh2';
%         omega = sinc(omega./kernel_pars(1));
%     end
    
elseif strcmp(kernel_type,'lin_kernel')
    if nargin<4,
        omega = Xtrain*Xtrain';
    else
        omega = Xtrain*Xt';
    end
    
elseif strcmp(kernel_type,'poly_kernel')
    if nargin<4,
        omega = (Xtrain*Xtrain'+kernel_pars(1)).^kernel_pars(2);
    else
        omega = (Xtrain*Xt'+kernel_pars(1)).^kernel_pars(2);
    end
    
% elseif strcmp(kernel_type,'wav_kernel')
%     if nargin<4,
%         XXh = sum(Xtrain.^2,2)*ones(1,nb_data);
%         omega = XXh+XXh'-2*(Xtrain*Xtrain');
%         
%         XXh1 = sum(Xtrain,2)*ones(1,nb_data);
%         omega1 = XXh1-XXh1';
%         omega = cos(kernel_pars(3)*omega1./kernel_pars(2)).*exp(-omega./kernel_pars(1));
%         
%     else
%         XXh1 = sum(Xtrain.^2,2)*ones(1,size(Xt,1));
%         XXh2 = sum(Xt.^2,2)*ones(1,nb_data);
%         omega = XXh1+XXh2' - 2*(Xtrain*Xt');
%         
%         XXh11 = sum(Xtrain,2)*ones(1,size(Xt,1));
%         XXh22 = sum(Xt,2)*ones(1,nb_data);
%         omega1 = XXh11-XXh22';
%         
%         omega = cos(kernel_pars(3)*omega1./kernel_pars(2)).*exp(-omega./kernel_pars(1));
%     end
end