function  varargout = olop(xroot, model, cfg)
% OLOP Open-loop optimistic planning
%   U = OLOP(XROOT, MODEL, CFG)
%   [U, PLANINFO] = OLOP(XROOT, MODEL, CFG)
%   [U, PLANINFO, PLANDATA] = OLOP(XROOT, MODEL, CFG)
%   Detailed explanation goes here

% Author: Lucian Busoniu
% Limitation: no handling of terminal states
% NOTE: first discount is gamma^1 here, not gamma^0

CFG.gamma = [];             % discount factor
CFG.n = [];                 % simulation budget
CFG.U = [];                 % discrete, flat action space
CFG.exptype = 'one';        % expansion type: simulate 'one' child or 'all' children
CFG.examineucb = [];        % stop and ask to examine UCB after these trajectories
CFG.cucb = 1;               % constant by which the conf interval sqrt(log n / n(u)) should be multiplied
CFG.k = -1;                 % discrete time step
CFG.actiontype = 'one';     % 'one' or 'seq'
CFG.maxseqlen = Inf;        % by default return longest sequence
% display config
CFG.verb = 0;               % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);

% quick-access to config parameters
M = size(cfg.U, 2);      % number of discrete actions
cucb = cfg.cucb;

% start counting time
t = cputime;

H0 = 20;                        % initial max depth (may grow later)
% init tree information (for speed, everything is represented flatly, as vectors)
Ut = zeros(H0, 0, 'uint8');     % sequences (uint8 as M > 256 considered unfeasible)
    % they aren't really used for anything right now, but OK why not store them
Pt = zeros(H0, 0, 'uint32');    % paths (root to current node)
Ct = zeros(M, 0, 'uint32');     % indices of children
Ht = zeros(1, 0, 'uint16');     % depths (uint16 as depth > 65536 considered unfeasible)
Tt = zeros(1, 0, 'double');     % counters (kept double as they are used in math)
Rt = zeros(1, 0, 'double');     % rewards
UCBt = Rt;                      % UCBs
Bt = Rt;                        % B-values
% history of simulated trajectories
Xh = zeros(model.p, H0+1, 0); Uh = zeros(model.q, H0+1, 0); Rh = zeros(1, H0+1, 0);
% and of final states and rewards for all actions, if we expand 'all' children
if cfg.exptype(1) == 'a', Xallh = zeros(model.p, M, 0); Rallh = zeros(1, M, 0); end;

if cfg.verb >= 1, disp(sprintf('Starting planning, n=%d', cfg.n)); end;

% init a template sequence of actions

% init tree root...
Ut(:, 1) = zeros(H0, 1);
Pt(:, 1) = zeros(H0, 1); 
Ct(:, 1) = 0; 
Ht(1) = 0;
Tt(1) = NaN; Rt(1) = NaN;       % undefined for the root
UCBt(1) = NaN; Bt(1) = NaN;     % undefined for the root
% we thus have 1 nodes in the initial tree
nt = 1;

% main simulation loop
tau = 1;                        % sequence/trajectory index
nsamp = 0;                      % number of samples
while nsamp < cfg.n,
    
    % process sequence up to an uninitialized new leaf node
    it = 1;                     % current index in tree (root)
    seq = []; h = 0;            % current sequence and its depth (empty)
    itseq = [];                 % index of all nodes along subsequence
    Xh(:, 1, tau) = xroot;      % init trajectory
    stoptraj = 0;
    while ~stoptraj,
        
        % if the current node is uninitialized, initialize it and all its siblings 
        if ~any(Ct(:, it)),
            % we expand the node and either simulate one, or all children -- either way, we will
            % then stop the simulation
            stoptraj = 1;
            % init all children in the 'unsampled' state
            Ut(1:h+1, nt+1:nt+M) = [repmat(seq(:), 1, M); 1:M];
            Pt(1:h+1, nt+1:nt+M) = [repmat(itseq(:), 1, M); nt+1:nt+M]; 
            Ct(:, nt+1:nt+M) = 0; 
            Ht(nt+1:nt+M) = h+1;
            Tt(nt+1:nt+M) = 0; Rt(nt+1:nt+M) = 0;
            UCBt(nt+1:nt+M) = Inf; Bt(nt+1:nt+M) = Inf;
            % finally, set the children pointers appropriately
            Ct(:, it) = nt+1:nt+M;
            % tree has grown
            nt = nt+M;
        end;

        % we are one step deeper in the traj
        h = h + 1; 

        % either a regular, one-step simulation   
        if ~stoptraj || cfg.exptype(1) == 'o',
            % find a child with the largest B-value
            [bmax jmax] = max(Bt(Ct(:, it)));
            it = Ct(jmax, it);
            seq(h) = jmax; itseq(h) = it;

            % if this child has not yet been simulated, this is also where we stop the trajectory
            % (note this condition only gets the chance to become true for exptype = 'one')
            stoptraj = Tt(it) == 0;

            % simulate the action
            Uh(:, h, tau) = cfg.U(:, jmax);
            [Xh(:, h+1, tau), Rh(1, h+1, tau), term] = feval(model.fun, model, Xh(:, h, tau), Uh(:, h, tau));
            nsamp = nsamp + 1;
            if term, error('OLOP does not currently support terminal states'); end;

            % increment counter, add obtained reward
            Tt(it) = Tt(it) + 1;
            Rt(it) = Rt(it) + Rh(1, h+1, tau);
            
        % or a simulation of all the children (for all actions)
        elseif stoptraj && cfg.exptype(1) == 'a',
            % set special markers signifying 'all actions' in the history
            Xh(:, h+1, tau) = nan;
            Rh(1, h+1, tau) = nan;
            Uh(:, h, tau) = nan;            
            for j = 1:M,
                ic = Ct(j, it);
                seq(h) = j; itseq(h) = ic;
                
                % simulate the action
                [Xallh(:, j, tau), Rallh(1, j, tau), term] = feval(model.fun, model, Xh(:, h, tau), cfg.U(:, j));
                if term, error('OLOP does not currently support terminal states'); end;
                nsamp = nsamp + 1;
                
                % increment counter, add obtained reward
                Tt(ic) = Tt(ic) + 1;
                Rt(ic) = Rt(ic) + Rallh(1, j, tau);
            end;
        end;
    end;
    htraj = h;  % length of simulated traj
    
    % recompute UCB and B for the whole tree
    H = double(max(Ht));            % convert to double as it goes into the math
    discount = cfg.gamma.^(1:H+1);  % note we start from gamma^1, not gamma^0
    
    for h = H:-1:1,
        ih = find(Ht == h);
        leaves = ih(~any(Ct(:, ih), 1));
        innernodes = ih(any(Ct(:, ih), 1));
        nonempty = ih(Tt(ih) > 0);
        
        % update UCB of nonemtpy nodes (the rest remain infinite)
        ihseq = Pt(1:h, nonempty);
        Rth = reshape(Rt(ihseq), size(ihseq)); Tth = reshape(Tt(ihseq), size(ihseq));
        UCBt(nonempty) = discount(1:h) * ( Rth ./ Tth  +  cucb .* sqrt(2*log(tau) ./ Tth) )...
            + discount(h+1) / (1-cfg.gamma);
        
        % set B of leaf nodes to UCB
        Bt(leaves) = UCBt(leaves);
        
        % compute B of inner nodes as min[UCB(node), max(B(children))]
        Bth = reshape(Bt(Ct(:, innernodes)), M, length(innernodes));
        Bt(innernodes) = min( UCBt(innernodes), max(Bth) );
    end;
    
    % log
    if cfg.verb >= 2, disp(sprintf('Trajectory tau=%d processed, n=%d out of %d', tau, nsamp, cfg.n)); end;
    if (cfg.verb >= 3) ...
            || ((cfg.verb == 2.5) && any(Bt(2:nt) ~= UCBt(2:nt))) ...
            || any(cfg.examineucb == tau), % if we examine sequences we also need to show them
        processnodes = 1:nt; % find(isfinite(Bt));
        % show detailed stats
        disp('U/R traj'); disp([Uh(:, 1:htraj, tau); Rh(1, 2:htraj+1, tau)]);
        disp('Counters, rewards, UCBs, B-values for existing sequences');
        S = cell(length(processnodes), 1);
        spec = ['%-' num2str(H) 's: T=%3d, R=%2.2f, UCB=%2.2f, B=%2.2f'];
        for is = 1:length(processnodes),
            i = processnodes(is);
            S{is} = sprintf(spec, char(double(Ut(Ut(:, i) > 0, i))+'0'), Tt(i), Rt(i), UCBt(i), Bt(i));
            if (i>1) && (Bt(i) ~= UCBt(i)), S{is} = [S{is} '****']; end;
        end;
        disp(S);
        pause;
    end;
    if any(cfg.examineucb == tau),
        while true,
            seq = input('Enter a sequence: ', 's');
            if isempty(seq), break; end;
            seqnum = uint8(seq - '0');
            h = length(seqnum);
            it = findflat([seqnum(:); zeros(size(Ut, 1) - h, 1, 'uint8')], Ut);
            ucbparent = discount(1:h-1) * ( Rt(Pt(1:h-1, it)) ./ Tt(Pt(1:h-1, it))  +  cucb .* sqrt(2*log(tau) ./ Tt(Pt(1:h-1, it))) )';
            ucbthis = discount(h) * ( Rt(it) ./ Tt(it)  +  cucb .* sqrt(2*log(tau) ./ Tt(it)) );
            gammah = cfg.gamma^h; 
            terminal = cfg.gamma^(h+1)/(1-cfg.gamma);
            disp(sprintf('UCB(parent: %s) = UCBrewuptop + gamma^h + gamma^h+1/1-gamma = %.2f + %.2f + %.2f = %.2f', ...
                seq(1:end-1), ucbparent, gammah, terminal, ucbparent+gammah+terminal));
            disp(sprintf('UCB(this: %s) = UCBrewuptop + UCBrew(this) + gamma^h+1/1-gamma = %.2f + %.2f + %.2f = %.2f', ...
                seq, ucbparent, ucbthis, terminal, ucbparent+ucbthis+terminal));
        end;
    end;
        
    % increment traj index
    tau = tau + 1;
    
end;    % end main loop

if cfg.actiontype(1) == 'o',    % 'one', root action
    % return a first-level action that has been played the most times
    [tmax, jmax] = max(Tt(Ct(:, 1)));
    ustar = cfg.U(:, jmax);
else                            % 'seq' of actions
    % return a sequence of the desired length that has been played the most times
    if cfg.maxseqlen >= 1, hseq = min(cfg.maxseqlen, H);        % integer max #steps
    elseif cfg.maxseqlen > 0, hseq = ceil(H * cfg.maxseqlen);   % (note < 1) fraction of max length
    else error('Invalid maxseqlen');
    end;
    nodesath = find(Ht == hseq);
    [tmax, imax] = max(Tt(nodesath)); node = nodesath(imax);
    ustar = cfg.U(:, Ut(1:hseq, node)); 
    % recall Ut contains the sequence of action indices corresponding to the node
end;
    
% output stats if requested
if nargout >= 2,
    planinfo.n = nsamp;
    planinfo.tau = tau;
    planinfo.depth = max(Ht);
    planinfo.treesize = nt;
    planinfo.trun = cputime - t;
    planinfo.message = sprintf('Planning done, tau=%d traj, n=%d out of %d, treesize=%d, depth=%d', ...
        planinfo.tau, planinfo.n, cfg.n, planinfo.treesize, planinfo.depth); 
    if cfg.verb >= 1, disp(planinfo.message); end;
end;
if nargout >= 3,
    % compress Ut and Pt to the maximum depth
    Ut = Ut(1:planinfo.depth, :); Pt = Pt(1:planinfo.depth, :);
    plandata = varstostruct('Ut', 'Ht', 'Pt', 'Ct', 'Tt', 'Rt', 'UCBt', 'Bt', 'Xh', 'Uh', 'Rh');
    if cfg.exptype(1) == 'a', plandata.Xallh = Xallh; plandata.Rallh = Rallh; end;
end;

% construct the output based on calling mode
switch nargout,
    case 1, varargout = {ustar};
    case 2, varargout = {ustar, planinfo};
    case 3, varargout = {ustar, planinfo, plandata};
end;
end % main function

% code to:
%     % return a path that has been played the most times
%     node = 1; h = 1;
%     seq = [];     
%     while any(Ct(:, node)),
%         [tmax, jmax] = max(Tt(Ct(:, node)));
%         if tmax == 0, error('tmax = 0, at least one child should have been sampled'); end;
%         seq(h) = jmax; h = h+1;
%         node = Ct(jmax, node);
%     end;
%     ustar = cfg.U(:, seq);
