function  [u, planinfo, plandata] = opmdp(xroot, model, cfg)
% OPMDP Optimistic planning for MDPs / sparsely stochastic systems (OP-MDP, OPSS)
%   U = OPDMP(XROOT, MODEL, CFG)
%   [U, PLANINFO] = OPDMP(XROOT, MODEL, CFG)
%   [U, PLANINFO, PLANDATA] = OPDMP(XROOT, MODEL, CFG)
% This is a revised version of opss, which uses b-values over states rather
% than over state-action pairs (like OPSS did). This is closer to the
% algorithm as given in papers.
% Version using plain arrays for speed. Implements b-function learning
% OPMDP does not implement the following features available in OPSS: 
%   - tree reuse
%   - returning (policy) trees instead of single-actions
%   - compatibility-mode calls

% Author: Lucian Busoniu
% Not yet tested: terminal state handling
CFG.gamma = [];                 % discount factor
CFG.n = [];                     % expansion budget OR:
CFG.ntrans = [];                % number of simulations OR:
CFG.depth = [];                 % target effective depth
CFG.criterion = 'contrib';      % expansion criterion: based on the whole tree: 'unif'
                                % based on optimistic subtree: 'contrib'
CFG.L = [];                     % maximum number of successors for each state-action pair
CFG.U = [];                     % discrete, flat action space
CFG.blearn = [];                % function handle to b-value approximator
    % also be sure to give the parameters of the approx on the config
CFG.k = -1;                     % discrete time step
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed

persistent bapprox              % the approximator must be remembered between calls

cfg = checkparams(cfg, CFG);

if ~xor(xor(~isempty(cfg.n), ~isempty(cfg.ntrans)), ~isempty(cfg.depth)), 
    error('One of: budget n, #sims ntrans, target effective depth must be specified'); 
end;
if ~isempty(cfg.depth) && cfg.criterion(1) ~= 'c', 
    error('Only optimistic variant allowed in effective-depth mode'); 
end;

% quick-access to config parameters, and derived variables
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.gamma; n = cfg.n; criterion = cfg.criterion; L = cfg.L; U = cfg.U; 
blearn = cfg.blearn; blearnflag = ~isempty(blearn);
% derived variables: number of discrete actions, number of state vars
M = size(U, 2); dimx = model.p;

% start counting time
t = cputime;

% 1a) Approximator initialization: if uninitialized, or we are just starting a new trajectory
% must be done before tree, since root init may use the approximator
if blearnflag,      % add some common approximator parameters
    cfg.dimx = dimx;
end;
if blearnflag && (isempty(bapprox) || cfg.k <= 1),
    bapprox = blearn('init', cfg);
end;

% bound on tree size, accounting for budget
if ~isempty(cfg.n), maxtreesize = M*L*n + 1;
elseif ~isempty(cfg.ntrans), maxtreesize = ntrans + 1;
else maxtreesize = 500; % no budget given, use a default value
end;
% prepare tree data (note 0 = null pointer)
Tparent = zeros(1, maxtreesize, 'uint32');        % parent index
Tchildren = zeros(M, L, maxtreesize, 'uint32');   % indices of children for each action
Tdepth = zeros(1, maxtreesize, 'uint16');         % depth (uint16 as depth > 65536 considered unfeasible)
Tleaf = false(1, maxtreesize);                    % leaf flag
Tx = NaN(dimx, maxtreesize, 'double');            % state
Tf = NaN(1, maxtreesize,'double');                % single-step probability
Tr = NaN(1, maxtreesize, 'double');               % reward
Tterm = false(1, maxtreesize);                    % terminal flag
TP = NaN(1, maxtreesize, 'double');               % probability of reaching this node
Tb = NaN(1, maxtreesize, 'double');               % b-values
Tutil = NaN(1, maxtreesize, 'double');            % u-values
% TR = NaN(1, maxtreesize, 'double');               % cumulative rewards - for debugging

% initialize the root node
root = 1;
Tparent(root) = 0;
Tchildren(:, :, root) = 0;
Tdepth(root) = 0;
Tleaf(root) = true;
Tx(:, root) = xroot;
Tf(root) = 1;
Tr(root) = 0;
Tterm(root) = false;    % presumably, we are not asked to make a decision in a terminal state
TP(root) = 1;
% TR(root) = 0;           % debugging
if blearnflag,  Tb(root) = blearn('estimateb', cfg, bapprox, xroot);
else            Tb(root) = 1/(1-discount);
end;
Tutil(root) = 0;
treesize = 1;
    
iter = 0; optimdiam = Inf; stopflag = false;
% 2) Main tree construction loop
while ~stopflag,
    iter = iter + 1;  % increment counter
    
    % 2a) find the candidate leaves for expansion and choose one
    if criterion(1) == 'c', 
        % optimistic leaves (forward traversal)
        frontier = root;  % start at root
        while ~all(Tleaf(frontier)),            % there are still inner nodes to process
            % find all children of the (inner) nodes on the frontier
            innernodes = frontier(~Tleaf(frontier));
            allchildren = Tchildren(:, :, innernodes);
            % inner term of the Bellman backup for each child
            b = Tf(allchildren) .* (Tr(allchildren) + discount .* Tb(allchildren));
            % sum = expected value over next-states, max = best exp value over actions
            [bmax, jmax] = max(sum(b, 2), [], 1);
            % update frontier, letting current leaves unchanged
            frontier = frontier(Tleaf(frontier));
            % and adding all nonzero-proba children of max-b nodes
            for i = 1:length(innernodes), 
                children = allchildren(jmax(i), :, i);
                frontier = [frontier, children(TP(children) > 0)];
            end;
        end;
        % eliminate the terminal nodes and continue with the rest
        % if ALL are terminal, then nothing left to do
        frontier = frontier(~Tterm(frontier)); if isempty(frontier), break; end;    
        % largest contribution to the bound (breaking ties in favor of first)
        contrib = TP(frontier) .* (discount .^ double(Tdepth(frontier))) ./ (1-discount);
        optimdiam = sum(contrib);       % also compute diameter
        [cmax, imax] = max(contrib);
        nodetoexpand = frontier(imax);
    elseif criterion(1) == 'u',             % uniform expansion
        % get all leaves, checking if any are non-terminal
        leaves = find(Tleaf);
        leaves = leaves(~Tterm(leaves));  if isempty(leaves), break; end;    
        % uniform: expand a node with the smallest depth (breaking ties determ)
        [depthmin imin] = min(Tdepth(leaves));
        nodetoexpand = leaves(imin);
    end;
    
    % 2b) expand the chosen node
    Tleaf(nodetoexpand) = false;        % no longer a leaf
    % run through the discrete actions, adding L next-state children for each
    for j = 1:M,
        % use model to determine distribution over next states (+ states, rewards, terminal flags)
        [x, r, term, f] = feval(model.fun, model, Tx(:, nodetoexpand), U(:, j));
        if any(r < 0) || any(r > 1), error('Rewards must be in [0, 1]'); end;
        % pad with default values for zero-proba nodes, so that always L children
        pad =(length(f)+1):L;
        if ~isempty(pad), x(:, pad) = NaN; r(pad) = 0; term(pad) = true; f(pad) = 0; end;
        % find index where new node will be, save it on the children table of current node
        children = treesize+(1:L);
        Tchildren(j, :, nodetoexpand) = children;
        % create new nodes
        Tparent(children) = nodetoexpand;               % parent is expanded node
        Tchildren(:, :, children) = 0;                  % init to "null"
        Tdepth(children) = Tdepth(nodetoexpand) + 1;    % one deeper than the parent
        Tleaf(children) = true;                         % all new nodes are leaves
        Tx(:, children) = x;
        Tf(children) = f;
        Tr(children) = r;
        Tterm(children) = term;
        TP(children) = TP(nodetoexpand) .* f;           % prob = parent prob * single-step prob
%         TR(children) = TR(nodetoexpand) + discount^double(Tdepth(nodetoexpand)) * r;    % debugging
        if blearnflag,  
            for child = children,
                Tb(child) = blearn('estimateb', cfg, bapprox, Tx(:, child));
            end;
        else
            Tb(children) = 1/(1-discount);
        end;
        % override the values of terminal nodes with exactly 0
        Tb(children(logical(term))) = 0;                           
        Tutil(children) = 0;

        % we just added L nodes
        treesize = treesize + L;
    end;    % loop over actions
    
    % 2d) update b-values affected by expanded node
    % these are the b-values of the expanded node itself for all the actions, 
    % and those on the state-ACTION path from the expanded added node to the root
    node = nodetoexpand;
    while node ~= 0,       % when node is 0, it means the root has already been processed
        children = Tchildren(:, :, node);
        b = Tf(children) .* (Tr(children) + discount .* Tb(children));
        Tb(node) = max(sum(b, 2));
        util = Tf(children) .* (Tr(children) + discount .* Tutil(children));
        Tutil(node) = max(sum(util, 2));
        node = Tparent(node);
    end;
    
    % verify stopping condition -- note ntrans condition ignores that sometimes fewer than L
    % successors exist
    stopflag = (~isempty(cfg.n) && (iter >= n)) ...
        || (~isempty(cfg.ntrans) && (iter * L * M >= n)) ...
        || (~isempty(cfg.depth) && (optimdiam <= discount^cfg.depth / (1-discount)));
end;    % end main tree construction loop

% 3) Find action at the tree root
children = Tchildren(:, :, root);
util = Tf(children) .* (Tr(children) + discount .* Tutil(children));
[utilmax, jmax] = max(sum(util, 2));
u = U(:, jmax);

% 4) Update memory, if using it
% note by eliminating leaves we also eliminate training with zeros in terminal states; if we
% wouldn't we would have to be careful about not including the zero-proba nodes
if blearnflag,
    indices = find(~Tleaf(1:treesize));
    bapprox = blearn('update', cfg, bapprox, Tx(:, indices), Tb(indices));
end;

% output stats if requested
if nargout >= 2,
    planinfo.n = iter;
    planinfo.ntrans = iter * L * M; % note this ignores any transitions with fewer than L successors
    planinfo.depth = max(Tdepth);
    if criterion(1) == 'c', planinfo.optimdiam = optimdiam; end;
    planinfo.trun = cputime - t;
    planinfo.treesize = treesize;
    planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d, treesize=%d', ...
        iter, planinfo.depth, planinfo.treesize); 
end;
% output only actually used tree data (no memory)
if nargout >= 3,
    plandata.Tparent = Tparent(1:treesize);
    plandata.Tchildren = Tchildren(:, :, 1:treesize);
    plandata.Tdepth = Tdepth(1:treesize);
    plandata.Tleaf = Tleaf(1:treesize);
    plandata.Tx = Tx(:, 1:treesize);
    plandata.Tf = Tf(1:treesize);
    plandata.Tr = Tr(1:treesize);
    plandata.Tterm = Tterm(1:treesize);
    plandata.TP = TP(1:treesize);
    plandata.Tb = Tb(1:treesize);
    plandata.Tutil = Tutil(1:treesize);
end;

end % main function


