function  varargout = sooplp(x0, model, cfg)
% SOOPLP Planning with SOO or LP
%   U0 = SOOP(X0, MODEL, CFG)
%   [U0, PLANINFO] = SOOP(X0, MODEL, CFG)
%   [U0, PLANINFO, PLANDATA] = SOOP(X0, MODEL, CFG)
%   [USEQ, XSEQ, PLANINFO, PLANDATA] = SOOP(X0, MODEL, CFG)
% In the first three modes, just the root action is returned
% In the last two modes, the action sequence and the corresponding state sequence are returned

% Author: Lucian Busoniu
% Versions:

CFG.gamma = [];                 % discount factor
CFG.n = [];                     % simulation budget
CFG.method = 'soop';            % set selection to use: 'soop', 'lp', 'opc', 'SOOP2'
% SOOP options
CFG.dimselect = [];             % selection method for which dimension to split:
    % 'disc' for discount-factor-based selection
    % 'all' for splitting all dimensions
    % 'lp' for LP criterion
CFG.alpha = [];                 % discount factor for dimselect='disc'
CFG.lipschitz = [];             % Lipschitz constants [L_f, L_rho] (if only one given, assumed L_f = L_rho for back-compatibility)
CFG.aconst = [];                % multiplier in the theoretical criterion "opc"
CFG.dmax = @(tau) sqrt(tau);    % for criterion "SOOP2", function dmax(iter), by default sqrt(iter)
CFG.k = -1;                     % discrete time step
CFG.removeduplicates = 0;        % whether to check for duplicates
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed

% options that would maybe make sense, but are not implemented so far
% CFG.refx = [];                  % ref model state at root
% CFG.reusetree = 0;              % try reusing previous tree (only when k is set and > 1)
% CFG.reusetree_maxsize = 0;      % do not reuse tree if larger (has more nodes) than this; 0 = unlimited size

cfg = checkparams(cfg, CFG);

% parameter and error checks
if strcmp(cfg.method, 'lp') && isempty(cfg.dimselect), cfg.dimselect = 'lp'; end;   % default
if ~any(strcmp(cfg.method, {'soop', 'SOOP2', 'lp', 'opctheor'})), error('Unsupported method [%s]', cfg.method); end;
if ~any(strcmp(cfg.dimselect, {'disc', 'all', 'lp'})), error('Unsupported dimselect [%s]', cfg.dimselect); end;
% judging from the SOOP analysis the disc criterion results from a Lipschitz metric so forbidding
% the combination doesn't make sense
% if strcmp(cfg.method, 'lp') && ~strcmp(cfg.dimselect, 'lp'), error('dimselect must be [lp] for method [lp]'); end;
if model.q > 1, error('SOOP/LP currently supports only scalar actions'); end;
if any(strcmp(cfg.method, {'lp', 'opc'})) && strcmp(cfg.dimselect, 'disc')...
        && cfg.alpha ~= cfg.gamma, disp('Warning! lp or opc used with dimselect=discount, but alpha ~= gamma.'); end;   % since theory says they should match

% quick-access to config parameters, and other helper vars
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
p = model.p; maxu = model.maxu; discount = cfg.gamma; 
method = cfg.method(1); dimselect = cfg.dimselect(1);
L = cfg.lipschitz; if length(L) == 1, L = [L L]; end;   % else, L = [L_f, L_rho]

% start counting time
t = cputime;

% init data structures
initlen = 50;                                   % an initial max length guess
initsize = ceil(1.1 * cfg.n);                   % rough upper bound on #seq
% initlen = 5; initsize = 10;                   % smaller values for debugging
% starting points of U domains, for each depth and sequence: sequence length X # of sequences
SU = zeros(initlen, initsize);                    
% number of splits: sequence length X # of sequences (could be uint16, would require conversion)
SS = zeros(initlen, initsize);     
% state trajectory for center action samples: dimx X sequence length + 1 X # of sequences
% (always includes the root state on first position)
SX = NaN(p, initlen+1, initsize);             
% rewards RESULTING from center action samples: sequence length X # of sequences
% (note they are not shifted by 1, as is usual in the trajectory standard)
SR = zeros(initlen, initsize);
if method == 'l', LIPSCHITZ = lipschitzmatrix(L, initlen-1); end;

% this is to support time-dependent models; k is the real-time in genmpc, not the simulation time
% which would be cfg.k + depth
model.k = cfg.k-1;      % cfg.k is 1-based, model.k is zero-based

% start by trisecting the first dimension
for j = 1:3,
    SU(1, j) = (j-1)/3;
    SS(1, j) = 1;
    SX(:, 1, j) = x0; [SX(:, 2, j), SR(1, j)] = model.fun(model, x0, unnormu((j-0.5)/3, maxu));
end;
Ns = 3;             % number of sequences 
n = 3;              % budget consumed so far
longest = 1;        % length of longest sequence

% initialize depth and dominating value for SOOP2 criterion;
% cycle is the flag for cycling through the entire depth range without finding any set to expand --
% that means infinite loop, and the algo is stopped with an error
if method == 'S', hcrt = 1; rmax = -Inf; cycle = true; end;      
dupcounter = 0;

%#### Main loop
stopflag = false; tau = 0; 
while ~stopflag,
    % iteration counter; equal to number of expansions for all algos except "soop" which may expand
    % many sets at one iteration
	tau = tau + 1;     

    %## Select sets to expand
    expand = false(1, Ns);
    if method == 's',        % SOOP criterion
        % compute returns of all sequences (note all uninitialized rewards are 0)
        ret = (discount .^ (0:longest-1)) * SR(1:longest, 1:Ns);
        for is = 1:Ns,
            % find all sets j >> i (partially greater than i)
            % where j >> i  if  splits j <= # splits i
            greater = all(SS(1:longest, 1:Ns) <= repmat(SS(1:longest, is), 1, Ns), 1);
            % a set is expanded if has a better (or equal) return than all sets partially greater than it
            expand(is) = all(ret(is) >= ret(greater));
            % view ranking if configured
            if cfg.verb >= 5, viewranking(SS, Ns, ret, is, greater, expand); pause; end;
        end;
    elseif method == 'S',
        allh = sum(SS(1:longest, 1:Ns), 1);             % compute all depths
        hmax = max(allh);
        boxfound = false;
        while ~boxfound,
            % test if depth limit or tree depth not exceeded; note tau starts at 1
            if hcrt <= min(cfg.dmax(tau), hmax),        
                % compute returns for all boxes at this depth, pick the maximum one
                boxesath = find(allh == hcrt);
                if ~isempty(boxesath), 
                    reth = (discount .^ (0:longest-1)) * SR(1:longest, boxesath);
                    [rmaxh, imaxh] = max(reth);
                    if rmaxh > rmax,                            
                        % box not dominated by boxes at earlier depth => expand it
                        expand(boxesath(imaxh)) = true; 
                        rmax = rmaxh;                       % increase dominating value
                        boxfound = true; cycle = false;     % set flags
                    end;
                end;
                hcrt = hcrt + 1;
            else
                if cycle, error('SOOP2: No set to expand at iter %d, h=%d', tau, hcrt); end;
                hcrt = min(allh); rmax = -Inf; cycle = true;
            end;
        end;
    elseif method == 'l',    % LP criterion
        % vectorized - but faster - formula
        % a [longest x Ns] matrix containing set half-sizes, measured from the center sample
        subsetsizes = 0.5 * (1/3).^SS(1:longest, 1:Ns);
        % a [longest x Ns] matrix containing the reward-metric half-diameters of each action set
        % note that junk (subsetsize 0.5) is inserted beyond the length of each sequence, but
        % all its influence is eliminated when rewardbounds for 0 splits are capped to 1 below
        rewarddiameters = LIPSCHITZ(1:longest, 1:longest) * subsetsizes;
        % then we add the reward samples
        rewardbounds = SR(1:longest, 1:Ns) + rewarddiameters;
        % bounds greater than 1, as well as those based on uninitialized sets (up to now invalid), are set to 1
        rewardbounds((rewardbounds > 1) | (SS(1:longest, 1:Ns) == 0)) = 1;
        % finally the bound, with the cap gamma^longest/1-gamma
        bvalues = (discount .^ (0:longest-1)) * rewardbounds + discount^longest / (1-discount);
        
%         % simple (but slower/nonvectorized) formula to check the computation above
%         SB = zeros(Ns, 1);
%         for js = 1:Ns,
%             newlen = find(SS(1:longest, js), 1, 'last');
%             deltas = 0.5 * (1/3).^SS(1:newlen, js);
%             rbounds = zeros(newlen, 1);
%             for ik = 1:newlen, 
%                 rbounds(ik) = SR(ik, js) + L(2) .* L(1).^(ik-1:-1:0) * deltas(1:ik);
%             end;
%             SB(js) = (discount .^ (0:newlen-1)) * min(1, rbounds) + discount^newlen / (1-discount);
%         end;
%         idiff = find(abs(SB(1:Ns) - bvalues(1:Ns)') > 1e-10);
%         if ~isempty(idiff),
%             error('mismatch');
%         end;
        
        [bmax, imax] = max(bvalues(1:Ns));  % imax = only the first maximal index
        expand(imax) = true;
    elseif method == 'o',   % theoretical formula for OPC
        for is = 1:Ns,
            T = find(SS(1:longest, is), 1, 'last'); 
            deltas = (1/3).^SS(1:T, is);
            bvalues(is) = discount.^(0:T-1) * (SR(1:T, is) + cfg.aconst * deltas) + cfg.aconst * discount^T / (1-discount);
        end;

        [bmax, imax] = max(bvalues);        % imax just the first index
        expand(imax) = true;        
    end;
    
    %## Expand the selected sets
    for is = find(expand),
        
        if cfg.verb >= 4, viewsets(SU, SS, 1:Ns, is, tau == 1, 'all'); end;   % view focusing on expanded set
        
        len = find(SS(1:longest, is), 1, 'last'); 
        % store the data of the parent set, all subsets inherit from it
        parentsu = SU(1:longest+1, is); parentss = SS(1:longest+1, is); parentsx = SX(:, 1:len+1, is); parentsr = SR(1:len, is);
        
        if dimselect == 'd',     % 'disc'ount based selection
            % rank dimensions by discounted size; note we also rank len+1 dimension where SS is 0
            dimscores = (cfg.alpha .^ (0:len)) .* (1/3).^(parentss(1:len+1)');
            [maxscore, splitdims] = max(dimscores); % splitdims = single index
        elseif dimselect == 'a'  % 'all' dimensions, including the next one
            splitdims = 1:len+1;
        elseif dimselect == 'l', % 'lp', LP algorithm selection
            % rank each existing dimension by how much it would decrease the upper bound
            % we will explicitly recompute the bound, since due to the bounding of rewards by 1, 
            % an analytical computation of the scores would be ugly or impossible
            discounting = discount .^ (0:len-1);
            % construct a matrix with new subset sizes, ith column is the new size after splitting
            % the ith dimension
            newsubsetsizes = repmat(subsetsizes(1:len, is), 1, len);
            newsubsetsizes(1:len+1:len*len) = newsubsetsizes(1:len+1:len*len) ./ 3;
            % compute the new reward bounds, a column corresponds to a dimension split
            newrewbnd = repmat(SR(1:len, is), 1, len) + LIPSCHITZ(1:len, 1:len) * newsubsetsizes;
            % we did not add the cap gamma^d/1-gamma, so remove it from the b-value as well
            dimscores = (bvalues(is) - discount^len / (1-discount)) - discounting * min(newrewbnd, 1);
            % rank the next dimension: assuming the next dimension is split and next reward is 0            
            dimscores(end+1) = max(0, ...
                discount.^len * (1 - LIPSCHITZ(len+1, 1:len+1) * [subsetsizes(1:len, is); 0.5/3]));
            % choose the dimension that maximizes the decrease in upper bound
            [maxscore, splitdims] = max(dimscores);
        end;
        
        % split each of the selected dimensions
        for isplit = 1:length(splitdims),
            ksplit = splitdims(isplit);     % dimension to split
            for j = 1:3,
                % first create the box without saving it or simulating the system
                childsu = parentsu; childss = parentss; 
                % note these updates also work for ksplit = len+1, where parentss and parentsu are both zero
                childss(ksplit) = childss(ksplit) + 1; 
                childsu(ksplit) = childsu(ksplit) + (j-1) * (1/3)^childss(ksplit);  
                if cfg.removeduplicates,
                    % check if the box is not duplicated (accounting for some numerical math errors)
                    dup = all(abs(repmat(childss, 1, Ns) - SS(1:longest+1, 1:Ns)) < 1e-10, 1) ...
                        & all(abs(repmat(childsu, 1, Ns) - SU(1:longest+1, 1:Ns)) < 1e-10, 1);
                    if any(dup), dupcounter = dupcounter+1; continue; end;     % if duplicated, skip the box
                end;
                % otherwise, start creating a new box
                % init new database column EXCEPT when MIDDLE set for the FIRST split, when we overwrite the parent set
                if isplit == 1 && j == 2, js = is; else Ns = Ns+1; js = Ns; end;
                % set new box info as computed above
                SU(1:longest+1, js) = childsu; SS(1:longest+1, js) = childss; 
                % initially, save all state and reward data of the parent
                SX(:, 1:len+1, js) = parentsx(:, 1:len+1); SR(1:len, js) = parentsr(1:len);
                % then it depends which dimension we split:
                if ksplit <= len,       % existing dimension
                    % x & r only change for left and right sets, from the ksplit transition onwards
                    if j ~= 2,
                        for k = ksplit:len,
                            [SX(:, k+1, js), SR(k, js)] = model.fun(model, SX(:, k, js), ...
                                unnormu(SU(k, js) + 0.5*(1/3)^SS(k, js), maxu));
                        end;
                        % update model calls counter
                        n = n + (len-ksplit+1); 
                    end;
                    % for middle set, x & u remain valid (they were in the middle anyway), so nothing to do 
                else                    % ksplit = len+1, the next, undiscretized dimension
                    % x & r data are valid so far, but we need to append a new transition 
                    [SX(:, ksplit+1, js), SR(ksplit, js)] = model.fun(model, SX(:, ksplit, js), unnormu((j-0.5)/3, maxu));
                    % update model calls counter
                    n = n + 1;
                end;    % new-or-existing-dimension branch
                
                % if there would be no more room for the next sequence, expand the data structures
                % -- with current initialization this should happen only rarely
                if Ns+1 > size(SU, 2),
                    [crtlen, crtsize] = size(SU); 
                    delta = ceil(0.5*crtsize);  % expand to 150%
                    SU(:, crtsize+(1:delta)) = zeros(crtlen, delta); SS(:, crtsize+(1:delta)) = zeros(crtlen, delta);
                    SX(:, :, crtsize+(1:delta)) = NaN(p, crtlen+1, delta); SR(:, crtsize+(1:delta)) = zeros(crtlen, delta);
                end;
                
            end;    % loop over left-middle-right subset
            
            % max seq length may have increased after expanding the dimension
            if ksplit > longest, longest = ksplit; end;

        end;    % loop over dimensions to expand for current sequence
        
        % expand length if some set has the last dimension initialized, so we always have extra room
        % note: sequence length may only increase by 1 per expanded set, thus we just need to
        % perform this test once every such set
        if longest == size(SU, 1),
            [crtlen, crtsize] = size(SU); 
            delta = ceil(0.5*crtlen);   % expand to 150%
            SU(crtlen+(1:delta), :) = zeros(delta, crtsize); SS(crtlen+(1:delta), :) = zeros(delta, crtsize);
            SX(:, crtlen+1+(1:delta), :) = NaN(p, delta, crtsize); SR(crtlen+(1:delta), :) = zeros(delta, crtsize);
            % for Lipschitz planning, we must also enlarge the Lipschitz coeff matrix
            if method == 'l', LIPSCHITZ = lipschitzmatrix(L, crtlen+delta-1); end;
        end;

        if cfg.verb >= 4, 
            nnew = 2+3*(length(splitdims)-1); 
            viewsets(SU, SS, [is Ns-nnew+1:Ns], [], tau==1, 'expanded'); 
            pause; 
        end;   % view with expanded seqs in focus
            
    end;    % loop over sequences to expand
    
    % stop when the budget has been exhausted: 
    % note we may use more than the allowed budget to expand the last-selected sets
    stopflag = n >= cfg.n;    
    
    if cfg.verb >= 2,
        fprintf('Iter %d: %d expanded, n=%d trans, #sets=%d, longest=%d; dupsofar=%d\n', ...
            tau, nnz(expand), n, Ns, longest, dupcounter); 
    end;

end;    % end main tree construction loop

%#### Find best action sequence, breaking ties in favor of the first found
ret = (discount .^ (0:longest-1)) * SR(1:longest, 1:Ns);
[retstar, istar] = max(ret);
lenstar = find(SS(1:longest, istar), 1, 'last'); 
useq = (unnormu(SU(1:lenstar, istar) + 0.5 .* ((1/3).^SS(1:lenstar, istar)), maxu))';
xseq = SX(:, 1:lenstar+1, istar);
if cfg.verb >= 3, viewsets(SU, SS, 1:Ns, 1, istar, []); pause; end;  % view selected sequence

% create stats if requested on output
if nargout >= 2,
    planinfo.n = n;
    planinfo.nsets = Ns;
    planinfo.longest = longest;
    planinfo.trun = cputime - t;
    planinfo.duplicates = dupcounter;
    planinfo.message = sprintf('Planning done, n=%d transitions, len_returned=%d, longest=%d, #sets=%d, trun=%.2f, duplicates=%d', ...
        planinfo.n, lenstar, planinfo.longest, planinfo.nsets, planinfo.trun, planinfo.duplicates); 
end;

% create plandata if requested on output
if nargout >= 3,
    plandata.SU = SU(1:longest, 1:Ns);
    plandata.SS = SS(1:longest, 1:Ns);
    plandata.SX = SX(:, 1:longest+1, 1:Ns);
    plandata.SR = SR(1:longest, 1:Ns);
    plandata.useq = useq;
    plandata.xseq = xseq;
end;

% construct the output based on calling mode
switch nargout,
    case 1, varargout = {useq(:, 1)};
    case 2, varargout = {useq(:, 1), planinfo};
    case 3, varargout = {useq(:, 1), planinfo, plandata};
    case 4, varargout = {useq, xseq, planinfo, plandata};
end;

end % main function


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local functions
% normalize / unnormalize actions
% function u = normu(u, maxu)   % currently not needed
% u = 0.5 + 0.5 * u / maxu;
% end
function u = unnormu(u, maxu)
u = (u - 0.5) * 2 * maxu;
end
function LIPSCHITZ = lipschitzmatrix(L,d)
LIPSCHITZ = zeros(d+1, d+1);
for i=1:d+1, LIPSCHITZ(i, 1:i) = L(1) .^ (i-1:-1:0); end;
LIPSCHITZ = L(2) .* LIPSCHITZ;
end

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Visualization
% sequences in the U space
function viewsets(SU, SS, indices, focus, reset, which)
longest = size(SS, 1);
if which(1) == 'a',
    if reset, figurex('h=10 size=[1100,800]'); end;
    figure(10);
    clf;
elseif which(1) == 'e'
    if reset, figurex('h=11 size=[800,600]'); end;
    figure(11);
    clf;
end;
for i = 1:length(indices),
   is = indices(i);
   len = find(SS(1:longest, is), 1, 'last'); 
   if any(is == focus), c = 'r'; else c = 'k'; end;
   for k = 1:len
       % plot subset
       plot(i-1 + (SU(k, is) + [0 (1/3)^SS(k, is)]), [k k], c, 'Marker', '.', 'MarkerFaceColor', c);
       hold on;
       scatter(i-1 + SU(k, is) + 0.5*(1/3)^SS(k, is), k, [], c, 'x');
   end;
   plot([i i], [0 longest], 'k:');    % seq separator
end;
end

% ranking of sequences by partial ordering / return
function viewranking(SS, Ns, ret, is, greater, expand)
greater = greater'; greater(is) = false;
figurex('h=20 size=[900,800]');
clf;
offset_x = (max(ret) - min(ret)) * .01;
offset_y = Ns * .01;
% generate labels
labels = cell(Ns, 1);
for i = 1:Ns, 
    len = find(SS(:, i), 1, 'last');
    labels{i} = num2str(SS(1:len, i)');
    labels{i}(labels{i} == ' ') = [];
end;
%
regular = true(Ns, 1);
regular = regular & ~greater; regular(is) = false;
scatter(ret(regular), find(regular), [], 'k', 'MarkerFaceColor', 'k');
hold on;
text(ret(regular) + offset_x, find(regular) + offset_y, {labels{regular}});
%
scatter(ret(greater), find(greater), [], 'b', 'MarkerFaceColor', 'b');
h = text(ret(greater) + offset_x, find(greater) + offset_y, {labels{greater}});
set(h, 'Color', 'b');
%
if expand(is), c = 'r'; else c = 'g'; end;
scatter(ret(is), is, [], c, 'MarkerFaceColor', c);
h = text(ret(is) + offset_x, is + offset_y, {labels{is}});
set(h, 'Color', c);
grid on;
xlabel('return / util'); ylabel('seq'); title('Blue: partially greater // Green: dominated // Red: undominated (expanded)');
end

% Old LP code
% the reward capping described in JF's thesis: (should be in set selection)
% make sure everything after the first rewardbound > 1 or uninitialized set is 1;
% this is equivalent so stopping at the last step k so that bound <= 1 (or at first
% uninitialized set) and then taking gamma^k/1-gamma    
% for is = 1:Ns, rewardbounds(find(rewardbounds(:, is) > 1 | SS(1:longest, is) == 0, 1, 'first'):end, is) = 1; end; 

% the set selection described in JF's thesis
% % note T in the paper notation = T - 1
% T = find(parentsr + rewarddiameters(1:len, is) <= 1, 1, 'last');
% if isempty(T), splitdims = 1;   % the first dimension is still too coarse, split that one
% else
%     % a [1 x T] vector containing the coefficients to which subsets 0, 1, ..., T-1 intervene
%     % in the bound
%     subsetcoeff = (discount .^ (0:T-1)) * LIPSCHITZ(1:T, 1:T);
%     % splitting a set at some dimension j in 3 would result in a decrease of the bound 
%     % equal to 2/3 * coeff * subsetsize
%     dimscores = 2/3 .* subsetcoeff .* subsetsizes(1:T, is)';
%     if T == len,    % rank the last dimension too
%         % old contribution of set T+1 was gamma^T * 1 (it was in gamma^T/1-gamma)
%         % new contribution is gamma^T * reward diameter with the usual formula, 
%         % where the new dimension is split once so it has a size of 0.5/3
%         dimscores(T+1) = discount.^T * (1 - LIPSCHITZ(T+1, 1:T+1) * [subsetsizes(1:T, is); 0.5/3]);
%     end;
%     % choose the dimension that maximizes the decrease in upper bound
%     [maxscore, splitdims] = max(dimscores);    
% end;

% % bug-check explicit computation for B-values in LP
% for is = 1:Ns,
%     bval = 0;
%     d = find(SS(1:longest, is), 1, 'last'); 
%     rewbound = zeros(1, d);
%     for k = 1:d,
%         sizes = 0.5 * (1/3).^SS(1:k, is);
%         rewbound(k) = min(SR(k, is) + sum( (cfg.lipschitz .^ (k:-1:1)) .* sizes'), 1);
%         bval = bval + discount^(k-1) * rewbound(k);
%     end;
%     bval = bval + discount^d / (1-discount);
%     if abs(bval - bvalues(is)) > 1e-10,
%         error('bval mismatch');
%     end;
% end;

% % old unvectorized code for ranking dimensions in LP
%             t2 = cputime;
%             dimscores = zeros(1, len+1);
%             for k = 1:len,
%                 newsubsetsizes = subsetsizes(1:len, is);
%                 % kth subset would be split in 3
%                 newsubsetsizes(k) = newsubsetsizes(k) / 3;
%                 % compute the new truncated) bound; note the capping by 1
%                 newrewbnd = SR(1:len, is) + LIPSCHITZ(1:len, 1:len) * newsubsetsizes;
%                 dimscores(k) = discounting * min(newrewbnd, 1);
%             end;
%             % we did not yet add the cap gamma^d/1-gamma
%             dimscores(1:len) = bvalues(is) - (dimscores(1:len) + discount^len / (1-discount));
%             % rank the next dimension: assuming the next dimension split and next reward is 0
%             newrewbnd = [SR(1:len, is); 0] + LIPSCHITZ(1:len+1, 1:len+1) * [subsetsizes(1:len, is); 0.5/3];
%             dimscores(k+1) = [discounting discount^len] * min(newrewbnd, 1) + discount^(len+1) / (1-discount);
%             dimscores(k+1) = bvalues(is) - dimscores(k+1);
%             T_DIMSEL = T_DIMSEL + [1 cputime-t2]; % 1st elem count, 2nd elem total