function  [uroot, planinfo, plandata] = uct(xroot, model, cfg)
% UCT Upper confidence bounds on trees

% Author: Lucian Busoniu
% Version history:
%   1.1, 5 July 2011: 
%       - fixed bug where action with largest UCB was selected 
%           (now providing either largest empirical return, or most-often selected action)
%       - tree data preinitialization

CFG.gamma = [];             % discount factor
CFG.n = [];                 % simulation budget
CFG.U = [];                 % discrete, flat action space
CFG.Cp = 1;                 % constant multiplying confidence term
CFG.actsel = 'ret';         % 'ret' for largest average return, 'count' for most-selected
CFG.depth = [];             % develop tree up to this depth, empty = default, i.e. log n / log(1/gamma)
CFG.k = -1;                 % discrete time step
% display config
CFG.verb = 0;               % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);
if isempty(cfg.depth), cfg.depth = ceil(log(cfg.n) / log(1/cfg.gamma)); end;

% quick-access
M = size(cfg.U, 2);      % number of discrete actions

% start counting time
t = cputime;

% init tree information (for speed, everything is represented flatly, as vectors)
% the maximum tree size is reached if we don't encounter any repeated states, i.e. each depth-long
% trajectory is unique
ntraj = ceil(cfg.n / cfg.depth);      % assuming continuing task and allowing an overrun of max 1 traj
maxtreesize = ntraj * cfg.depth + 1;  % extra 1 for root
Pt = zeros(1, maxtreesize, 'uint32');           % parent
Ct = zeros(ntraj, M, maxtreesize, 'uint32');    % indices of children for each action (first dimension = #children;
                                                % upper bound ntraj, if we choose same action at root)
Xt = zeros(model.p, maxtreesize,'double');      % state
Ht = zeros(1, maxtreesize, 'uint16');           % depth (uint16 as depth > 65536 considered unfeasible)
Tt = zeros(M, maxtreesize, 'double');           % counters for each action (kept double as they are used in math)
Rt = zeros(M, maxtreesize, 'double');           % return accumulator for each action
% init history of simulated trajectories
% preinit assuming continuing task i.e. ntraj trajectories -- otherwise they will have to grow, but
% there is no way to predict how many trajectories in that case
Xh = zeros(model.p, cfg.depth+1, ntraj); 
Uh = zeros(model.q, cfg.depth+1, ntraj); 
Rh = zeros(1, cfg.depth+1, ntraj);

if cfg.verb >= 1, disp(sprintf('Starting planning, n=%d', cfg.n)); end;

% init tree root
Pt(1, 1) = 0; 
Ct(:, :, 1) = 0; 
Xt(:, 1) = xroot;
Ht(1) = 0;
Rt(:, 1) = 0;      
Tt(:, 1) = 0; 
nt = 1;     % we have 1 node in the initial tree

% main simulation loop
tau = 1;                        % sequence/trajectory index
nsamp = 0;                      % total number of samples generated
iseq = zeros(cfg.depth+1, 1);   % sequence of node indices
jseq = zeros(cfg.depth+1, 1);   % sequence of action indices
UCB = zeros(M, 1);              % UCBs for the actions
while nsamp < cfg.n,
    
    % 1) Run a trajectory initializing nodes along the way
    it = 1;                     % current index in tree (root)
    Xh(:, 1, tau) = xroot;      % init trajectory
    h = 1; term = 0;
    while h <= cfg.depth && ~term,
        
        % 1a) Choose action and simulate
        % compute UCBs
        junsampled = Tt(:, it) == 0;
        if ~any(junsampled),
            UCB(:) = Rt(:, it) ./ Tt(:, it) + ...
                cfg.Cp .* sqrt(2 .* log(sum(Tt(:, it))) ./ Tt(:, it));
        else
            % save some computation, we have some inf bounds so the rest don't matter anyway
            UCB(junsampled) = Inf;
            UCB(~junsampled) = 0;
        end;
        
        % choose randomly an action with largest UCB and simulate it
        jmax = find(UCB == max(UCB)); jmax = jmax(unidrnd(length(jmax)));
        Uh(:, h, tau) = cfg.U(:, jmax);
        [Xh(:, h+1, tau), Rh(1, h+1, tau), term] = feval(model.fun, model, Xh(:, h, tau), Uh(:, h, tau));
        nsamp = nsamp + 1;
        
        % save indices
        iseq(h) = it; jseq(h) = jmax;
        
        % 1b) Go along tree or add new node
        % search for the next state in currently existing children
        % (NOTE this requires a finite number of possible next states for each transition)
        ic = Ct(:, jmax, it);                 % all children (including uninitialized)
        iilast = find(ic > 0, 1, 'last'); 
        if isempty(iilast), iilast = 0; end;  % last initialized index
        ic = ic(1:iilast);                    % only initialized children
        iixplus = findflat(Xh(:, h+1, tau), Xt(:, ic)); % search child with this state
        ixplus = Ct(iixplus, jmax, it);       % remember its index in the tree
        
        if ~isempty(ixplus),
            % we continue with the found node
            it = ixplus;
        else
            % new node must be added
            Pt(1, nt+1) = it;
            Ct(:, :, nt+1) = 0;             % node has no children
            Xt(:, nt+1) = Xh(:, h+1, tau);  % node is associated to next state
            Ht(nt+1) = h;                   % h is 1-based, so no need to add 1 even though we're at next level
            Rt(:, nt+1) = 0;                % return accumulator
            Tt(:, nt+1) = 0;
            % set child pointer in /current/ node
            Ct(iilast+1, jmax, it) = nt+1;
            % we continue with the newly initialized node
            it = nt + 1;
            % tree has grown
            nt = nt + 1;
        end;
        
        % increment depth / time step
        h = h + 1;
    end;
    % the trajectory was H steps long
    H = h - 1;
    % save final indices
    iseq(H+1) = it; jseq(H+1) = NaN;
        
    % 2) Update tree information as result of current trajectory
    
    % compute returns:
    discounting = cfg.gamma .^ (0:H-1);
    % ret_k = gamma^k r_k+1
    ret = discounting .* Rh(1, 2:H+1);
    % ret_k = sum_{j=k}^K gamma^j r_j+1
    ret = cumsum(ret(end:-1:1)); ret = ret(end:-1:1);
    % ret_k = sum_{j=k}^K gamma^{j-k} r_j+1 -- i.e., return starting at step k
    ret = ret ./ discounting;
    
    % update info of each node along the trajectory
    for h = 1:H,
        % increment sample counter
        Tt(jseq(h), iseq(h)) = Tt(jseq(h), iseq(h)) + 1;
        % add return
        Rt(jseq(h), iseq(h)) = Rt(jseq(h), iseq(h)) + ret(h); 
    end; 
    
    % log
    if cfg.verb >= 2, disp(sprintf('Trajectory tau=%d processed, n=%d out of %d', tau, nsamp, cfg.n)); end;
    if cfg.verb >= 3, visualizetree; pause; end;
    
    % increment traj index
    tau = tau + 1;
    
end;    % end main loop

% finally, return a first-level action according to selection criterion
junsampled = Tt(:, 1) == 0;
if any(junsampled), disp('UCT: Warning! Not all actions have been sampled at root'); end;
if cfg.actsel(1) == 'r',            % Return (largest empirical return)
    % don't rely on unsampled actions (shouldn't be any, warning is thrown above if any found)
    uvalues(junsampled) = 0;    
    uvalues(~junsampled) = Rt(~junsampled, 1) ./ Tt(~junsampled, 1); 
elseif cfg.actsel(1) == 'c',        % Count (most selected action)
    uvalues = Tt(:, 1);
else
    error('UCT: Unknown action selection method [%s]', cfg.actsel);
end;
jmax = find(uvalues == max(uvalues)); jmax = jmax(unidrnd(length(jmax)));
uroot = cfg.U(:, jmax);

% output stats if requested
if nargout >= 2,
    planinfo.n = nsamp;
    planinfo.tau = tau;
    planinfo.depth = max(Ht(:));
    planinfo.treesize = nt;
    planinfo.trun = cputime - t;
    if nargout >= 3,
        % output only initialized tree and trajectories
        plandata.Pt = Pt(1:nt); plandata.Ct = Ct(:, 1:nt);
        plandata.Xt = Xt(:, 1:nt); plandata.Ht = Ht(1:nt);
        plandata.Rt = Rt(:, 1:nt); plandata.Tt = Tt(:, 1:nt);
        plandata.Xh = Xh(:, :, 1:tau-1); plandata.Uh = Uh(:, :, 1:tau-1); plandata.Rh = Rh(1, :, 1:tau-1);
    end;
    planinfo.message = sprintf('Planning done, tau=%d traj, n=%d out of %d, treesize=%d, depth=%d', ...
        planinfo.tau, planinfo.n, cfg.n, planinfo.treesize, planinfo.depth); 
    if cfg.verb >= 1, disp(planinfo.message); end;
end;


    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Nested function to plot a graphical representation of the current tree
    function visualizetree
    persistent figh;
    figsize = [800,600];
    if isempty(figh), figh = figurex(struct('size', figsize)); else figure(figh); end;
    clf;

    % initialize helper variables for nested functions
    nodestoexpand = {};
    % build and customize tree
    if verLessThan('Matlab', '7.10'),   guitree = uitree('Root', rendernode(1));
    else                                guitree = uitree('v0', 'Root', rendernode(1));
    end;
    guitree.Position(3) = figsize(1);   % expand to figure width
    % expand nodes
    for i = 1:length(nodestoexpand), guitree.expand(nodestoexpand{i}); end;
    % done

        % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Nested function to render a node recursively
        function guinode = rendernode(i, jp)
        % create label
        if i == 1,  label = 'root: ';
        else        label = ['u=' num2strx(cfg.U(:, jp)', 'vector')];
        end;
        % find if this node is along last trajectory
        iontraj = find(i == iseq);
        % if yes show reward (note last node has no associated reward)
        if ~isempty(iontraj) && iontraj > 1, label = [label '  r=' num2str(Rh(1, iontraj, tau))]; end;
        % add state and return data
        label = [label '   #-->   x=' num2strx(Xt(:, i)', 'vector')];
        label = [label '   R/T=' num2strx(Rt(:, i)', 'vector') '/' num2strx(Tt(:, i)', 'vector')];
        % highlight if along traj
        if ~isempty(iontraj), label = ['<html><font color="blue">' label '</font></html>'];    % expanded
        end;
        leaf = ~any(any(Ct(:, :, i)));
        % add tree(i) to tree
        try
            if verLessThan('Matlab', '7.10'),   guinode = uitreenode(label, label, [], leaf);
            else                                guinode = uitreenode('v0', label, label, [], leaf);
            end;
        catch
            pause;
        end;
        if ~leaf,
            % remember we need to expand this node
            nodestoexpand{end+1} = guinode;
            for jj = 1:M,
                Lj = find(Ct(:, jj, i), 1, 'last');
                for ii = 1:Lj,
                    ic = Ct(ii, jj, i);
                    guinode.add(rendernode(ic, jj)); 
                end;
            end;
        end;
        end     % function rendernode
    end     % function visualizetree

end % main function




