function [theta, flag, delta] = rbfqiter(c, rad, model, cfg, theta0)
% Approximate Q-iteration with normalized radial basis functions approximation.
%   [THETA, FLAG, DELTA] = RBFQITER(C, RAD, MODEL, CFG, THETA0)
% This is not indended to be used as a standalone algorithm. See rbfqi for a wrapper.
%
% Parameters:
%   C       - RBF centers, p x N where p state variable size, N number of basis functions
%   RAD     - RBF radii, p x N where p state variable size, N number of basis functions
%   MODEL   - problem model
%   CFG     - configuration parameters. Requires fields (subject to change):
%       gamma, U, verb, N, M, qi_maxiter, qi_eps  - see cerbfqi for description
%   THETA0  - initial linear parameter vector. If not given, will start from identical 0
%
% Returns:
%   THETA   - computed parameter vector 
%   FLAG    - 1 if convergence, -1 if divergence, 0 if max iterations exhausted with on
%       convergence or divergence
%   DELTA   - the evolution of the infinity-norm difference over subsequent parameter vectors

% NOTE the name of this function has been changed 2007-05-31; only cerbfqi has been updated as a
% result, the test scripts etc. still call rbfqi; but that will generate errors so it can be
% fixed upon need.

BASEVERB = 4;
% shorthand dimension variables
N = cfg.N; M = cfg.M;

dispx('Computing MDP and activation data...', cfg.verb, BASEVERB+1);
% activation of each center
A = zeros(N, N);
% for each x and u, activation vector of f(x, u)
F = zeros(N * M, N);
% reward rho(x, u)
R = zeros(N, M);

% iterate over (x, u) in (centers, discrete actions) and populate A, F, R structures
for i = 1:N,
    x = c(:, i);
    % compute activation vector
    A(i, :) = nrbf(x, N, c, rad);    
    for j = 1:M,
        % apply f, rho, and store results
        [xplus R(i, j)] = feval(model.fun, model, x, cfg.U(:, j));
        F(i+(j-1)*N, :) = nrbf(xplus, N, c, rad);
    end;            % FOR over action discretization
end;                % FOR over state samples

% precompute the inverse of A
Ainv = inv(A);

dispx('Performing linear Q-value iteration...', cfg.verb, BASEVERB);

% initialize linear parameter matrix
if nargin < 5,  theta = zeros(N, M);
else            theta = theta0;
end;

k = 1;
delta = nan(1, cfg.qi_maxiter);
conv = 0; div = 0;
while ~conv && ~div && k <= cfg.qi_maxiter,       % main loop

    thetaold = theta;
    % apply least-squares update
    theta = Ainv * (R + cfg.gamma * reshape(max(F * theta, [], 2), N, M));
    
    % compute infty-norm of vector representation of theta difference
    delta(k) = max(max(abs(theta - thetaold)));        
    conv = delta(k) <= cfg.qi_eps;
    div = k > 10 && all(diff(delta(k-10:k)) > 0);

    k = k + 1;
end;        % while not converged and allowed more iterations

if conv,	flag = 0; dispx(['Convergence detected after ' num2str(k-1) ' iterations. Algorithm stopped.'], cfg.verb, BASEVERB);
elseif div, flag =-1; dispx(['Divergence detected after ' num2str(k-1) ' iterations. Algorithm stopped.'], cfg.verb, BASEVERB);
else        flag = 1; dispx(['qi_maxiter=' num2str(cfg.qi_maxiter) ' exhausted. Algorithm stopped'], cfg.verb, BASEVERB);
end;

end
% END rbfqi() -------------------------------------------------------------

% Removed code for:
%     % visual feedback of algorithm progress
%     if ~mod(k, cfg.qi_iterdisp),
%         dispx(['qi_k=' num2str(k) ' iteration completed, delta=' num2str(delta)], cfg.verb, BASEVERB+2);
%     end;

% Alternate update code
%     % "trick" to ensure convergence: pure assignment, no least-squares
%     theta = R + cfg.gamma * reshape(max(F * theta, [], 2), N, M);
