function [sys,x0,str,ts] = bike_sfun(t,x,u,flag)
% S-function for bike model
% No bounds are enforced
switch flag,

  %%%%%%%%%%%%%%%%%%
  % Initialization %
  %%%%%%%%%%%%%%%%%%
  case 0,
    % make a deterministic, oriented model
    model = bike_problem('model', 1, 1);
    % change it into a riding model
    model = bike_problem('riding', model);
    
    % define initial state
    model.x0 = [10*pi/180; 0; 0; 0; pi/4; 0; 0];
    
    % save in base workspace
    assignin('base', 'BIKE_MODEL', model);
    disp('BIKE_MODEL initialized');
    [sys,x0,str,ts]=mdlInitializeSizes(model);

  %%%%%%%%%%%%%%%
  % Derivatives %
  %%%%%%%%%%%%%%%
  case 1,
    model = evalin('base', 'BIKE_MODEL');
    sys=mdlDerivatives(t,x,u,model);

  %%%%%%%%%%%
  % Outputs %
  %%%%%%%%%%%
  case 3,
    model = evalin('base', 'BIKE_MODEL');
    sys=mdlOutputs(t,x,u,model);

  %%%%%%%%%%%%%%%%%%%
  % Unhandled flags %
  %%%%%%%%%%%%%%%%%%%
  case { 2, 4, 9 },
    sys = [];

  %%%%%%%%%%%%%%%%%%%%
  % Unexpected flags %
  %%%%%%%%%%%%%%%%%%%%
  otherwise
    error(['Unhandled flag = ',num2str(flag)]);

end
% end csfunc

%
%=============================================================================
% mdlInitializeSizes
% Return the sizes, initial conditions, and sample times for the S-function.
%=============================================================================
%
function [sys,x0,str,ts]=mdlInitializeSizes(m)

sizes = simsizes;
sizes.NumContStates  = m.p;
sizes.NumDiscStates  = 0;
sizes.NumOutputs     = m.p;
sizes.NumInputs      = m.q;
sizes.DirFeedthrough = 0;
sizes.NumSampleTimes = 1;

sys = simsizes(sizes);
x0  = m.x0;
str = [];
ts  = [0 0];

% end mdlInitializeSizes
%
%=============================================================================
% mdlDerivatives
% Return the derivatives for the continuous states.
%=============================================================================
%
function xdot=mdlDerivatives(t,x,u,m)

omega   = x(1);
domega  = x(2);
theta   = x(3);
dtheta  = x(4);
psi     = x(5);

% preinit xdot
xdot = x;

% (1) = d
if m.det,   % deterministic dynamics
    phi = omega + atan(u(1)) / m.h;
else        % noisy, noise w added to u(1)=d with w~uniform(-maxw, maxw)
    phi = omega + atan(u(1) + (2*rand - 1)*m.maxw) / m.h;        
end;
invrb = abs(tan(theta)) / m.l;
if theta == 0,      invrCM = 0;
else                invrCM = 1 / sqrt(m.lminuscsquared + (m.l/invrb)^2);
end;
% omega /omegadot dynamics
% if omega <= m.maxomega, % within bounds
    xdot(1) = domega;
    xdot(2) = 1 / m.Ibc * ( m.Mgh * sin(phi) ...
        - cos(phi) * ( m.Idcsigmadot * dtheta + ...
                       sign(theta)* m.vsquared * (m.Mdr * (abs(sin(theta)) / m.l + invrb) + m.Mh*invrCM) ) );
% else                    % outside bounds
%     xdot(1) = 0;
%     xdot(2) = 0;
% end;

% theta / dtheta dynamics
% u(2) = T, torque on the handlebars
% if abs(theta) <= m.maxtheta,
    xdot(3) = dtheta;
    xdot(4) = (u(2) - m.Idvsigmadot * domega) / m.Idl;
% else    % theta+ hit the bounds, its derivative and the speed derivative are 0
%     xdot(3) = 0;
%     xdot(4) = 0;
% end;

% psi dynamics, wrapping in the interval [-pi, pi)
xdot(5) = sign(theta) * m.v * invrb;
% position dynamics
xdot(6) = m.v * cos(psi); 
xdot(7) = m.v * sin(psi); 

% end mdlDerivatives
%
%=============================================================================
% mdlOutputs
% Return the block outputs.
%=============================================================================
%
function y=mdlOutputs(t,x,u,m)

y = x;

% end mdlOutputs
