function varargout = ipsetup_rtproblem(mode, sys, varargin)
% Interact with the real-time Inverted Pendulum Setup.
%   OUT = IPSETUP_RTPROBLEM(MODE, SYS, VARARGIN)
% This function conforms to the specifications established by SAMPLE_RTPROBLEM (TODO create this function)
%
% Calling modes:
% SYS = IPSETUP_RTPROBLEM('system', SYSCFG)
% Create system structure. 
%
% TODO document other modes

% Requirements on the call sequence and its timing:
%   - 'init' must be called before anything else
%
%   - 'reset' is called at the beginning of each new trajectory, 
%           immediately followed by 'writeu'
%   - thereafter, 'readx' immediately followed by 'writeu' are called exactly once every sample time
%
%   - 'close' is called before stopping the algorithm
%
% 'idle' should be called if the system should wait for longer than 1 sample time

% remember the old angle values, in order to estimate velocities
% note angle values are not wrapped, so that velocity calculations by discrete diff are correct
persistent angold resetcount

varargout = {};
switch mode,
    
    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Time-critical operations (come first to avoid "case" overhead)
    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    
    % read angles, estimate angular velocities
    % this is a time-critical operation
    case 'readx',
        % read angle
        DATA = mops('Read', sys.H); ang = DATA(3);
        if ~isempty(angold),    
            % wrap angle, bound velocity (computed by discrete difference)
            x = [mod(ang, 2*pi)-pi; min(max((ang-angold)./sys.Ts, -sys.maxx(2)), sys.maxx(2))]; 
        else
            % wrap angle, velocity is zero (i.e., unknown)
            x = [mod(ang, 2*pi)-pi; 0];
        end;
        angold = ang;
        varargout = {x};      
        
    % write actions to actuators
    % this is a time-critical operation
    case 'writeu',
        u = varargin{1};
        % saturate control action
        u = min(sys.maxu, max(-sys.maxu, u));
        % send control input to process
        mops('Write', sys.H, 0, 1, u, 0);   
        % output the actual command values of the considered joints (unchanged for this system)
        varargout = {u};
                
    % set the system to idle
    case 'idle',
        % set actuator to zero
        mops('Write', sys.H, 0, 1, 0, 0);
    
    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Non-time-critical operations
    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % create system structure
    case 'system',    
        % save sys config variable; 'sys' will get overwritten below
        if nargin >= 2, syscfg = sys; 
        else syscfg = struct;
        end;
        
        % 1) fixed fields, e.g., given by the hardware
        sys = struct;
        sys.id = 'ipsetup'; 
        sys.problem = @ipsetup_rtproblem;
        sys.det = [];   % not sure yet if it's deterministic or not (there will be noise...)
        sys.fun = @ipsetup_evaltrans;
        sys.p = 2;
        sys.q = 1;

        % 2) configurable fields and options: SYSCFG contains the defaults
        % bounds, sampling time
        SYSCFG.maxx = [pi; 15*pi];
        SYSCFG.maxu = 3;
        SYSCFG.Ts = .05;       % sampling time
        % reward config
        SYSCFG.rewtype = 'lqr';% reward type: 'lqr'
        % params for 'lqr'
        SYSCFG.Qrew = diag([5, 0.1]);   % state weights in the reward
        SYSCFG.Rrew = 1;                % action weights in the reward
        % process the config, copy it to the system structure
        syscfg = parseconfig(syscfg, SYSCFG);
        sys = copyfields(syscfg, sys);
               
        % reward config, derived fields
        switch sys.rewtype,
            case 'lqr',
                sys.maxr = sys.maxx' * sys.Qrew * sys.maxx + sys.Rrew * sys.maxu * sys.maxu;
            otherwise,
                error('Reward type [%s] unknown', sys.rewtype);
        end;
        
        % prepare output
        varargout = {sys};
        
    % initialize hardware system
    case 'init',
        % initialize DCMOTOR - USB init
        mops('CloseAll');                % close port
        H = mops('Open', 'mops1');       % open port
        H.WatchdogTimeout = 10;          % watchdog timeout
        mops('SetParams', H);            % set the parameter
        mops('Write', H, 1, 1, 0, 0);    % dummy write to sync interface board
        mops('Write', H, 1, 1, 0, 0);    % reset position
        mops('Write', H, 0, 1, 0, 0);    % end reset
        for i = 1 : 100,                 % read a couple of times
            mops('Read',H);              % dummy read sensor data
        end;        
        % place handle on system structure
        sys.H = H;
        % reset persistent variables
        angold = []; resetcount = 0;
        varargout = {sys};
       
        
    %  finalize and close connection with hardware
    case 'close', 
        % initialize DCMOTOR - USB init
        mops('CloseAll');                
        H = mops('Open', 'mops1');
        mops('Write', H, 0, 1, 0, 0.0);
        mops('CloseAll');
        % output system
        sys.H = [];     % there is no handle
        varargout = {sys};
        
    % reset system to an initial state
    case 'reset',
        reset = varargin{1};
        % support calling using only reset type
        if ischar(reset) && any(strcmp(reset, {'randpos'})),
            reset = struct('type', reset);
        end;
        
        % reset config
        RESET.type = 'randpos';
        RESET.rand_maxu = sys.maxu; % use different (usually larger) u bound for rand
        RESET.rand_time = 1;        % for how long to apply random actions (in seconds)
        RESET.rand_keep = 0.5;      % for how long to keep random action constant
        reset = parseconfig(reset, RESET);
        
        % the modes in the outer switch work at trial start or for replay
        switch reset.type,
            case 'rand',        % apply some random actions
                x0 = ipsetup_randcontrol(sys, reset);
            case 'down',        % stabilize pointing down
                x0 = ipsetup_pointdown(sys, reset);   
            case 'up',          % stabilize pointing up
                x0 = ipsetup_swingup(sys, reset);
            case 'randdown',    % half downward, half random
                if ~mod(resetcount, 2), x0 = ipsetup_pointdown(sys, reset);
                else                    x0 = ipsetup_randcontrol(sys, reset);
                end;
                resetcount = resetcount + 1;
            otherwise,
                error('Reset method [%s] unknown.', reset.type);
        end;
        % save (nonnormalized) angle value
        angold = x0(1);
        
        % normalize and output x0
        x0 = [mod(x0(1), 2*pi)-pi; min(max(x0(2), -sys.maxx(2)), sys.maxx(2))]; 
        varargout = {x0};
        
end;    % main 'mode' switch

end




% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local functions
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% Evaluate ipsetup transition
% Compute reward and terminal flag
function [rplus, term] = ipsetup_evaltrans(sys, x, u, xplus)
    % call common reward function (for setup and RT system)
    rplus = ipsetup_reward(sys, x, u, xplus);   
    % task is continuing (not episodic)
    term = 0;   
end


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% Apply some random control actions
function xfinal = ipsetup_randcontrol(sys, reset)
    oldang = []; 
    x = NaN(2, 1); 
    tic;
    for k = 1:reset.rand_time / sys.Ts,                            
        data = mops('Read', sys.H);
        x(1) = data(3); 
        if k > 1, x(2) = (x(1) - oldang) / sys.Ts; 
        else x(2) = 0;
        end;
        oldang = x(1);
        % once every rand_keep seconds, change the action to be applied
        % note we use a rand_maxu instead of sys.maxu as action bound
        if ~mod((k-1)*sys.Ts, reset.rand_keep), 
            u = (2*rand - 1) * reset.rand_maxu; 
            % disp(sprintf('k=%d, new action', k));
        end;
        mops('Write', sys.H, 0, 1, u, 0);
        while toc < sys.Ts, end;       
        tic;                       
    end;
	mops('Write', sys.H, 0, 1, 0, 0);   % zero actuator
    xfinal = x;
end         % ipsetup_randcontrol()


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% Stabilize pointing downward ('reset' input is ignored)
function xfinal = ipsetup_pointdown(sys, reset)
    T = 3; kstable = 20; epsstable = .001; PD = [25 1];
    X = NaN(2, ceil(T/sys.Ts));     % raw
    Xn = NaN(2, ceil(T/sys.Ts));    % normalized
    k = 1; stable = 0;
    tic; 	% reset Matlab's tic-toc timer
    while ~stable && k < T/sys.Ts,
        DATA = mops('Read', sys.H);        % read sensor data
        % current angle, estimate velocity
        X(1, k) = DATA(3);
        if k > 2,  X(2, k) = (X(1, k) - X(1, k-1)) / sys.Ts;
        else       X(2, k) = 0;
        end;
        % wrap the angle so we don't care about # full rotations
        Xn(1, k) = mod(X(1, k) + pi, 2*pi) - pi; 
        % compute error, derror, control action
        err = 0 - Xn(1, k); 
        if k > 2,  derr = (err - (0 - Xn(1, k-1))) / sys.Ts;
        else       derr = 0;
        end;
        u = PD(1) * err + PD(2) * derr;         % PD control law
        % bound control action and send to process
        mops('Write', sys.H, 0, 1, min(max(u, -sys.maxu), sys.maxu), 0);
        % check if we are stable
        stable = k > kstable && all(abs(X(2, k-kstable:k)) < epsstable);
        k = k + 1;
        while toc < sys.Ts, end;                % synchronize with real time
        tic;                                % reset Matlab's tic-toc timer
    end;
    mops('Write', sys.H, 0, 1, 0, 0);   % zero actuator
    xfinal = X(:, k-1);
    % disp(sprintf('Stabilized in k=%d steps', k-1));
    % H = ipsetup_reset;          % reset the setup / sensor
end         % ipsetup_pointdown()


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% Swing up using a heuristic policy (which btw works quite well -- thanks to Maarten)
function xfinal = ipsetup_swingup(sys, reset)
    T = 3; kstable = 20; epsstable = .001; PD = [10 1];
    X = NaN(2, ceil(T/sys.Ts));     % raw
    Xn = NaN(2, ceil(T/sys.Ts));    % normalized
    k = 1; stable = 0;
    tic;
    while ~stable && k < T/sys.Ts,
        DATA = mops('Read', sys.H); X(1, k) = DATA(3);
        % velocity = dangle/dt
        if k > 2,  X(2, k) = (X(1, k) - X(1, k-1)) / sys.Ts;
        else       X(2, k) = 0;
        end;
        % normalized state (note we don't care about the velocity)
        Xn(1, k) = mod(X(1, k), 2*pi) - pi;
        if abs(Xn(1, k)) > pi/2,    
            if X(2, k) >= 0, u = sys.maxu; else u = -sys.maxu; end;
        else
            u = - PD(1) * Xn(1, k) - PD(2) * X(2, k);
        end;
        % bound control action and send to process
        mops('Write', sys.H, 0, 1, min(max(u, -sys.maxu), sys.maxu), 0);
        % check if we are stable
        stable = k > kstable && all(abs(X(2, k-kstable:k)) < epsstable);
        k = k + 1;
        while toc < sys.Ts, end;                % synchronize with real time
        tic;                                % reset Matlab's tic-toc timer
    end;
    % disp(sprintf('Stabilized in k=%d steps', k-1));
    mops('Write', sys.H, 0, 1, 0, 0);   % zero actuator
    xfinal = X(:, k - 1);
end         % ipsetup_swingup()



