% Example control loop implementation for the inverted pendulum

p = 2;              % order of the system
Tend = 5;           % time to run
Ts = 0.005;         % sampling time
maxu = 3;           % maximum allowed control 
maxomega = 15*pi;   % maximum angular velocity

K = Tend/Ts;        % # of steps to run

% state "normalization" function
% 0 angle is originally stable equilibrium; we change angles so that it's the unstable
% equilibrium, and wrap all the angles into [-pi,pi)
% also bound the angular velocity
normalizex = @(x, maxomega) [mod(x(1, :), 2*pi) - pi; min(max(x(2, :), -maxomega), maxomega)];

% ===================================
% Initialize DCMOTOR - USB init
% sensors structure: [status, time, angle, velocity, current, voltage, dangle/dt, ???]
mops('CloseAll');                % close port
H = mops('Open', 'mops1');       % open port
H.WatchdogTimeout = 10;          % watchdog timeout
mops('SetParams', H);            % set the parameter
mops('Write', H, 1, 1, 0, 0);    % dummy write to sync interface board
mops('Write', H, 1, 1, 0, 0);    % reset position
mops('Write', H, 0, 1, 0, 0);    % end reset
for i = 1 : 100,                 % read a couple of times
    mops('Read',H);              % dummy read sensor data
end;

% History variables
t = 0:Ts:K*Ts;
DATA = zeros(8, K+1);       % data history
X = zeros(p, K+1);          % initialize state history
U = zeros(1, K);            % initialize input history

Xinit = [0;0];

% ============================
% MAIN CONTROL LOOP
tic; 	% reset Matlab's tic-toc timer
for k = 1:K,                            
    DATA(:, k) = mops('Read',H);        % read sensor data
    % current angle
    X(1, k) = DATA(3, k);
    % velocity = dangle/dt
    if k > 2,   X(2, k) = (DATA(3,k) - DATA(3,k-1)) / Ts; %DATA(2,k);
    else        X(2, k) = (DATA(3,k) - Xinit(2)) / Ts; %DATA(2,k);    
    end;
        
    % heuristic swing-up control law
    xn = normalizex(X(:, k), maxomega);
    if abs(xn(1)) > pi/2,   % destabilizing control
        U(k) = maxu * (sign(xn(2)) >= 0);
    else                    % stabilizing control -- PD
        U(k) = -10 * xn(1) -0.3 * xn(2);
    end;
	
    U(k) = sat(U(k), -maxu, maxu);      % saturate control action
    mops('Write',H,0,1,U(k),0);         % send control input to process

    while toc < Ts, end;                % synchronize with real time
    tic;                                % reset Matlab's tic-toc timer
end;

% measure last state
DATA(:, k+1) = mops('Read',H);        % read sensor data
X(1, k+1) = DATA(3, k+1);
X(2, k+1) = (DATA(3,k+1) - DATA(3,k)) / Ts; %DATA(2,k+1);

% give a small push to move away from unstable eqn (which is actually stable due to static
% friction)
mops('Write',H,0,1,2,0); tic; while toc < 0.2, end; mops('Write',H,0,1,0,0); 
% stop motor -- reset actuators, close the port
mops('Write', H, 0, 1, 0, 0.0);
mops('CloseAll');

% normalize process state: shift angle by pi, wrap into -pi,pi
XN = normalizex(X, maxomega);

% plot stuff
figh = figure; clf;
subplot(311);
plot(t, XN(1, :), 'k');
legend('\alpha');
subplot(312);
plot(t, XN(2, :), 'k');
legend('\alpha''');
subplot(313);
stairs(t(1:end-1), U, 'k');
legend('u');
xlabel('t');
