function varargout = machrepl_mdp(m, x, u, xplus)
% Implements the MDP of the machine replacement problem.
%   [OUT1, OUT2, ...] = MACHREPL_MDP(MODEL, X, U, [XPLUS])
% 
% The input parameters can be:
%   MODEL   - model
%   X       - current state
%   U       - applied action
%   XPLUS   - next state
% Not all calling modes take all parameters, and the output arguments
% depend on calling mode.
% 
% Calling modes:
%   [PPLUS, RPLUS, TERM] = MACHREPL_MDP(MODEL, X, U, XPLUS)
%   Transition probability and reward for an (X, U, XPLUS) tuple. Returns
%   the probability ftilde(X, U, XPLUS), and the reward rhotilde(X, U,
%   XPLUS) from the MDP description. TERM is an optional argument.
%
%   [XPLUS, RPLUS, TERM] = MACHREPL_MDP(MODEL, X, U)
%   Regular, stochastic transition. Returns a sample next state XPLUS and the
%   corresponding reward RPLUS, together with a terminal flag TERM. TERM is
%   an optional argument.
%
%   [XPLUS, RPLUS, TERM, PPLUS] = MACHREPL_MDP(MODEL, X, U)
%   Full set ot stochastic transitions. XPLUS is a matrix of possible next
%   states, one per row; RPLUS a vector of corresponding rewards; TERM a
%   vector of termination flags for each next state; and PPLUS a vector of
%   probabilities of reaching each of these states. TERM is an optional
%   argument.
%
% See also: machrepl_problem

% determine calling mode
if nargin == 3 && nargout <= 3,     % regular, stochastic transition
    if u == 2,  % replace  
        xplus = 1; rplus = -m.rew_c + m.rew_r(1);       % replace
    else        % wait
        cdf = cumsum(m.prob(x, :)); cdf = [0 cdf(1:end-1)];
        xplus = find(rand >= cdf, 1, 'last');
        rplus = m.rew_r(x);
    end;
	terminal = 0;
    varargout = {xplus, rplus, terminal};
elseif nargin == 3 && nargout == 4, % full set of outcomes for an x,u pair
    if u == 2,  % replace 
        xplus = 1; pplus = 1; rplus = -m.rew_c + m.rew_r(1);
    else        % wait
        xplus = x:m.n; 
        pplus = m.prob(x, x:m.n); 
        rplus = repmat(m.rew_r(x), 1, length(xplus));
    end;
    terminal = zeros(1, length(xplus));
    varargout = {xplus, rplus, terminal, pplus};
elseif nargin == 4,                 % proba and reward for an x, u, x' pair
    if u == 2,  % replace 
        if xplus == 1, pplus = 1; rplus = -m.rew_c + m.rew_r(1);
        else pplus = 0; rplus = 0;
        end;
    else        % wait
        pplus = m.prob(x, xplus); 
        rplus = m.rew_r(x);
    end;
    terminal = 0;
    varargout = {pplus, rplus, terminal};
end;

end     % outcome local function
