function out = machrepl_problem(mode, varargin)
% Machine replacement problem.
%   OUT = MACHREPL_PROBLEM(MODE, [PARAM1, [PARAM2, ...]])
%
% See the lectures for a complete description of the problem, or reference
% [1, Section 1.1]. The state signal is a usage state in the set 1,...n, and the actions
% are 1, corresponding to using the machine in the current state, and 2,
% corresponding to replacing the machine.
%
% Calling modes:
%   MODEL = MACHREPL_PROBLEM('model', CFG)
%   Creates the model
%   - CFG specifies the model config. The following fields are important (all are optional):
%         n         - number of states
%         rew_r     - revenue produced in each state, 1xn vector
%         rew_c     - replacement cost, scalar
%         prob      - transition probas for action "wait", nxn matrix
%
%   - MODEL is the model structure returned. Representative fields:
%         p         - number of states
%         q         - numer of inputs (controls)
%         fun       - the mdp function of the problem
%         visualizefun - visualization function of the problem
%         X         - discrete state space, cell array with grids 
%               for each state variable (in our case, there's just one)  
%         U         - discrete action space, cell array with vector of
%               values for each variable
%       The following can be configured:
%           size    - the grid size, a vector [width, height]
%           x_goal  - the location of the goal, a column vector containing the 2 coordinates
%           x_obst  - the locations of obstacles, one per column, each in the same format as the goal
%           rew_step, rew_obst, rew_goal
%                   - the rewards obtained on, respectively, making a regular step, hitting an
%                   obstacle, and getting to the goal
%   
%   X0 = MACHREPL_PROBLEM('model', MODEL, RESET)
%   Resets the system to some initial state.
%   - RESET can be 'rand' for random reset to a nonterminal state, or a 
%       matrix containing reset states in the columns; the choice between
%       these reset states will be random in this case (if there's just one
%       column, deterministic reset to that state)
%   - X0 is the resulting state
% 
%   INFO = MACHREPL_PROBLEM('info')
%   Information about the model. 
%   - INFO representative fields:
%           id      - problem identifier
%           problem - handle to model function
%           det 	- whether problem is deterministic or stochastic
%           p       - # of state variables
%           q       -  # of control variables
%
%  The state is the 2-D position on the grid, and the action ranges in {1, 2, 3, 4}, signifying
%  respectively, {move left, right, down, up}.
% 
% [1] D. Bertsekas, Dynamic Programming and Optimal Control, vol 2 4th
% edition, Athena Scientific.

switch mode,

    % Offer basic info about the model
    case 'info',    
        info.id = 'machrepl';               % an identifier
        info.problem = @machrepl_problem;   % handle to model function
        info.det = 0;                       % stochastic
        info.p = 1;                         % # of state variables
        info.q = 1;                         % # of control variables
        out = info;
        
    case 'shortid', out = 'machrepl';

    % create the model
    case 'model',
        % these are set in stone
        m.id = 'machrepl';               % an identifier
        m.problem = 'machrepl_problem';  % handle to model function
        m.det = 0;                       % stochastic
        m.p = 1;                         % # of state variables
        m.q = 1;                         % # of control variables
        m.nw = 0;                        % number of noise variables
        m.Ts = 1;                        % no sample time per se, just a discrete time index
        m.fun = 'machrepl_mdp';          % MDP function
        % visualization function
        m.visualizefun = 'machrepl_visualize';
        m.U = {[1 2]};                   % wait, replace: this is set in stone
        
        % configurable properties
        CFG.gamma = 0.5;
        CFG.n = 5;                        % number of states
        CFG.prob = [];                    % transition probas, nxn matrix
        CFG.rew_r = [];                   % revenue produced in each state, 1xn vector
        CFG.rew_c = 1;                    % replacement cost
        
        if nargin >= 2, cfg = varargin{1};
        else cfg = struct; 
        end;
        cfg = parseconfig(cfg, CFG);
        
        m = copyfields(cfg, m);

        % explicit state space
        m.X = {1:m.n};

        % automatic creation of default reward, proba structure if not given on config
        if isempty(m.rew_r), for i = 1:m.n, m.rew_r(i) = max(.1, 1 - (i-1)*.1); end; end;
        if isempty(m.prob),
            m.prob = zeros(m.n, m.n);
            for i = 1:m.n,
                if i == m.n,        m.prob(i, i) = 1;
                elseif i == m.n-1,  m.prob(i, i:i+1) = [.7 .3];
                else                m.prob(i, i:i+2) = [.6 .3 .1];
                end;
            end;
        end;
            
        % determine minima and maxima
        m.minx = min(m.X{1}); m.maxx = max(m.X{1});
        m.minu = min(m.U{1}); m.maxu = max(m.U{1});
        
        out = m;
        
    % reset to initial state
    case 'reset',
        m = varargin{1};
        reset = varargin{2};
        if isempty(reset), reset = 'rand'; end;
        if isnumeric(reset),
            if size(reset, 2) > 1,      x0 = reset(:, unidrnd(size(reset, 2)));
            else                        x0 = reset;
            end;
        elseif strcmp(reset, 'rand'),
            x0 = unidrnd(m.n);
        end;

        out = x0;

end;        % mode SWITCH

% END sample_problem() RETURNING out ===================================================