function varargout = machrepl_visualize(cfg)
% Visualization function for the machine replacement problem.
%   [FIGH, GVIEW] = MACHREPL_VISUALIZE(CFG)
% Many modes are implemented to visualize various objects and the progress of DP and RL
% algorithms.
% 
% Q-function visualization:
%   CFG.model; should contain the model
%   CFG.Q should contain the Q-function, as an n x 2 matrix with n the
%   number of states
%
% Policy visualization:
%   CFG.model should contain the model
%   CFG.h should contain the policy, as an n x 1 vector with n the
%   number of states
%
% The two modes above can be combined to show both Q and h at the same
% time.
%
% State visualization:
%   CFG.model should contain the model
%   CFG.x should contain the state
%
% Other modes should not be used explicitly, they are called internally by
% various algorithms.

% TODO visualizing from data file: not tested
% output to sequence of figures: not tested

% default arguments
if nargin < 1, cfg = ''; end;

% where from to load the data, one of:
%       'caller'    - if the variables should be taken from the calling function
%       filename    - if data should be loaded from a file with name <filename>
CFG.datasource = 'caller';
CFG.gview = [];           % if nonempty: update existing; if empty: create new; graphical view
% CFG.datafile   = '';    % data source can be given as a data file here as well
% update/create view with:
CFG.model = []; CFG.x = []; CFG.Q = []; CFG.h = [];
CFG.qiter = 0;            
CFG.qevaliter = 0;
CFG.piter = 0;
% CFG.mpcstep = 0;
% CFG.mpctrial = 0;
CFG.qlearnstep = 0;             
% CFG.qlearnerstep = 0;
CFG.qlearntrial = 0;
CFG.sarsastep = 0;   
CFG.sarsaerstep = 0;
CFG.sarsatrial = 0;
CFG.mcstep = 0;
CFG.mctrial = 0;

CFG.finalize = 0;       % special option to finalize the output (namely, movie)
% which iter/trial/step to plot
CFG.ell = [];
CFG.tau = [];
% CFG.trial = [];
% CFG.k = [];
% CFG.ertrial = [];
% CFG.erk = [];
% CFG.eri = [];
% target Q-function/policy for performance plots
CFG.Qstar = [];
CFG.hstar = [];
CFG.Qhstar = [];

CFG = setfigprop(CFG, 'addfields');  % add figure settings fields to CFG
% output configuration
CFG.pause = .1;
CFG.consoleout = 0;
CFG.plottarget = 'screen';      % 'latex', 'beamer', 'screen', 'movie'
CFG.snapshottarget = 'screen';  % 'latex', 'beamer'
CFG.savedir = ''; 
CFG.savefig = 'machrepl';       % filename for saving figure
% if output is "movie":
CFG.compression = 'Cinepak';    % quality with Cinepak is pretty crappy...
CFG.quality = 100;
CFG.fps = 1;
CFG.snapshot = [];              % still snapshot config
% snapshot has the form [trial], [outer_iter], [trial step], or [outer_iter inner_iter]
% [outer_iter] fires a snapshot for qiter and piter when ell = outer_iter
% [trial] fires a snapshot for qlearn, sarsa when trial = trial
% [outer_iter inner_iter] fires a snapshot for qevaliter ell = outer_iter, tau = inner_iter
% [trial step] fires a snapshot for qlearn, sarsa when trial = trial, k = step

% process config
cfg = parseconfig(cfg, CFG);

% needed vars
vars = {'model'};
if cfg.qiter, 
    vars = {vars{:}, 'cfg', 'Qh'};
elseif cfg.qevaliter,
    vars = {vars{:}, 'cfg', 'Qhh', 'hh'};
elseif cfg.piter % || cfg.mctrial,
    vars = {vars{:}, 'cfg', 'Qh', 'hh'};
% elseif cfg.mpcstep || cfg.mpctrial,
%     vars = {vars{:}, 'X', 'U', 'R'};
elseif cfg.mcstep,
    vars = {vars{:}, 'cfg', 'Qh', 'Q', 'X', 'U', 'R'};
elseif cfg.qlearnstep || cfg.sarsastep % || cfg.qlearnerstep || cfg.sarsaerstep,
    vars = {vars{:}, 'cfg', 'Qh', 'Q', 'E', 'X', 'U', 'R'};
elseif cfg.qlearntrial || cfg.sarsatrial,
    vars = {vars{:}, 'cfg', 'Qh', 'deltah'};
elseif ~isempty(cfg.Q) || ~isempty(cfg.h) || ~isempty(cfg.x), 
    cfg.datasource = 'cfg';
end;

if (cfg.qlearnstep || cfg.sarsastep)... % || cfg.mpcstep || cfg.qlearnerstep || cfg.sarsaerstep) ...
        && ~strcmp(cfg.datasource, 'caller'),
    error('CLEANROB_PLOT: [***step] updates can only be performed while algorithm is running');
end;

% retrieve needed variables from the data source
cfg1 = cfg;
switch cfg.datasource,
    case 'caller',              % load needed vars from caller space
        for i=1:length(vars), 
            eval(sprintf('%s=evalin(''caller'', ''%s'');', vars{i}, vars{i})); 
        end;
    case 'cfg',
        structtovars(cfg, vars);
    otherwise,
%         load(cfg.datasource, vars{:});
end;
expcfg = cfg; cfg = cfg1;

% graphical style configuration
sty = struct;
% colors / linestyles
sty.title = {'HorizontalAlignment', 'left', 'FontSize', 12, 'FontWeight', 'bold'};
sty.cellc = [1 1 1];    % cell background
sty.eligc = [1 1 0];    % eligibility trace (background)
sty.cell = {'FaceColor', sty.cellc, 'LineWidth', 2, 'EdgeColor', [.3 .3 .3]};    % cell style
sty.Q = {{'b', 'LineWidth', 2}, {'m', 'LineWidth', 2}};
sty.goodactionc = [0 .5 0];
sty.badactionc = [.5 .5 .5];
sty.action = {'Color', sty.goodactionc, 'VerticalAlignment', 'middle', 'HorizontalAlignment', 'center', 'FontSize', 30, 'LineStyle', 'none'};
sty.perfQ = {'Color', 'k', 'LineWidth', 2, 'Marker', 'o'};
sty.perfh = {'Color', [0 .6 0], 'LineWidth', 2, 'Marker', 's'};
% sizing (cell is unit size)
sty.axismargin = .01;   % add this to axis limits around the graphics, to avoid clipping edges
sty.cellmargin = .01;   % add this around cells to allow thicker lines
sty.imgsize = .93;      % image size
sty.bulletsize = .4;     % state bullet size
sty.actionpadding = .1;  % padding to left and right of text
% images
fileinfo = what('machrepl'); imgpath = [fileinfo.path '/images/'];
sty.img_state = [imgpath 'state.png'];
sty.img_new = [imgpath 'new.png'];
sty.img_damaged = [imgpath 'damaged.png'];
% sty.img_statedimmest = [imgpath 'robot_dimmest.png'];
% sty.img_statedimmer = [imgpath 'robot_dimmer.png'];
% sty.cm = gray(128); sty.cm = sty.cm(33:end, :);    % w/o strong blacks and whites
% sty.mesh  = {'EdgeColor', [.3 .3 .3]};   % use dark meshes for readability

% labels
labels.x = 'state, x'; labels.u = 'u';
labels.Q = {'Q(x, Wait)', 'Q(x, Replace)'};
labels.h = {'h(x)'}; labels.V = {'V(x)'};

% ===================================
% ==== Prepare data for plotting ====

% flags for what we are going to show
show = struct;
show.world = 0;
show.state = [];
show.ertransition = [];
show.Q = [];
show.E = []; show.cleanupE = 0;
show.h = [];
show.perfQ = [];
show.perfh = [];
show.nothing = 0;

% compute quantities to show
% Q-iteration options
if cfg.qiter,
    show.world = 1;
    show.Q = Qh{cfg.ell+1};
    [Qmax jmax] = max(show.Q, [], 2);
    show.h = model.U{1}(jmax);
    if ~isempty(cfg.Qstar),
        show.perfQ = zeros(1, cfg.ell+1);
        for i = 1:cfg.ell+1, show.perfQ(i) = max(max(abs(cfg.Qstar - Qh{i}))); end;
        plotsize = [4 1];
        perfx = 0:cfg.ell; labels.xperf = 'iteration, ell'; labels.Qperf = 'Q - Q*';
    else
        plotsize = [3 1];
    end;
    gtitle = sprintf('Q-iteration, ell=%d', cfg.ell);
    if cfg.consoleout, dispQ(model, show.Q, cfg.ell); end;
    
% Policy iteration options
elseif cfg.qevaliter,
    show.world = 1;
    show.h = hh{cfg.ell};   % policy that is being evaluated
    show.Q = Qhh{cfg.ell, cfg.tau+1};
    if ~isempty(cfg.Qhstar),
        show.perfQ = zeros(1, cfg.tau+1);
        for i = 1:cfg.tau+1, show.perfQ(i) = max(max(abs(cfg.Qhstar{cfg.ell} - Qhh{cfg.ell, i}))); end;
        plotsize = [4 1];
        perfx = 0:cfg.tau; labels.xperf = 'policy evaluation iteration, tau'; labels.Qperf = 'Q - Q^h';
    else
        plotsize = [3 1];
    end;
    gtitle = sprintf('Policy evaluation, tau=%d (at policy iteration ell=%d)', cfg.tau, cfg.ell);
    if cfg.consoleout, dispQ(model, show.Q, cfg.tau); end;
elseif cfg.piter,
    show.world = 1;
    show.Q = Qh{cfg.ell};
    show.h = hh{cfg.ell+1};
    if ~isempty(cfg.hstar),
        show.perfh = zeros(1, cfg.ell+1);
        for i = 1:cfg.ell+1, show.perfh(i) = sum(cfg.hstar ~= hh{i}); end;
        plotsize = [4 1];
        perfx = 0:cfg.ell; labels.xperf = 'iteration, ell'; labels.hperf = 'h - h*';
    else
        plotsize = [3 1];
    end;
    gtitle = sprintf('Policy iteration, ell=%d', cfg.ell);
    if cfg.consoleout, disph(model, show.h, cfg.ell); end;

% Q-learning and SARSA options
elseif cfg.qlearnstep || cfg.sarsastep,
    show.world = 1;
    show.Q = Q;
    [Qmax jmax] = max(Q, [], 2);
    show.h = model.U{1}(jmax);
    show.state = X(:, cfg.k+1, cfg.trial);
    plotsize = [4 1];   % reserve the space for the performance plot
    if cfg.qlearnstep, algname= 'Q-learning'; else algname = 'SARSA'; end;
%     if expcfg.lambda > 0, show.E = E; end;
%     if cfg.qlearnstep,
%         if expcfg.ertype(1) ~= 'n', algname = 'ER-Q-learning';
%         elseif expcfg.lambda > 0,   algname = 'Q(lambda)-learning';
%         else                        algname = 'Q-learning';
%         end;
%     elseif cfg.sarsastep,
%         if expcfg.ertype(1) ~= 'n', algname = 'ER-SARSA';
%         elseif expcfg.lambda > 0,   algname = 'SARSA(lambda)';
%         else                        algname = 'SARSA';
%         end;
%     end;
    gtitle = sprintf('%s, trial %d, step %d', algname, cfg.trial, cfg.k);
% elseif cfg.qlearnerstep || cfg.sarsaerstep,
%     show.world = 1;
%     show.Q = Q;
%     % show.state = X(:, cfg.k+1, cfg.trial);
%     show.ertransition = X(:, [cfg.erk cfg.erk+1], cfg.ertrial);
%     plotsize = [4 1];    
%     if cfg.qlearnerstep,    algname = 'ER-Q-learning';
%     elseif cfg.sarsaerstep, algname = 'ER-SARSA';
%     end;
%     gtitle = sprintf('%s, trial %d, step %d [replaying trial %d, step %d]', algname, cfg.trial, cfg.k, cfg.ertrial, cfg.erk);
elseif cfg.qlearntrial || cfg.sarsatrial,
    show.world = 1;
    show.Q = Qh{cfg.trial+1};
    [Qmax jmax] = max(show.Q, [], 2);
    show.h = model.U{1}(jmax);
    if ~isempty(cfg.Qstar),
        show.perfQ = zeros(1, cfg.trial+1);
        for i = 1:cfg.trial+1, show.perfQ(i) = max(max(abs(cfg.Qstar - Qh{i}))); end;
        perfx = 0:cfg.trial; labels.xperf = 'trial'; labels.Qperf = 'Q - Q^*';
    elseif cfg.trial > 1,
        show.perfQ = deltah(1:cfg.trial+1);
        perfx = 0:cfg.trial; labels.xperf = 'trial'; labels.Qperf = 'Q_{trial+1} - Q_{trial}';
    end;
    plotsize = [4 1];
    if cfg.qlearntrial, algname = 'Q-learning'; else algname = 'SARSA'; end;
%     % may need to clean up eligibility trace
%     show.cleanupE = expcfg.lambda > 0;
%     if cfg.qlearntrial,
%         if expcfg.ertype(1) ~= 'n', algname = 'ER-Q-learning';
%         elseif expcfg.lambda > 0,   algname = 'Q(lambda)-learning';
%         else                        algname = 'Q-learning';
%         end;
%     elseif cfg.sarsatrial,
%         if expcfg.ertype(1) ~= 'n', algname = 'ER-SARSA';
%         elseif expcfg.lambda > 0,   algname = 'SARSA(lambda)';
%         else                        algname = 'SARSA';
%         end;
%     end;
    gtitle = sprintf('%s, trial %d completed', algname, cfg.trial);
%     
% % MPC options
% elseif cfg.mpcstep || cfg.mpctrial,
%     show.world = 1;
%     show.state = X(:, cfg.k+1, cfg.trial);      % note index order
%     plotsize = [1 1];
%     algname = 'MPC';
%     gtitle = sprintf('%s, trial %d, step %d', algname, cfg.trial, cfg.k);
%     
% Monte Carlo options
elseif cfg.mcstep,
    show.world = 1;
    show.state = X(:, cfg.k+1, cfg.trial);
    plotsize = [4 1];
    algname = 'Monte Carlo';
    gtitle = sprintf('%s, trial %d, step %d', algname, cfg.trial, cfg.k);
elseif cfg.mctrial,
    show.world = 1;
    if cfg.policyimproved || cfg.ell == 0, lastell = cfg.ell + 1;
    else lastell = cfg.ell;
    end;
    show.Q = Qh{cfg.trial+1};
    show.h = hh{lastell};
    if ~isempty(cfg.hstar),
        show.perfh = zeros(1, lastell);        
        for ell = 1:lastell, show.perfh(ell) = sum(cfg.hstar ~= hh{ell}); end;
        plotsize = [4 1];
        perfx = 0:lastell-1; labels.xperf = 'policy iteration'; labels.hperf = 'h - h*';
    else
        plotsize = [3 1];
    end;
    algname = 'Monte Carlo';
    if cfg.policyimproved, piterstatus = 'done'; else piterstatus = ''; end;
    gtitle = sprintf('%s, trial %d [piter %d %s, peval %d]', ...
        algname, cfg.trial, cfg.ell, piterstatus, cfg.peval_trial);
    if cfg.policyimproved && cfg.consoleout, 
        dispQ(model, show.Q, cfg.ell-1);
        disph(model, show.h, cfg.ell); 
    end;
elseif cfg.x,
    show.world = 1; show.state = x; plotsize = [1 1]; gtitle = 'Current state';
elseif ~isempty(cfg.Q) && ~isempty(cfg.h),
    show.world = 1; show.Q = cfg.Q; show.h = cfg.h; plotsize = [3 1]; gtitle = 'Q-function & Policy';
elseif ~isempty(cfg.Q),
    show.world = 1; show.Q = cfg.Q; plotsize = [3 1]; gtitle = 'Q-function';
elseif ~isempty(cfg.h),
    show.world = 1; show.h = cfg.h; plotsize = [1 1]; gtitle = 'Policy';
else
    show.nothing = 1;       % signal nothing is in fact being shown
end;

if isempty(cfg.gview);  
    create = 1; gview = struct; gview.figh = figure; setfigprop(cfg);
    if strcmp(cfg.plottarget, 'movie'),
        avifilename = [cfg.savedir cfg.savefig '.avi'];
        if exist(avifilename, 'file'), delete(avifilename); end;
        gview.aviobj = avifile(avifilename, 'compression', cfg.compression, 'quality', cfg.quality, 'fps', cfg.fps);
    end;
else
    create = 0; gview = cfg.gview; figure(gview.figh);
end;

% ===================================
% ==== Create world if needed =======
if show.world && create,
    subplot(plotsize(1), plotsize(2), 1);
    N = length(model.X{1}); gview.N = N;
    % --- draw cells; note cells are unit size; leave some room around to avoid clipping the edges
    set(gca, 'xlim', [-sty.axismargin N+sty.axismargin], 'ylim', [-sty.axismargin 1+sty.axismargin]);
    gview.cells = zeros(N, 1);
    for i = 1:N,
        gview.cells(i) = rectangle('Position', ...
            [i-1+sty.cellmargin, 0+sty.cellmargin, 1-2*sty.cellmargin, 1-2*sty.cellmargin], ...
            sty.cell{:});
    end;
    % --- draw goal objects
    gview.img_new = image('CData', imread(sty.img_new), ...
        'XData', .5+[-sty.imgsize/2 sty.imgsize/2], ...
        'YData', .5+[sty.imgsize/2 -sty.imgsize/2]);
    gview.img_damaged = image('CData', imread(sty.img_damaged), ...
        'XData', model.n-1+.5+[-sty.imgsize/2 sty.imgsize/2], ...
        'YData', .5+[sty.imgsize/2 -sty.imgsize/2]);
    % --- draw policy
    gview.actions = zeros(N, 2);
    for i = 1:N,
        % transform from data coords to figure coords (required to draw the action)
        [ax ay] = dsxy2figxy(gca, [i-1+sty.actionpadding, i-sty.actionpadding], [sty.actionpadding, 1-sty.actionpadding]);
        % "left" (action 1); then "right" (action 2) actions
        gview.actions(i, 1) = annotation('textbox', [ax(1), ay(1), ax(2)-ax(1), ay(2)-ay(1)], 'String', 'W', sty.action{:});
        gview.actions(i, 2) = annotation('textbox', [ax(1), ay(1), ax(2)-ax(1), ay(2)-ay(1)], 'String', 'R', sty.action{:});
    end;
    set(gview.actions(:), 'Visible', 'off');     % hide the actions    
    % --- draw state images on top of the actions
    gview.img_state = image('CData', imread(sty.img_state), ...
        'XData', 1.5+[-sty.bulletsize/2 sty.bulletsize/2], ...
        'YData', .5+[sty.bulletsize/2 -sty.bulletsize/2]);
%     gview.img_statedimmest = image('CData', imread(sty.img_statedimmest), ...
%         'XData', 1.5+[-sty.imgsize/2 sty.imgsize/2], ...
%         'YData', .5+[sty.imgsize/2 -sty.imgsize/2]);
%     gview.img_statedimmer = image('CData', imread(sty.img_statedimmer), ...
%         'XData', 1.5+[-sty.imgsize/2 sty.imgsize/2], ...
%         'YData', .5+[sty.imgsize/2 -sty.imgsize/2]);
    set(gview.img_state, 'Visible', 'off');    % hide the robot, it's not always needed
    % set(gview.img_statedimmer, 'Visible', 'off');    % hide the dreamy robot, it's not always needed
    % set(gview.img_statedimmest, 'Visible', 'off');    % hide the dreamy robot, it's not always needed
    % move this after the calls to dsxy2figxy to ensure axis limits are not
    % changed (recent MTB behavior)
    % credit for finding this bug: Yao Xiao
    axis image off manual;
    % --- create title and left-align it (remember handle)
    gview.title = title(gtitle);
    tpos = get(gview.title, 'Position'); 
    tpos(1) = 0;
    set(gview.title, 'Position', tpos, sty.title{:});
end;

if show.world,
    % change the title
    set(gview.title, 'String', gtitle);
end;

% ===================================================
% ==== Perform the update of the graphical view =====

if ~isempty(show.state),
    % show agent at updated (next) position
    set(gview.img_state, 'XData', show.state-1+.5+[-sty.bulletsize/2 sty.bulletsize/2], ...
        'YData', .5+[sty.bulletsize/2 -sty.bulletsize/2], 'Visible', 'on');
else set(gview.img_state, 'Visible', 'off'); 
end;

% if ~isempty(show.ertransition),
%     % don't show previous position , it's confusing
% %     set(gview.img_robotdimmest, 'XData', ...
% %         find(model.X{1}==show.ertransition(1, 1))-1+.5+[-sty.imgsize/2 sty.imgsize/2], ...
% %         'YData', .5+[sty.imgsize/2 -sty.imgsize/2], 'Visible', 'on');
%     % show agent at updated (next) position
%     set(gview.img_robotdimmer, 'XData', ...
%         find(model.X{1}==show.ertransition(1, 2))-1+.5+[-sty.imgsize/2 sty.imgsize/2], ...
%         'YData', .5+[sty.imgsize/2 -sty.imgsize/2], 'Visible', 'on');
% else set([gview.img_robotdimmer, gview.img_robotdimmest], 'Visible', 'off'); 
% end;

if ~isempty(show.Q),
    subplot(plotsize(1), plotsize(2), [2 3]);
    cla; 
    % pad & shift by .5 to have bars fall on top of integers
    xx = [model.X{1} model.X{1}(end)+1] - .5;
    stairs(xx, [show.Q(:, 1)' show.Q(end, 1)], sty.Q{1}{:}); hold on;
    stairs(xx, [show.Q(:, 2)' show.Q(end, 2)], sty.Q{2}{:});
    set(gca, 'xlim', [model.minx-.5 model.maxx+.5], 'XTick', model.X{1});
    xlabel(labels.x);
    legend(labels.Q{:}, 'Location', 'NorthEast');
end;

% if ~isempty(show.E),
%     show.E = max(show.E, [], 2);    % mean for the two actions
%     subplot(411);
%     % (debugging, show E in figure title)
%     % set(gcf, 'Name', sprintf('E=[%.2f, %.2f, %.2f, %.2f]', show.E(2), show.E(3), show.E(4), show.E(5)));
%     for i = 2:gview.N-1,
%         set(gview.cells(i), 'FaceColor', sty.cellc + show.E(i)*(sty.eligc - sty.cellc));
%     end;
% elseif show.cleanupE,
%     % not showing elig trace but may have shown it in the past -- cleanup
%     set(gview.cells, 'FaceColor', sty.cellc);
% end;

if ~isempty(show.h),
    % reset the actions to their default state
    set(gview.actions(:), 'Visible', 'off', 'Color', sty.goodactionc);
    % make right actions visible in states that take the respective action
    % visactions = false(1, length(show.h));
    visactions(show.h == 1) = gview.actions(show.h == 1, 1);
    visactions(show.h == 2) = gview.actions(show.h == 2, 2);
    set(visactions, 'Visible', 'on');
    % if optimal policy known: mark incorrect actions 
    if ~isempty(cfg.hstar),
        set(visactions(cfg.hstar ~= show.h), 'Color', sty.badactionc);
    end;
end;

if ~isempty(show.perfQ),
    subplot(plotsize(1), plotsize(2), 4);
    plot(perfx, show.perfQ, sty.perfQ{:});
    if length(perfx) < 10, set(gca, 'XTick', perfx); end;
    xlabel(labels.xperf);
    legend(labels.Qperf, 'Location', 'NorthEast');
end;

if ~isempty(show.perfh),
    subplot(plotsize(1), plotsize(2), 4);
    plot(perfx, show.perfh, sty.perfh{:});
    if length(perfx) < 10, set(gca, 'XTick', perfx); end;
    xlabel(labels.xperf);
    legend(labels.hperf, 'Location', 'NorthEast');
end;

% ==========================================
% ==== Save graphical output (if any) ======

if ~show.nothing,
    switch cfg.plottarget,
        case 'screen',
            drawnow;
            if isscalar(cfg.pause), 
                if cfg.pause >= 0, pause(cfg.pause),
                else pause;
                end;
            else
                % assuming cell with structure 
                % {condition pause; condition pause; ...}; should include at
                % least a default where the condition evaluates true
                for i = 1:size(cfg.pause, 1),
                    if eval(cfg.pause{i, 1}),
                        if cfg.pause{i, 2} >= 0, pause(cfg.pause{i, 2}),
                        else pause;
                        end;
                        break;
                    end;
                    % otherwise try the next condition
                end;
            end;
        case 'movie',
            % add the figure as a frame to the movie
            drawnow;
            gview.aviobj = addframe(gview.aviobj, gview.figh);
            % determine if we need to take a snapshot at this stage
            if ~isempty(cfg.snapshot),
                if length(cfg.snapshot) == 1,
                    takesnapshot = ((cfg.qiter || cfg.piter) && (cfg.ell == cfg.snapshot)) || ...
                        ((cfg.qlearntrial || cfg.sarsatrial || cfg.mctrial) && (cfg.trial == cfg.snapshot));
                elseif length(cfg.snapshot) == 2,
                    takesnapshot = (cfg.qevaliter && (cfg.ell == cfg.snapshot(1)) && (cfg.tau == cfg.snapshot(2))) || ...
                        ((cfg.qlearnstep || cfg.sarsastep || cfg.mcstep) && (cfg.trial == cfg.snapshot(1)) && (cfg.k == cfg.snapshot(2)));
                else
                    takesnapshot = 0;
                end;
            else
                takesnapshot = 0;
            end;
            if takesnapshot,
                % disp('snapshot');
                saveplot(gview.figh, [cfg.savedir cfg.savefig], cfg.snapshottarget);
            end;
        case {'latex', 'beamer'},
            % determine if we need to take a snapshot at this stage
            if ~isempty(cfg.snapshot),
                if length(cfg.snapshot) == 1,
                    takesnapshot = ((cfg.qiter || cfg.piter) && (cfg.ell == cfg.snapshot)); % || ...
                        % ((cfg.qlearntrial || cfg.sarsatrial || cfg.mctrial) && (cfg.trial == cfg.snapshot));
                elseif length(cfg.snapshot) == 2,
                    takesnapshot = (cfg.qevaliter && (cfg.ell == cfg.snapshot(1)) && (cfg.tau == cfg.snapshot(2))); % || ...
                        %((cfg.qlearnstep || cfg.sarsastep || cfg.mcstep) && (cfg.trial == cfg.snapshot(1)) && (cfg.k == cfg.snapshot(2)));
                else
                    takesnapshot = 0;
                end;
            else
                takesnapshot = 0;
            end;
            if takesnapshot,
                saveplot(gview.figh, [cfg.savedir cfg.savefig], cfg.snapshottarget);
            end;
    end;
end;

% Finalize movie if requested to do so
if cfg.finalize,
    if strcmp(cfg.plottarget, 'movie'),
        gview.aviobj = close(gview.aviobj);
    end;
end;

% save last figure if indicated
% if ~isempty(figh),
%     if ~isempty(cfg.datafile),
%         set(figh(end), 'Name', cfg.datafile, 'NumberTitle', 'off');
%     end;
%     saveplot(figh(end), [cfg.savedir cfg.savefig], cfg.plottarget);
% end;

varargout = {gview.figh, gview};

end
% END main function =================================================

% local function to display Q-function to console, in Latex format
function dispQ(model, Q, itx)
s = sprintf('$\\Q_{%d}$', itx);
for i = 1:size(Q, 1),
    s = [s sprintf(' & $%.3g\\,;\\,%.3g$', Q(i, 1), Q(i, 2))];
end;
s = [s '\\'];
disp(s); 
end

% and for policy
function disph(model, h, itx)
s = ['$\h_{' num2str(itx) '}$'];
for i = 1:length(h),
    s = [s ' & $' num2str(h(i)) '$'];
end;
s = [s '\\'];
disp(s);
end 

