function [xplus, rplus, terminal, pplus, refxplus] = switchedlin_mdp(m, x, u, w, refx)
% Implements the dynamics of a switched linear system
%  [XPLUS, RPLUS, TERMINAl] = SWITCHEDLIN_MDP(M, X, U)
% This function conforms to the specifications established by SAMPLE_MDP.

if ~m.stoch,
    if ~any(m.U == u), error('Invalid mode'); end;

    cost = 0;
    for i = 1:m.dwelltime,
        % saturate state
        xsat = min(max(x, -m.maxx), m.maxx);
        if any(x ~= xsat), disp('Warning: bounds reached'); end;

        % add stage cost to cost accumulator, possibly using the reference model
        if m.userefmodel,
            % saturate
            refxsat = min(max(refx, -m.maxx), m.maxx);
            if any(refx ~= refxsat), disp('Warning: ref state bounds reached'); end;
            % add cost
            cost = cost + (xsat - refxsat)' * m.Q * (xsat-refxsat);
        else
            % add cost
            cost = cost + xsat' * m.Q * xsat;
        end;

        % system dynamics
        if m.type(1) == 'l',    % linear
            x = m.A{u} * x;
        else                    % affine
            x = m.A{u} * x + m.b{u};
        end;

        % ref system dynamics
        if m.userefmodel,
            if m.type(1) == 'l',    % linear
                refx = m.Aref * refx;
            else                    % affine
                refx = m.Aref * refx + m.bref;
            end;
        end;
    end;
    xplus = x;
    if any(abs(xplus) > m.maxx), m.maxx, xplus, error('Bounds exceeded'); end;

    if m.userefmodel,
        refxplus = refx;
        if any(abs(refxplus) > m.maxx), m.maxx, refxplus, error('Ref state bounds exceeded'); end;
        pplus = [];
    end;

    if m.rewtype(1) == 'o' || m.rewtype(1) == 'r', %optimal variant or reference following: must optimize
        rplus = 1 - cost / m.maxcost; 
    elseif m.rewtype(1) == 's', % safety variant to get the worst possible cost and check the bound
        rplus = cost / m.maxcost; 
    end;

    terminal = 0;
    
else    % stochastic
    
    % saturate state
    xsat = min(max(x, -m.maxx), m.maxx);
    if any(x ~= xsat), disp('Warning: bounds reached'); end;
    cost = xsat' * m.Q * xsat;
    
    if nargout == 4, 
        % next states, rewards, terminal flags
        xplus = nan(m.p, m.M);
        for mode = 1:m.M,
            if m.type(1) == 'l',    % linear
                xplus(1:m.p-1, mode) = m.A{mode} * x(1:end-1);
            else                    % affine
                xplus(1:m.p-1, mode) = m.A{mode} * x(1:end-1) + m.b{u};
            end;
            xplus(m.p, mode) = mode;
            if any(abs(xplus(:, mode)) > m.maxx), xp = xplus(:, mode), maxx = m.maxx, error('Bounds exceeded'); end;
        end;
        terminal = zeros(1, m.M);   % no terminal states
        rplus = repmat(cost / m.maxcost, 1, m.M);     % the reward does not depend on the next state
        % note that if a dependence is introduced, the mode is one-step delayed: formula says
        % the NEXT mode should come in THIS reward, so take that into account!

        % set probas
        if x(end) == 0,         % first step, initial mode distr
            pplus = m.prob0;
        else                    % generic Markov step
            pplus = m.prob(x(end), :);
        end;        
    else
        % draw mode sample from either the first-state distribution if current mode is special value
        % 0 = initial state; or else according to current mode's distribution in the Markov chain
        if x(end) == 0,     cdf = cumsum(m.prob0);
        else                cdf = cumsum(m.prob(x(end), :));
        end;
        mode = 1; s = rand; while s > cdf(mode), mode = mode + 1; end;

        % next states, rewards, terminal flags
        if m.type(1) == 'l',    % linear
            xplus = [m.A{mode} * x(1:end-1); mode];
        else                    % affine
            xplus = [m.A{mode} * x(1:end-1) + m.b{u}; mode];
        end;
        % if any(abs(xplus) > m.maxx), xp = xplus(:, mode), maxx = m.maxx, error('Bounds exceeded'); end;
        terminal = 0;
        rplus = cost / m.maxcost;
        % note that if a dependence is introduced, the mode is one-step delayed: formula says
        % the NEXT mode should come in THIS reward, so take that into account!

    end;
    
end;

% END FUNCTION linsys_mdp() RETURNING xplus, rplus ==================================================
