function varargout = switchedlin_problem(what, varargin)
% Switched linear system where
% - in the deterministic case, the discrete input is the current mode.
% - in the stochastic case the modes evolve stochastically and there is no input
% For stochastic mode, the initial mode should be 0
% This function conforms to the specifications established by SAMPLE_PROBLEM.

switch what,
    
    case 'model'    
        % fixed params
        m.id = 'swlin'; 
        m.problem = @switchedlin_problem;
        m.fun = @switchedlin_mdp;
        m.plotfun = @switchedlin_plot;
        m.q = 1;                    % always one (integer) action: mode

        % configurable params
        PAR.type = 'lin';           % 'lin', 'affine'
        PAR.stoch = 0;              % whether the model is stochastic
        PAR.prob = [];              % Markov probabilities, if stochastic model
        PAR.prob0 = [];             % initial mode probabilities
        PAR.A = {};                 % cell array of system matrices A_1 A_2 ... A_M
        PAR.b = {};                 % for affine model: cell array of bias vectors b_1 b_2 ... b_M
        PAR.Q = [];                 % cost function matrix
        PAR.maxx = [];              % state bounds must be estimated by user
        PAR.Ts = 1;                 % conventional unit sample time
        PAR.dwelltime = 1;          % dwell time (how many samples to spend in the same mode)
        % only set this if you wish ALL modes to be set for exactly dwelltime timesteps
        % (so that effectively the system is lifted to a multiple-step variant)
        % if an algorithm that can impose a dwell time constraint on its own is used, set this to 1 and handle
        % things via the algorithm
        PAR.rewtype = 'opt';
            % 'opt'imize (minimize cost/maximize reward)
            % 'safe'ty bound (maximize cost/minimize reward)
            % 'ref'erence tracking
        
        PAR.refp = [];                        % state count for reference model (inferred automatically from size of Aref)
        PAR.refx0 = [];                       % initial state for reference model
        PAR.Aref = [];      
        PAR.bref = [];

        par = parseconfig(varargin{1}, PAR);
        m = copyfields(par, m, fieldnames(par));    % copy params to the model
        
        % process params and error checks
        m.userefmodel = strcmp(m.rewtype, 'ref');
        if m.stoch && (m.userefmodel || m.dwelltime > 1 || m.rewtype(1) ~= 's'), 
            error('Stochastic system only implemented for: no refmodel, dwelltime=1, rewtype=safe'); 
        end;
        
        m.M = length(m.A);
        if ~m.stoch, 
            m.p = size(m.A{1}, 1);
            m.U = 1:m.M; 
        else
            m.p = size(m.A{1}, 1) + 1;  % since we need to store the mode as an extra state
            m.U = -1;                   % singleton, dummy action for stochastic modes
            % increase vectors and matrices accordingly
            m.maxx = [m.maxx(:); m.M];
            m.Q(m.p, m.p) = 0;          % initialize with zeros the extra row and column
        end; 
        if any(size(m.Q) ~= m.p) || length(m.maxx) ~= m.p, error('Model size mismatch'); end;
        if m.userefmodel,
            m.refp = size(m.Aref, 1);
            if m.refp ~= m.p || length(m.refx0) ~= m.refp, error('Reference model size mismatch'); end;
        end;
        if m.stoch && (any(size(m.prob) ~= m.M) || length(m.prob0) ~= m.M || any(sum(m.prob, 2) ~= 1) || sum(m.prob0) ~= 1),
            error('Incorrect probability settings');
        end;
        % process reward
        switch m.rewtype,
            case {'opt', 'safe'},
                m.maxcost = m.dwelltime * m.maxx' * m.Q * m.maxx;   % valid both for + or - quad reward
                m.maxr = 1;
            case 'ref',
                m.maxcost = m.dwelltime * (2*m.maxx)' * m.Q * (2*m.maxx);   % the reference can be 2maxx away from the state
                m.maxr = 1;
        end;

        varargout = {m};
        
    case 'reset',   % handle state reset
        m = varargin{1};
        res = varargin{2};
        if ~isnumeric(res), error('Only direct numerical reset supported'); end;
        if ~m.stoch, 
            x0 = res;
            if m.userefmodel, varargout = {x0, m.refx0}; 
            else varargout = {x0};
            end;
        else
            x0 = [res(:); 0];   % "initial mode" quantifier
        end;

end;


% END doubleint_problem(), RETURNING out ====================================