function varargout = quadmulti_problem(what, varargin)
% Switched linear system where
% - in the deterministic case, the discrete input is the current mode.
% - in the stochastic case the modes evolve stochastically and there is no input
% For stochastic mode, the initial mode should be 0
% This function conforms to the specifications established by SAMPLE_PROBLEM.

switch what,
    
    case 'model'    
        % fixed params
        m.id = 'quad'; 
        m.problem = @quadmulti_problem;
        m.fun = @quadmulti_mdp;
        m.visualizefun = @quad_visualize;
        % m.plotfun = @switchedlin_plot;
        m.p = 13;                   % 13 states: x,y,z, phi, theta, psi, dx, dy, dz, dphi, dtheta, dpsi, wptidx
        m.q = 4;                    % 4 actions: collective, phi, theta, psi

        % configurable params
        PAR.type = 'standard';      %
        PAR.stoch = 0;              % whether the model is stochastic
%         PAR.m = 0.4365;             % vehicle total mass [kg]
%         PAR.Ix = 2.04016e-5;        % moment of inertia [kg*m^3]
%         PAR.Iy = 1.56771e-5;        % moment of inertia [kg*m^3]
%         PAR.Iz = 3.51779e-5;        % moment of inertia [kg*m^3]
        PAR.m=0.4472;          % drone from paper :15
        %PAR.m=0.490;           % our drone, in validateDynamics flights
        PAR.Ix = 0.0020;        % moment of inertia [Nms^2]
        PAR.Iy = 0.0016;        % moment of inertia [Nms^2]
        PAR.Iz = 0.0035;        % moment of inertia [Nms^2]

        PAR.b  = 1.114e-7;          % thrust coefficient [Ns^2]
        PAR.d = 3e-9;               % drag coefficient [Nms^2]
        PAR.l = 0.178;              % arm length, motor-to-centerOfMass [m]
        PAR.g = 9.80665;
        PAR.maxx = [20 20 20 pi pi pi 15 15 15 2*pi 2*pi 2*pi 100]';
%         PAR.minu = [7.244e-3 -6.835e-3 -6.835e-3 -1.466e-3]';
%         PAR.maxu = [1.159e-1  6.835e-3  6.835e-3  1.466e-3]';
        PAR.minu = [1.2508   -1.1807   -1.1807   -0.0488]';
        PAR.maxu = [20.0126   1.1807    1.1807    0.0488]';
        % numerical integration
        PAR.Ts = 0.1;                 % scalar, sampling time      
        PAR.odemethod = 'fixedode';    % 'varode', 'fixedode', 'eul'
        PAR.odesolver = @ode4;         % solver function
        PAR.odesteps = 5;              % # of steps per sample time (ignored for varode)        
        PAR.odet = 0 : PAR.Ts / PAR.odesteps : PAR.Ts;
        % reward
        PAR.rewtype = 'nlqr';
        % PAR.Q = diag([1 1 0, 0.1 0.1 0, 0.05 0.05 0, 0 0 0, 0]);
        % PAR.Q = diag([1 1 0.1, 0.3 0.3 0, 0.1 0.1 0.1, 0 0 0.1, 0]);
        PAR.Q = diag([1 1 0, 0.1 0.1 0.1, 0.1 0.1 0.1, 0.1 0.1 0.1, 0]);
        % Lucian: I am not sure that the weights with 0.1 everywhere make sense; e.g. we don't want
        % to stabilize yaw at any particular value. I used the weights below:
        % PAR.Q = diag([1 1 1, 0.3 0.3 0, 0.1 0.1 0.1, 0 0 0.1, 0]);
        PAR.xgoal = zeros(13, 10);
        PAR.wptNr = 1;                  % # of waypoints, i.e. xgoals of 13 states

        if nargin < 2, 
            par = PAR;
        else
            par = parseconfig(varargin{1}, PAR);
        end;
        m = copyfields(par, m, fieldnames(par));    % copy params to the model
        
        % process params and error checks
        m.userefmodel = 0;      % no reference model
        
        if any(size(m.Q) ~= m.p) || length(m.maxx) ~= m.p, error('Model size mismatch'); end;
        
        % process reward
        switch m.rewtype,
            case 'nlqr',
                if any(max(m.xgoal,[],2) > m.maxx | min(m.xgoal,[],2) < -m.maxx), error('Goal state not within bounds'); end;
                % Lucian: this max(m.xgoal,[],2) seems problematic, what if the goal from which maxx
                % is furthest away is different from the maximum value among goals?
                maxdx = max(m.maxx - max(m.xgoal,[],2), max(m.xgoal,[],2) - (-m.maxx)); %k select max from xgoal for each state
                m.unnorm_maxr = maxdx' * m.Q * maxdx;
                m.maxr = 1;
        end;

        varargout = {m};
        
    case 'reset',   % handle state reset
        m = varargin{1};
        res = varargin{2};
        if ~isnumeric(res), error('Only direct numerical reset supported'); end;
        if ~m.stoch, 
            x0 = res;
            varargout = {x0};
        else
            x0 = [res(:); 0];   % "initial mode" quantifier
        end;

end;


% END quadmulti_problem(), RETURNING out ====================================