function [xplus, rplus, terminal] = quadrotor_mdp(M, x, u)
% Implements the discrete-time dynamics of a quadrotor
%  [XPLUS, RPLUS, TERMINAL] = INV_MDP(M, X, u)

%k I commented this out, though it might be needed, but then we have to
%perform it for all the inputs
% u = [coll, phi, theta, psi]'
u = sat(u, M.minu, M.maxu);

if M.odemethod(1) == 'v',          % Variable-step ODE
	[odet, odey] = M.odesolver(@quadrotor_trans, M.odet, x, M.odeopt, M, u);
    xplus = odey(end, :)';

elseif M.odemethod(1) == 'f',      % Fixed-step ODE
    odey = M.odesolver(@quadrotor_trans, M.odet, x, M, u);
    xplus = odey(end, :)';

elseif M.odemethod(1) == 'e',      % Euler, x(k+1) = x(k) + Ts * f(x(k), u(k)) (possibly multistep)
    dt = M.Ts / M.odesteps; xplus = x;
    for i = 1:M.odesteps,
        xplus = xplus + dt * quadrotor_trans(0, xplus, M, u);
    end;
end;

% x,y,z, phi, theta, psi, dx, dy, dz, dphi, dtheta, dpsi
% 1 2 3  4    5      6    7   8   9   10    11      12
xplus(4) = mod(xplus(4) + pi, 2*pi) - pi;
xplus(5) = mod(xplus(5) + pi, 2*pi) - pi;
xplus(6) = mod(xplus(6) + pi, 2*pi) - pi;

%k should rethink what saturation in our case means, I think it is
%important only for x(7)-x(12), the velocities, but it is hard to set
%limits for them, as we have limits for the rotor speeds only

switch M.rewtype(1),
    case 'n',       % nlqr stabilization to zero
        % saturated state for the benefit of the reward (but the model keeps it as-is)
        xsat = sat(xplus, -M.maxx, M.maxx);
        rplus = 1 - ((xsat-M.xgoal)' * M.Q * (xsat-M.xgoal)) / M.unnorm_maxr;
    otherwise,
        error('Reward type [%s] unknown', M.rewtype);
end;

terminal = 0;   % task is continuing

end     % quadrotor_singletransition local function



function dx = quadrotor_trans(t, x, M, u)
%  Implements the transition function of the inverted pendulum setup
%
%  Parameters:
%   T           - the current time. Ignored, added for compatibility with
%               the ODEXX function style.
%   X           - the current state
%   M           - the model structure specifying physical params
%   u           - the command
%
%  Returns:
%   DX          - the derivative of the state
%

% x,y,z, phi, theta, psi, dx, dy, dz, dphi, dtheta, dpsi
% 1 2 3  4    5      6    7   8   9   10    11      12
dx(1,1) = x(7);
dx(2,1) = x(8);
dx(3,1) = x(9);
dx(4,1) = x(10);
dx(5,1) = x(11);
dx(6,1) = x(12);

% u = [coll, phi, theta, psi]'

dx(7,1) = ( cos(x(4)) * sin(x(5)) * cos(x(6)) + sin(x(4)) * sin(x(6)) )* u(1) / M.m;
dx(8,1) = ( cos(x(4)) * sin(x(5)) * sin(x(6)) - sin(x(4)) * cos(x(6)) )* u(1) / M.m;
dx(9,1) = -M.g + ( cos(x(4)) * cos(x(5))  )* u(1) / M.m;
%? plus or minus M.g. Seems to be fine with minus, actually it opposes the
%input u(1) so it's fine
dx(10,1) = x(11) * x(12) * (M.Iy - M.Iz) / M.Ix + 1 / M.Ix * u(2);
dx(11,1) = x(10) * x(12) * (M.Iz - M.Ix) / M.Iy + 1 / M.Iy * u(3);
dx(12,1) = x(10) * x(11) * (M.Ix - M.Iy) / M.Iz + 1 / M.Iz * u(4);

% end QUADROTOR_TRANS returning DX
end