function varargout = rarm_problem(what, varargin)
% Robotic arm problem setup.
%   OUT = RARM_PROBLEM(WHAT)
% Conforms to the specifications of SAMPLE_PROBLEM (but does not implement 'info' mode for now).
%
% In 'model' mode:
%   MODEL = RARM_PROBLEM(MODE, [MODELCFG])
%       MODELCFG    - model config, see code below for commented defaults
% 
% Supports 'reset' mode
% Implements rewards based on a reference model 

% default state for replays
% x0 = [];
gamma = 0.98;

switch what

    case 'shortid', varargout = {'rarm'};
    
    case 'model',        
        % Note that hooks exist for different model types
        % However, for now, the model type has no effect 
        if nargin >= 2 && ~isempty(varargin{1}),
            cfg = varargin{1};
        else cfg = struct;
        end;

        MODEL.type = 'default';
            % 'default' or 'acrobot' to actuate only middle joint; if so dimensionality of R must be changed
        % physics as in FUZZ-IEEE paper
        MODEL.wrap = 1;                          % wrapping enabled (1 -- default) or disabled (0)
        MODEL.len =      [.4     .4];            % [m] link lenghts
        MODEL.m =        [1.25    .8];           % [kg] link masses
        MODEL.I =        MODEL.m .* (MODEL.len.^2) / 3; % [kg.m^2] inertias, proportional with mass 
        MODEL.c =        MODEL.len / 2;
        MODEL.b =        [0.08     0.02];        % [kg/sec] joint dampings 
        MODEL.maxomega = [2*pi;  2*pi];          % [rad/sec] maximum angular speeds of links
        MODEL.maxtau =   [1.5;    1];            % [Nm] maximum torques
        MODEL.orientation = 'horiz';             % 'horiz' or 'vert', arm orientation
        % stochasticity
        MODEL.stoch_prob = [];                   % if stochastic: probabilities with which...
        MODEL.stoch_gain = [];                   % the action variables are multiplied with these values
            % size: (q or 1) x Noutcomes, where q = dim(u); if 1 then the same values apply to all actuators)
        % reward/goal
        MODEL.rewtype = 'lqr';
        % params for 'lqr' reward
        MODEL.Q = diag([1 0.05 1 0.05]);
        MODEL.R = zeros(2, 2);
        % params for 'vectorfield' - following v.f. in cartesian position coordinates
        MODEL.vectorfield = @(t, pos, c, rad, vel, slope)...
            -4 * slope * ((pos-c)'*(pos-c) - rad^2) * (pos-c) + vel*[-pos(2)+c(2); pos(1)-c(1)];
        % default center, rad, and velocity
        MODEL.vectorfield_params = {...
            [.3; .3], ...                       % center
            .4, ...                             % radius (cartesian)   
        	10,...                              % multiplier for rotation-inducing factor in vector field
            15};                                % multiplier for "slope" towards circle region
        % parameters for ref traj in cartesian coordinates
        MODEL.refmodel = @refmodel_circle;      % function for reference model
        MODEL.refmodel_hull = @refmodel_circle_hull;% function for convex polygon including reftrajs
        MODEL.refp = 1;                         % state count for reference model
        MODEL.refx0 = 0;                        % initial state for reference model
        MODEL.refmodel_params = {...
            [.3; .3], ...                       % center (cartesian)
            .3, ...                             % radius (polar)
        	0.5                                 % angular velocity (polar)
        };
        % integration
        MODEL.Ts = .05;
        MODEL.odemethod = 'fixed-ode';
        MODEL.odesolver = 'ode4';
        MODEL.odesteps = 5;
        
        % process any user-configured fields
        model = parseconfig(cfg, MODEL);
        
        % create standard model variables, derived quantities 
        model.p = 4;        % always 4 states
        model.maxx = [pi; model.maxomega(1); pi; model.maxomega(2)]; 
        switch model.type,
            case 'default'
                model.q = 2;        % both joints actuated, 2 actions
                model.maxu = model.maxtau(:); 
            case 'acrobot',
                model.q = 1;
                model.maxu = model.maxtau(2);
                if ~strcmp(model.rewtype, 'nlqr'), disp('WARNING! "Acrobot" actuation only tested with nlqr rewards'); end;
            case 'rotpend',
                model.q = 1;
                model.maxu = model.maxtau(1);
                if ~strcmp(model.rewtype, 'nlqr'), disp('WARNING! "Rotpend" actuation only tested with nlqr rewards'); end;
        end;
        model.fun = @rarm_mdpstandalone;
        model.plotfun = @rarm_plot;
        model.visualizefun = @rarm_visualize;
        model.tocartesian = @rarm_tocartesian;
        
        % process stochastic options
        model.det = isempty(model.stoch_prob);  % check if deterministic
        if ~model.det, model = stochact_mdpwrapper(model, 'prepare'); end;
        
        % flag if using reference model
        model.userefmodel = strcmp(model.rewtype, 'refmodel'); 
        % compute max reward
        switch model.rewtype,
            case 'lqr',     % LQR
                model.maxr = model.maxx' * model.Q * model.maxx + model.maxu' * model.R * model.maxu;
            case 'nlqr',    % LQR normalized to interval [0, 1]
                model.nonnorm_maxr = model.maxx' * model.Q * model.maxx + model.maxu' * model.R * model.maxu;
                model.maxr = 1;
            case 'vectorfield',
                model.maxr = [];   % unknown
            case 'refmodel',    % reference model for end-effector trajectory in cartesian coords
                % estimate the maximal reward as the maximal squared distance between a simple convex polygon
                % including all possible end-effector trajectories, and a similar poligon for the reference
                % trajectory
                % maximum of distances between all combination of polygon vertices
                XYR = model.refmodel_hull(model, model.refmodel_params{:});
                r = model.len(1) + model.len(2);
                XY = [-r -r r r; -r r r -r];
                maxdiff = -Inf;
                for i = 1:size(XYR, 2), 
                    for j = 1:size(XY, 2),
                        maxdiff = max(maxdiff, sum((XYR(:, i)-XY(:, j)).^2));
                    end; 
                end;
                model.nonnorm_maxr = maxdiff;
                model.maxr = 1;
            % note box reward is currently unsupported 
        end;
        
        % create discretization helper vars
        switch(model.odemethod)
            case 'ode'
                model.odet = [0 model.Ts/2 model.Ts];
                model.odeopt = odeset;
            case 'fixed-ode'
                model.odet = 0 : model.Ts / model.odesteps : model.Ts;
            case 'euler'
                % do nothing
        end;
        
        % compute physical helper vars
        model.g = 9.81;
        model.P1 = model.m(1) * model.c(1)^2 + model.m(2) * model.len(1)^2 + model.I(1);
        model.P2 = model.m(2) * model.c(2)^2 + model.I(2);
        model.P3 = model.m(2) * model.len(1) * model.c(2);
        model.g1 = (model.m(1) * model.c(1) + model.m(2) * model.len(1)) * model.g;
        model.g2 = model.m(2) * model.c(2) * model.g;
        model.neglectG = strcmp(model.orientation, 'horiz');

        varargout = {model};
        
    case 'reset',
        model = varargin{1};
        reset = varargin{2};
        if isempty(reset), reset = 'rand'; end;
        if isnumeric(reset),
            if size(reset, 2) > 1,      x0 = reset(:, unidrnd(size(reset, 2)));
            else                        x0 = reset;
            end;
        elseif strcmp(reset, 'rand'),
            x0 = (2*rand(model.p, 1) - 1) .* model.maxx;
        end;
        if model.userefmodel, varargout = {x0, model.refx0};
        else varargout = {x0};
        end;
        
    case 'fuzzy',
        % use the default grids
        % Note: These grids are used by the original fuzzy QI experiments
        % Newer experiments define their own discretizations
        TH = [-180:50:-30 -15 -5 0 5 15 30:50:180] * pi/180;
        OM = [-360 -180 -30 0 30 180 360] * pi/180;
        % Note the cmd discretization should actually rely on the bounds set by the model/user,
        % but there currently is no mechanism to pass the model structure back to the function in
        % the non-model modes
        TAU1 = [-1.5 0 1.5];
        TAU2 = [-1 0 1];
        cfg.xgrids = {TH, OM, TH, OM};
        cfg.ugrids = {TAU1, TAU2};
        cfg.gamma = gamma;
        varargout = {cfg};

    case {'lspi', 'lspe'},
        cfg.gamma = gamma;
        % note the code on next line will only work if maxomega is different from 2*pi
        % cfg.x0 = (2*rand(4, 1) - 1) * [pi; 2*pi; pi; 2*pi];
        varargout = {cfg};
        
    case 'fittedqi';
        cfg.gamma = gamma;
        varargout = {cfg};
       
    case 'X0',      % representative set of states
        if nargin < 2, xt = 'zero';         % Default is simply the zero initial state
        else xt = varargin{1}; end;
        switch xt,
            case 'zero',
                X0 = {0, 0, 0, 0};
            case 'pos',
                X0 = {-pi:pi/3:pi, 0, -pi:pi/3:pi, 0};
            otherwise,
                error(['Unknown X0 type [' xt ']']);
        end;
        varargout = {X0};
end;

end
% END doubleint_problem(), RETURNING out ====================================


% %%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local functions

% find end-effector cartesian coords
function pos = rarm_tocartesian(model, x)
pos = [model.len(1)*cos(x(1)) + model.len(2)*cos(x(1)+x(3)); ...
       model.len(1)*sin(x(1)) + model.len(2)*sin(x(1)+x(3))];
end

% Circle reference trajectory
% Receives: model, current angle on circle, center, radius, angular velocity
% Returns: current reference cartesian coords, next angle on circle
% refx is the angle around the reference circle
function [refpos, refxplus] = refmodel_circle(model, refx, c, r, omegadot)
    refpos = [c(1) + r * cos(refx); c(2) + r * sin(refx)];
    % Euler integration is sufficient for this
    refxplus = refx + model.Ts * omegadot;
end
% Convex polygon including set of reference trajectories
function XY = refmodel_circle_hull(model, c, r, omegadot)
    XY = [c(1)-r c(1)-r c(1)+r c(1)+r; c(2)-r c(2)+r c(2)+r c(2)-r];
end

% code for model type switch
%         % pre-parse for model type
%         TYPE.type = 'lqr'; tt = parseconfig(cfg, TYPE);
%         switch (tt.type),
%             case 'lqr',     % do nothing
%             otherwise,      error(['Unknown model type [' tt.type ']']);
%         end;


%         MODEL.vectorfield_params = {...
%             [-MODEL.len(1)*sin(pi/6) + MODEL.len(2); MODEL.len(1)*cos(pi/6)], ...    % center
%             MODEL.len(2)/3,              ...    % radius (cartesian)   
%         	5,...                               % multiplier for rotation-inducing factor in vector field
%             10};                                % multiplier for "slope" towards circle region
