function  [uroot, planinfo, plandata] = opd_lex(xroot, model, cfg)
% OPD Optimistic planning for deterministic systems
% Variant in Lex's code tree

% Author: Lucian Busoniu
% Versions:

CFG.gamma = [];                 % discount factor
CFG.n = [];                     % expansion budget
CFG.criterion = 'optim';        % expansion criterion: optim(istic) or unif(orm)
CFG.U = [];                     % discrete, flat action space
% data reuse options
CFG.reusetree = 0;              % try reusing previous tree (only when k is set and > 1)
CFG.reusetree_maxsize = 0;      % do not reuse tree if larger (has more nodes) than this; 0 = unlimited size
CFG.usemem = 0;                 % use memory of b-values
CFG.reusemem = 0;    
CFG.lipschitz = [];             % Lipschitz constant to use in bound estimates
CFG.KK = [];
CFG.S = [];
CFG.LLSSVR=[];
CFG.mem_maxsize = 0;            % 0: no limit, as large as needed,
% <= 10: given as a multiple of the tree size (ignoring root)
% > 10: given as number of points in memory
CFG.k = -1;                     % discrete time step
% implementation options
CFG.compat = 0;                 % 0: use default, 1: use structures
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed

cfg = checkparams(cfg, CFG);

% forward call to compatibility version if configured
if cfg.compat == 1,
    [uroot, planinfo, plandata] = opd_structures(xroot, model, cfg);
    return;
end;

persistent Tparent Tchildren Tdepth Tleaf Tx Tr TR Tterm Tb treesize
persistent Mx Mb memsize
persistent LLSSVR
global LSSVR NET
cfg = checkparams(cfg, CFG);

if cfg.reusetree && cfg.usemem, error('Memory and tree reuse cannot be used together.'); end;

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.gamma; n = cfg.n; criterion = cfg.criterion;
U = cfg.U; usemem = cfg.usemem; lipschitz = cfg.lipschitz;
KK=cfg.KK; S=cfg.S;
LLSSVR=cfg.LLSSVR;      %??? Is this correct -- don't we overwrite the approximator at each call?

M = size(U, 2);      % number of discrete actions
dimx = model.p;      % number of state vars

% start counting time
t = cputime;

% 1a) Memory initialization: if uninitialized, or we are just starting a new trajectory
% must be done before tree, as otherwise tree may attempt to use existing memory
if usemem && (isempty(memsize) || cfg.k <= 1) && cfg.reusemem==0,
    % global methods only called once
    if usemem==4,
    LSSVR=cfg.LSSVR;
    end
    if usemem==6,
    NET=cfg.NET;
    end
    Mx = NaN(dimx, 0, 'double');      % states
    Mb = NaN(0, 'double');            % b-values
    memsize = 0;
    disp('Memory reuse off')
end;

if cfg.reusemem && (isempty(memsize)),
    Mx = NaN(dimx, 0, 'double');      % states
    Mb = NaN(0, 'double');            % b-values
    memsize = 0;
    disp('Memory reuse on, memory is empty')
end;


%if usemem && (isempty(memsize)==0),
% Mx = NaN(dimx, 0, 'double');      % states
%   Mb = NaN(0, 'double');            % b-values
%   memsize = 0;
%end

% 1b) Tree initialization


maxtreesize = M*n + 1;
% prepare tree data
Tparent = zeros(2, maxtreesize, 'uint32');        % parent and parent action
Tchildren = zeros(M, maxtreesize, 'uint32');      % child index for each action
Tdepth = zeros(1, maxtreesize, 'uint16');         % depth (uint16 as depth > 65536 considered unfeasible)
Tleaf = false(1, maxtreesize);                    % leaf flag
Tx = NaN(dimx, maxtreesize, 'double');            % state
Tr = NaN(1, maxtreesize, 'double');               % reward
TR = NaN(1, maxtreesize, 'double');               % partial return up to node (cumulative reward)
Tterm = false(1, maxtreesize);                    % terminal flag
Tb = NaN(1, maxtreesize, 'double');               % b-values

% initialize the root node (note 0 = null pointer)
root = 1;
Tparent(:, root) = 0;
Tchildren(:, root) = 0;
Tdepth(root) = 0;
Tleaf(root) = true;
Tx(:, root) = xroot;
Tr(root) = 0;
TR(root) = 0;
Tterm(root) = false;    % presumably, we are not asked to make a decision in a terminal state

if usemem && memsize > 0,   % memsize >0 is false also if memsize is empty
    if usemem==1,
        Tb(:, root) = estimate_b_Lipschitz(xroot, Mx, Mb, discount, lipschitz,S);
    elseif usemem==2,
        Tb(:, root) = estimate_b_knn(xroot, Mx, Mb,KK,S);
    elseif usemem==3,
        Tb(:, root) = estimate_b_llr(xroot, Mx, Mb,KK,S);
    elseif usemem==4,
        Tb(:, root)= estimate_b_lssvr(xroot,Mx,LSSVR);
    elseif usemem==5,
        Tb(:, root)= estimate_b_llssvr(xroot,Mx,Mb,LLSSVR);
    elseif usemem==6,
        Tb(:, root)= estimate_b_net(xroot,NET);
    end
    
else
    Tb(root) = 1/(1-discount);
end;
treesize = 1;

% 2) Main tree construction loop
for tau = 1:n,
    
    % 2a) select candidate leaves for expansion
    leaves = find(Tleaf);
    if criterion(1) == 'o',
        % optimistic: max b-value RELATIVE TO THE ROOT
        B = TR(leaves) + discount.^double(Tdepth(leaves)) .* Tb(leaves);
        leaves = leaves(B == max(B));
        % Notes:
        % - due to OPD structure we are able to directly go for the max-b
        % leaf, equivalent to iteratively developing all optimistic paths
        % - without memory use, Tb(leaves) will always be 1/1-gamma
    elseif criterion(1) == 'u',
        % uniform: smallest depth
        leaves = leaves(Tdepth(leaves) == min(Tdepth(leaves)));
    end;
    % select only nonterminal leaves; if none, nothing left to do
    leaves = leaves(~Tterm(leaves)); if isempty(leaves); break ;end;
    % break any ties between remaining candidates deterministically
    nodetoexpand = leaves(1);
    
    % 2b) expand the chosen node
    Tleaf(nodetoexpand) = false;        % no longer a leaf
    % run through the discrete actions
    for j = 1:M,
        % find index where new node will be, save it on the children table of current node
        child = treesize + 1;
        Tchildren(j, nodetoexpand) = child;
        
        % simulate transition and add data to tree
        [Tx(:, child), Tr(child), Tterm(child)] = ...
            feval(model.fun, model, Tx(:, nodetoexpand), U(:, j));
        % if Tr(child) < 0 || Tr(child) > 1, error('Rewards must be in [0, 1]'); end;
        
        % create new node
        Tparent(:, child) = [nodetoexpand; j];      % parent is expanded node, along jth action
        Tchildren(:, child) = 0;                    % init children pointers to "null"
        Tdepth(child) = Tdepth(nodetoexpand) + 1;   % one deeper than the parent
        Tleaf(child) = true;                        % new child is a leaf
        TR(child) = TR(nodetoexpand) + discount^double(Tdepth(nodetoexpand)) * Tr(child);
        if Tterm(child),
            Tb(child) = 0;             % values of terminal nodes exactly 0
        elseif usemem && memsize > 0,
            if usemem==1,
                Tb(child) = estimate_b_Lipschitz(Tx(:, child), Mx, Mb, discount, lipschitz, S);
            elseif usemem==2,
                Tb(child) = estimate_b_knn(Tx(:, child), Mx, Mb,KK,S);
            elseif usemem==3,
                Tb(child) = estimate_b_llr(Tx(:, child), Mx, Mb,KK,S);
            elseif usemem==4,
                Tb(child) = estimate_b_lssvr(Tx(:, child),Mx,LSSVR);
            elseif usemem==5,
                Tb(child) = estimate_b_llssvr(Tx(:, child),Mx,Mb,LLSSVR);
            elseif usemem==6,
                Tb(child) = estimate_b_net(Tx(:, child),NET);
                
            end
        else
            Tb(child) = 1/(1-discount);
        end;
        
        % we just added a new node, increase tree size
        treesize = treesize + 1;
    end;
    
    % 2c) update b-values affected by expanded node (path to root)
    node = nodetoexpand;
    while node ~= 0,       % when node is 0, it means the root has already been processed
        % account for the possibility of fewer successor states than the maximum L
        children = Tchildren(:, node);
        Tb(node) = max(Tr(children) + discount .* Tb(children));
        node = Tparent(1, node);
    end;
    
end;    % end main tree construction loop

% 3) Update memory, if using it
if usemem,
    
    innernodes = find(~Tleaf(1:treesize));
    memsize_initial= size(Mx,2)+length(innernodes);
    
    if usemem==2 || usemem==3 || usemem==4 || usemem==5
        [Mx Mb]=mem_man_llr(Mx, Mb, Tx(:, innernodes),Tb(innernodes));
        memsize_redundant=memsize_initial-size(Mx,2);
    end
    
    %remove duplicates and add samples to memory
    
    mem_init=size(Mx,2)+length(innernodes);
    [Mx Mb]=remove_duplicates(Mx, Mb, Tx(:, innernodes),Tb(innernodes));
    memsize=size(Mx,2);
    memsize_duplicates = mem_init - memsize;
    
    % memory management lipschitz done after removing duplicates
    if usemem==1
        [Mx Mb]=mem_man_lipschitz(Mx, Mb,discount, lipschitz,S);
        memsize_redundant=memsize-size(Mx,2);
    end
    
    if usemem==4
        % train SVR model
        LSSVR  = train_lssvr(Mx,Mb,LSSVR);
    end
    if usemem==6
        % train NET model
        memsize_redundant=0;
        NET  = train_net(Tx(:, innernodes),Tb(innernodes),NET);
    end;
    
    % No pruning
    memsize_pruned=0;
    
    
   
end;

% 4) Find action at the tree root
% choose a leaf with maximal u-value (i.e., truncated return)
leaves = find(Tleaf); [Rmax, jmax] = max(TR(leaves)); greedy = leaves(jmax);
% navigate greedy path upwards in the tree until level 1, just after root
while Tparent(1, greedy) ~= root, greedy = Tparent(1, greedy); end;
% choose the action that was taken at root to arrive at this level-1 node
% -- this is the action greedy in util
uroot = U(:, Tparent(2, greedy));
% Note: this avoids the need to store and backup util-values while building the tree

% output stats if requested
if nargout >= 2,
    planinfo.n = tau;
    planinfo.depth = max(Tdepth);
    planinfo.trun = cputime - t;
    planinfo.treesize = treesize;
    planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d, treesize=%d', ...
        tau, planinfo.depth, planinfo.treesize);
    if cfg.reusetree,
        planinfo.reusetree_clear = reusetree_clear;
        planinfo.reusetree_size = subtreesize;
        planinfo.message = [planinfo.message sprintf('; reusedtreesize=%d', planinfo.reusetree_size)];
    end;
    if usemem,
        planinfo = copyfields(...
            varstostruct('memsize', 'memsize_initial', 'memsize_duplicates', 'memsize_redundant', 'memsize_pruned'), ...
            planinfo);
        planinfo.message = [planinfo.message sprintf('; memsizes:init=%d/dup=%d/redund=%d/pruned=%d/final=%d', ...
            memsize_initial, memsize_duplicates, memsize_redundant, memsize_pruned, memsize)];
    end;
end;
% output tree data and memory
if nargout >= 3,
    plandata.Tparent = Tparent(:, 1:treesize);
    plandata.Tchildren = Tchildren(:, 1:treesize);
    plandata.Tdepth = Tdepth(1:treesize);
    plandata.Tleaf = Tleaf(1:treesize);
    plandata.Tx = Tx(:, 1:treesize);
    plandata.Tr = Tr(1:treesize);
    plandata.TR = Tr(1:treesize);
    plandata.Tterm = Tterm(1:treesize);
    plandata.Tb = Tb(1:treesize);
    plandata.Mx=Mx;
    plandata.Mb=Mb;
    plandata.LSSVR=LSSVR;
    plandata.LLSSVR=LLSSVR;
    plandata.NET=NET;
     % visualize including the points that we'll erase
%     if usemem && cfg.verb >= 3,
%         visualizememory(plandata, cfg);
%         set(gcf, 'Name', sprintf('k=%d: Visualizing memory and any cleaned up points', cfg.k));
%     end;
end;

end % main function

