function [eJ, J, hist] = sw_mc_rbfnearestpolicy(c, rad, uind, model, X0, cfg)
% Monte-Carlo evaluation of parameter vector for CE-PS with RBFs on the simple walker
%   [EJ, J] = SW_MC_RBFNEARESTPOLICY(C, RAD, U, MODEL, X0, CFG)
% Monte-Carlo evaluation of RBF parameters and action assignment for one parameter sample
% of cross-entropy policy search on the simple walker problem. 
% The policy is computing by choosing the action assigned to the nearest RBF.
%
% Parameters:
%   C       - RBF centers, p x N where p state variable size, N number of basis functions
%   RAD     - RBF radii, p x N where p state variable size, N number of basis functions
%   UIND    - assignment of discrete actions to centers, in 1-based integer indices
%   MODEL   - problem model
%   X0      - distribution of initial states. For now, assumed to be uniform over a 
%       set of discrete values, each given in a column of X0
%   CFG     - configuration parameters. Requires fields (subject to change):
%       U, mc_maxsteps, mc_nsim - see cerbfqi for description
%
% Returns:
%   EJ      - expected cost
%   E       - cost of every initial state sample

% precompute discounting vector, pre-initialize reward vector
disc = cumprod([1 cfg.gamma+zeros(1, cfg.mc_maxsteps)]);

% form explicit policy for ease of use
h = cfg.U(:, uind);
N0 = size(X0, 2); J = zeros(N0, 1);

% time
Ts = model.Ts; 
Tend = cfg.mc_maxsteps * Ts;
T = 0:Ts:Tend;
% ode settings
odeopt  = odeset('events', @sw_events);
% action has to be global
global SWMCRBFNEAREST_U
for i = 1:N0,
    % initialize simulation for i-th initial state 
    X = nan(cfg.mc_maxsteps+1, model.p);
    SWMCRBFNEAREST_U = nan(cfg.mc_maxsteps, 1);
    R = nan(cfg.mc_maxsteps+1, 1);
    % simulate
    timespan = T;       % make sure dynamics are called each sample time
    kstart = 1; koffset = 0;
    [t,x,te,ye,ie] = ode45(@sw_dynamics, timespan, X0(:, i), odeopt, model, c, rad, h); 
    tend = t(end);
    kend = floor(tend/Ts)+1;
    % resume simulation only as long as the timespan is not exhausted,
    % and the detected event was a swing leg impact (index 1) and not failure (index 2)
    while tend < Tend  &&  ie(1) == 1,
        X(kstart:kend, :) = x((kstart:kend) - koffset, :);
        % there was an impact, compute discontinuous state transfer
        x_afterimpact = sw_impact(x(end,:));
        % continue simulation
        firstsample = ceil(tend/Ts) * Ts;
        kstart = kend + 1; 
        if tend < firstsample,
            timespan = [tend firstsample:Ts:Tend];
            koffset = kstart - 2;
        else
            timespan = firstsample:Ts:Tend;
            koffset = kstart - 1;
        end;
        [t,x,te,ye,ie] = ode45(@sw_dynamics, timespan, x_afterimpact, odeopt, model, c, rad, h);
        tend = t(end);
        kend = floor(tend/Ts)+1;
    end;
    size(X(kstart:kend, :))
    size(x((kstart:kend) - koffset, :))
    X(kstart:kend, :) = x((kstart:kend) - koffset, :);
    % pad with failure state at the next sample time,
    % if simulation ended with failure
    if ~isempty(ie) && ie(1) == 2,
        Kend = kend+1;
        X(kend+1, :) = x(end, :);
        fail = 1;
    else
        Kend = kend;
        fail = 0;
    end;

    % compute reward
    R(2:Kend) = model.ru * abs(SWMCRBFNEAREST_U(1:Kend-1));
    if fail,
       R(Kend) = R(Kend) + model.rfall;
    end;

    % return of x0 is inner product of discounting and rewards
    J(i) = disc(1:Kend-1) * R(2:Kend);
    
    hist.x = X(1:Kend, :)';
    hist.u = [SWMCRBFNEAREST_U(1:Kend - 1); NaN];
    hist.t = 0:Ts:(Kend-1)*Ts;
    hist.r = [0; R(2:Kend)];
    figure(10); clf;
    subplot(221); hold on; plot(hist.t, hist.x(1, :), 'k', hist.t, hist.x(3, :), 'b'); legend('\theta', '\theta''');
    subplot(222); hold on; plot(hist.t, hist.x(2, :), 'k', hist.t, hist.x(4, :), 'b'); legend('\phi', '\phi''');
    subplot(223); hold on; stairs(hist.t, hist.u, 'k'); legend('\tau');
    subplot(224); hold on; plot(hist.t, hist.r, 'k'); legend('r');
end;

% expected value is the mean
eJ = mean(J);

end
% END rbfmceval() RETURNING expected cost eJ, cost vector J ===========================


% ------------------------------------
% LOCAL FUNCTIONS

% ------------------------------------
function xd = sw_dynamics(t, x, m, c, rad, h)
% Implements the dynamics of the simple walker
theta = x(1); phi = x(2);
% thetadot = x(3); phidot = x(4);

% determine if needing to compute action
global SWMCRBFNEAREST_U
k = floor(t/m.Ts)+1;    % 1-based discrete time index
if ~mod(t, m.Ts),       % we are precisely at a sample time, compute new u and store on global vector
	[actmax imax] = max( exp(-sum( ((repmat(x, 1, size(c, 2))  - c) ./ rad).^2 )) );
    SWMCRBFNEAREST_U(k) = h(:, imax);
end;

xd = [x(3)
      x(4)
      sin(theta-m.slope)
      sin(phi)*(x(3)^2-cos(theta-m.slope)) + sin(theta-m.slope) + SWMCRBFNEAREST_U(k)];
end

% -------------------------------------
function xnew = sw_impact(xold)
% Discontinuous state transfer and swapping swing - stance legs here
thetaold    = xold(1);
thetadotold = xold(3);
xnew = [-thetaold
        -2*thetaold
        cos(2*thetaold)*thetadotold
        cos(2*thetaold)*(1 - cos(2*thetaold))*thetadotold];
end

% ------------------------------------
function [value, isterminal, direction] = sw_events(t, x, m, c, rad, h)
% Swing leg impact detection; fall detection
theta = x(1); phi = x(2);
% thetadot = x(3); phidot = x(4);
value       = [cos(theta)-cos(phi-theta);  pi/2-abs(theta)];
isterminal  = [(theta < -0.03) & (x(3) < 0) & (phi < -0.01);  1];
direction   = [-1; -1];
end