function varargout = fuzzyqi_replay(cfg)
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see he str2cfg
% Outputs:
%   HIST, FIGH          - HIST is the replay history. 
%           FIGH contains figure handles if figures were created 
%           and not closed; and is an empty matrix if all the figures were closed.
%

if nargin < 1, cfg = struct(); end;

% -----------------------------------------------
% Process configuration structure

% default config
% script config
CFG.datafile = 'fzqidata';          % load data from this file
% replay config
CFG.interph = 1;                    % interpolated policy
CFG.x0 = [];                        % initial state for replay (otherwise zeros)
CFG.tend = 1;                       % end time for replay, in seconds
% stats & figure output config
CFG.plottarget = 'screen';          % 'screen', '', 'latex', or 'beamer'. If 'screen' figures will not be closed
CFG.savedir = '';
CFG.savefig = '';
% display config
CFG.verb = 1;
CFG.noplot = 0;                     % whether to suppress figure plots

% If caller provided string, parse it into a structure
if ischar(cfg),
    cfg = str2cfg(cfg, fieldnames(CFG));
end;
% Install function defaults for everything else
cfg = checkparams(cfg, CFG);

% Echo config
dispx('Fuzzy Q-iteration Replay called with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1);

if ~exist([cfg.datafile '.mat'], 'file'),
    error('File does not exist');
end;


% %%%%%%%%%%%%%%%%%%%%%
% START COPY HERE

load(cfg.datafile, 'model', 'X', 'U', 'theta', 'DIMS'); 
loadwithprefix(cfg.datafile, 'exp', 'cfg'); 

% -----------------------------------------------
% Replay
    
% %%%%%%%%%%%%%%%%%%%%%
% PREPARE POLICY STRUCTURE, NOTE INTERPH FLAG
H = fuzzyqi_prepareh(theta, X, U, DIMS, cfg.interph);

% %%%%%%%%%%%%%%%%%%%%%
% INITIALIZE PROCESS HERE
if ~isempty(cfg.x0),      % specified initial state
    x0 = cfg.x0(:);
else                      % zeros
    x0 = zeros(model.p, 1);
end;
dispx(['Controlling from x0=' num2str(reshape(x0, 1, [])) ], cfg.verb, 0);

% history
t = 0 : model.Ts : cfg.tend;
Ns = length(t)-1;       % number of samples / time instances at which control is applied
x = zeros(DIMS.p, length(t)); x(:, 1) = x0;
u = zeros(DIMS.q, length(t)); u(:, end) = NaN;
r = zeros(1, length(t)); r(1) = NaN;

% steps loop
for k = 1:Ns,
    % %%%%%%%%%%%%%%%%%%%%%
    % COMPUTE ACTION USING POLICY
    u(:, k) = fuzzyqi_h(x(:, k), [], H);
    % %%%%%%%%%%%%%%%%%%%%%
    % APPLY ACTION TO PROCESS HERE
    [x(:, k+1) r(k+1) terminal] = feval(model.fun, model, x(:, k), u(:, k));
    if terminal,
        Ns = k; break;      % entered terminal state, finish early
    end;
end;

% plot history & optionally save figures
cfg.gamma = expcfg.gamma; R = discreturn(cfg, r, Ns, terminal);
hist.t = t(1:Ns+1); hist.x = x(:, 1:Ns+1); hist.u = u(:, 1:Ns+1); hist.r = r(1:Ns+1); hist.R = R(1:Ns+1);
if ~cfg.noplot,
    toscreen = strcmp(cfg.plottarget, 'screen') || isempty(cfg.plottarget);
    figh = plothistory(hist);
    saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);
    % if the model supplies a plot function, use it to plot the trajectory
    if isfield(model, 'plotfun'),
        mpfigh = feval(model.plotfun, hist);
        % it might be that the model plot fun plotted nothing
        if ~isempty(mpfigh), figh(end+1) = mpfigh; end;
    end;
%             pause;
    % close the figures if their target was not the screen
    if ~toscreen, close(figh); end;
end;
        
% -----------------------------------------------
% set output
fig = ~cfg.noplot && (strcmp(cfg.plottarget, 'screen') || isempty(cfg.plottarget));
if fig,         varargout = {hist, figh};
else            varargout = {hist, []};
end;


% END fuzzyqi() RETURNING varargout =================================================================
